// ===================================================================
// $Id: shbasis.h,v 1.8 2005/11/29 15:51:50 xkrivanj Exp $
/// \file shbasis.h
///
/// \brief Abstraction of (hemi)spherical basis and various implementations
/// of spherical and hemispherical harmonics.
///
/// SPHAR_REPLACEMENT_STRING
///
/// Initial coding by Jaroslav Krivanek, June 2004.

#ifndef __SHBASIS_H__
#define __SHBASIS_H__

// SPHAR headers
#include "shconfig.h"
#include "sheval.h"
#include "shlib.h"
#include "shtab.h"

// standard headers

__BEGIN_SPHAR_HEADER

/////////////////////////////////////////////////////////////////////////////
//
//  class CSphBasis
//
/////////////////////////////////////////////////////////////////////////////

/// Interface encapsulating a (hemi)spherical basis.
template<class TV, class TA>
class CSphBasis {
  
  public: // types
    
    /// basis type
    enum EType {
      /// spherical harmonics
      ESH  = 1,
      /// hemispherical harmonics
      EHSH = 2
    };

    /// basis definition domain
    enum EDomain {
      /// full sphere
      ESphere     = 1,
      /// ypper hemisphere
      EHemisphere = 2
    };

    /// implementation of basis evaluation
    enum EImplementation {
      /// analytic formula
      EAnalytic    = 1,
      /// using tabulated legendre polynomials
      ETabLegendre = 2,
      /// full basis is tabulated
      ETabulated   = 3
    };

  protected: // data

    TV *internal_H;
    TV *internal_dHdt;
    TV *internal_dHdp;
    
  public: // methods

    /// allocate internal buffers
    CSphBasis(int order);
    
    /// Virtual destructor - delete the internal buffers
    virtual ~CSphBasis();

    /// Return the coefficient index for given l and m
    inline static int Index(int l, int m) { return shIndex(l,m); }
    
    /// Evaluate the basis for a given l,m at dir (theta, phi)
    virtual TV EvalBasisSph(int l, int m, const TA theta, const TA phi) const = 0;
    
    /// Evaluate the basis for a given l,m at dir v3d (v3d must be normalized).
    virtual TV EvalBasisCart(int l, int m, const TA* v3d)  const = 0;
    
    /// Evaluate basis functions up to a given order (exclusive) at dir 
    /// (theta,phi).
    virtual void  EvalBasisSphV(int order, const TA theta, const TA phi,
      TV* ret) const =0;
    
    /// Evaluate basis functions up to order at dir v3d 
    /// (v3d must be normalized).
    virtual void  EvalBasisCartV(const int order, const TA* v3d, 
      TV* ret)  const =0;
    
    /// Evaluate basis functions and their derivatives wrt theta and phi.
    virtual void  EvalDerivCartV(const int order, const TA* v3d, TV* y,
      TV* dytheta, TV* dyphi)  const = 0;
    
    /// Evaluate basis functions and their derivatives wrt theta and phi.
    virtual void  EvalDerivSphV(const int order, const TA theta, 
      const TA phi,TV* y, TV* dytheta, TV* dyphi) const = 0;
    
    /// Evaluate basis functions into an internal buffer
    virtual void  EvalBasisCartVInt(const int order, const TA* v3d, 
      const TV** H)  const;

    /// Evaluate basis functions and their derivatives into internal buffers
    virtual void  EvalDerivCartVInt(const int order, const TA* v3d, 
      const TV** H, const TV** dHdt, const TV** dHdp)  const;

    /// Get Spherical basis type.
    virtual EType GetType() const = 0;

    /// Get spherical basis type as a string;
    virtual const char* GetTypeString() const = 0;

    /// Returns the basis definition domain.
    virtual EDomain GetDomain() const = 0;

    /// return the implementation of the basis evaluation
    virtual EImplementation GetImplementation() const = 0;

    /// Returns the analytic version of the basis evaluator.
    virtual const CSphBasis<TV,TA>* GetAnalytic() const = 0;
};

// --------------------------------------------------------------------------
//  CSphBasis::CSphBasis()
// --------------------------------------------------------------------------
template <class TV,class TA>
CSphBasis<TV,TA>::CSphBasis(int order) 
{
  assert(order>0);
  int n = PAD_FOUR(order*order);
  internal_H = NEW_ALIGN16(TV, n);
  internal_dHdt = NEW_ALIGN16(TV, n);
  internal_dHdp = NEW_ALIGN16(TV, n);
  assert(internal_H);
  assert(internal_dHdt);
  assert(internal_dHdp);
  shv<TV>::setz(n,internal_H);
  shv<TV>::setz(n,internal_dHdt);
  shv<TV>::setz(n,internal_dHdp);
}

// --------------------------------------------------------------------------
//  CSphBasis::~CSphBasis() 
// --------------------------------------------------------------------------
template<class TV, class TA>
CSphBasis<TV,TA>::~CSphBasis() 
{
  FREE_ALIGN16(internal_H);
  FREE_ALIGN16(internal_dHdt);
  FREE_ALIGN16(internal_dHdp);
}

// --------------------------------------------------------------------------
//  CSphBasis::EvalBasisCartVInt() 
// --------------------------------------------------------------------------
template<class TV,class TA>
void  
CSphBasis<TV,TA>::EvalBasisCartVInt(const int order, const TA* v3d, 
                                const TV** H)  const
{
  assert(H);
  EvalBasisCartV(order,v3d,internal_H); 
  *H = internal_H;
}

// --------------------------------------------------------------------------
//  CSphBasis::EvalDerivCartVInt() 
// --------------------------------------------------------------------------
template<class TV,class TA>
void  
CSphBasis<TV,TA>::EvalDerivCartVInt(const int order, const TA* v3d, 
                                const TV** H, const TV** dHdt, 
                                const TV** dHdp)  const
{
  assert(H);
  assert(dHdt);
  assert(dHdp);
  EvalDerivCartV(order,v3d,internal_H,internal_dHdt,internal_dHdp);
  *H = internal_H;
  *dHdt = internal_dHdt;
  *dHdp = internal_dHdp;
}

/////////////////////////////////////////////////////////////////////////////
//
//  class CSphericalBasis
//
/////////////////////////////////////////////////////////////////////////////

/// Abstraction of all SPHERICAL bases.
template <class TV, class TA>
class CSphericalBasis : public CSphBasis<TV, TA> {
  public:

    CSphericalBasis(int order) : CSphBasis<TV, TA>(order) {}

    virtual typename CSphBasis<TV,TA>::EDomain GetDomain() const
    { return ESphere; }

    virtual ~CSphericalBasis() {};
};


/////////////////////////////////////////////////////////////////////////////
//
//  class CSphericalBasis
//
/////////////////////////////////////////////////////////////////////////////

/// Abstraction of all HEMIspherical bases.
template <class TV, class TA>
class CHemisphericalBasis : public CSphBasis<TV, TA> {

  public:
    CHemisphericalBasis(int order) : CSphBasis<TV, TA>(order) {}

    virtual typename CSphBasis<TV,TA>::EDomain GetDomain() const
    { return EHemisphere; }

    virtual ~CHemisphericalBasis() {};
};

/////////////////////////////////////////////////////////////////////////////
//
//  class CSHBasis
//
/////////////////////////////////////////////////////////////////////////////

/// Spherical harmonic analytical evaluation.
template <class TV, class TA>
class CSHBasis : public CSphericalBasis<TV,TA> {
  
  public: // methods

    CSHBasis(int order) : CSphericalBasis<TV,TA>(order) {}

    virtual TV EvalBasisSph(int l, int m, const TA theta, const TA phi) const
    { return (TV)shEvalSph(l,m,theta,phi); }
    
    virtual TV EvalBasisCart(int l, int m, const TA* v3d) const
    { return (TV)shEvalCart(l,m,v3d[0],v3d[1],v3d[2]); }
    
    virtual void  EvalBasisSphV(int order, const TA theta, const TA phi,
      TV* ret) const
    { shEvalSphV<TV>(order,theta,phi,ret); }
    
    virtual void EvalBasisCartV(const int order, const TA* v3d, TV* ret) const
    { shEvalCartV<TV,TA>(order,v3d,ret); }
    
    virtual void EvalDerivCartV(const int order, const TA* v3d, TV* y, 
      TV* dytheta, TV* dyphi)  const
    { dSHEvalCartV<TV,TA>(order, v3d, y, dytheta, dyphi); }
    
    virtual void  EvalDerivSphV(const int order, const TA theta, 
      const TA phi,TV* y, TV* dytheta, TV* dyphi) const
    { dSHEvalSphV<TV>(order, theta, phi, y, dytheta, dyphi); }
    
    virtual typename CSphBasis<TV,TA>::EType GetType() const 
    { return ESH; }

    virtual const char* GetTypeString() const { return "sh"; }

    virtual typename CSphBasis<TV,TA>::EImplementation 
    GetImplementation() const { return EAnalytic; }

    virtual const CSphBasis<TV,TA>* GetAnalytic() const {return this;};

    virtual ~CSHBasis() {};
};

// typedef CSHBasis<float,float>  CSHBasisf;
// typedef CSHBasis<double,double> CSHBasisd;

/////////////////////////////////////////////////////////////////////////////
//
//  class CHSHBasis
//
/////////////////////////////////////////////////////////////////////////////

/// Spherical harmonic analytical evaluation.
template <class TV, class TA>
class CHSHBasis : public CHemisphericalBasis<TV,TA> {

  public: // methods
    
    CHSHBasis(int order) : CHemisphericalBasis<TV,TA>(order) {}

    virtual TV EvalBasisSph(int l, int m, const TA theta, const TA phi) const
    { return (TV)hshEvalSph(l,m,theta,phi); }
    
    virtual TV EvalBasisCart(int l, int m, const TA* v3d) const
    { return (TV)hshEvalCart(l,m,v3d[0],v3d[1],v3d[2]); }
    
    virtual void  EvalBasisSphV(int order, const TA theta, const TA phi, 
      TV* ret) const
    { hshEvalSphV<TV,TA>(order,theta,phi,ret); }
    
    virtual void  EvalBasisCartV(const int order, const TA* v3d, TV* ret) const
    { hshEvalCartV<TV,TA>(order,v3d,ret); }
    
    virtual void  EvalDerivCartV(const int order, const TA* v3d, TV* y, 
      TV* dytheta, TV* dyphi) const
    { dHSHEvalCartV<TV,TA>(order, v3d, y, dytheta, dyphi);  }
    
    virtual void  EvalDerivSphV(const int order, const TA theta, const TA phi,
      TV* y, TV* dytheta, TV* dyphi) const
    { dSHEvalSphV<TV>(order, theta, phi, y, dytheta, dyphi); }
    
    virtual typename CSphBasis<TV,TA>::EType GetType() const 
    { return EHSH; }

    virtual const char* GetTypeString() const { return "hsh"; }

    virtual typename CSphBasis<TV,TA>::EImplementation 
    GetImplementation() const { return EAnalytic; }

    virtual const CSphBasis<TV,TA>* GetAnalytic() const {return this;};

    virtual ~CHSHBasis() {};
};

// typedef CHSHBasis<float,float>  CHSHBasisf;
// typedef CHSHBasis<double,double> CHSHBasisd;

/////////////////////////////////////////////////////////////////////////////
//
//  class CSHBasisApprox
//
/////////////////////////////////////////////////////////////////////////////

/// Abstraction of approximations to Spherical harmonics.
template <class TV,class TA>
class CSHBasisApprox : public CSHBasis<TV,TA> {

private: // data
  CSHBasis<TV,TA> *_analytic;
  
public: // methods

  CSHBasisApprox(int order) : 
        CSHBasis<TV,TA>(order), _analytic(new CSHBasis<TV,TA>(order)) {}

  virtual const CSphBasis<TV,TA>* GetAnalytic() const {return _analytic;};

  virtual ~CSHBasisApprox() { delete _analytic; };
};



/////////////////////////////////////////////////////////////////////////////
//
//  class CHSHBasisApprox
//
/////////////////////////////////////////////////////////////////////////////

/// Abstraction of approximations to Hemipherical harmonics.
template <class TV, class TA>
class CHSHBasisApprox : public CHSHBasis<TV, TA> {
  
private: // data
  CHSHBasis<TV,TA> *_analytic;
  
public: // methods
  CHSHBasisApprox(int order) : 
        CHSHBasis<TV,TA>(order), _analytic(new CHSHBasis<TV,TA>(order)) {}

  virtual const CSphBasis<TV,TA>* GetAnalytic() const {return _analytic;};

  virtual ~CHSHBasisApprox() { delete _analytic; };
};


/////////////////////////////////////////////////////////////////////////////
//
//  class CSHBasisLTab
//
/////////////////////////////////////////////////////////////////////////////

#define LEGENDRE_RES 5000

/// Spherical harmonics evaluated using tabulated Legendre polynomials.
template<class TV,class TA>
class CSHBasisLTab : public CSHBasisApprox<TV,TA> {
 
protected:
  CTabLegendre<TV,TA> leg;
  
public: // methods
		CSHBasisLTab(int maxOrder) : 
        CSHBasisApprox<TV,TA>(maxOrder), leg(LEGENDRE_RES, maxOrder,true) {}

    virtual void  EvalBasisSphV(int order, const TA theta, const TA phi, 
      TV* ret) const;

    virtual void  EvalBasisCartV(const int order, const TA* v3d, 
      TV* ret) const;

    virtual void  EvalDerivCartV(const int order, const TA* v3d, TV* y, 
      TV* dytheta, TV* dyphi) const;

    virtual void  EvalDerivSphV(const int order, const TA theta, 
      const TA phi, TV* y, TV* dytheta, TV* dyphi) const;

    virtual typename CSphBasis<TV,TA>::EImplementation 
    GetImplementation() const { return ETabLegendre; }

    virtual ~CSHBasisLTab() {};
};

/////////////////////////////////////////////////////////////////////////////
//
//  class CHSHBasisLTab
//
/////////////////////////////////////////////////////////////////////////////

/// Hemispherical harmonics evaluated using tabulated Legendre polynomials.
template<class TV,class TA>
class CHSHBasisLTab : public CHSHBasisApprox<TV,TA> {
  
protected:
  CTabLegendre<TV,TA> leg;
  
public: // methods
		CHSHBasisLTab(int maxOrder) : 
        CHSHBasisApprox<TV,TA>(maxOrder), leg(LEGENDRE_RES, maxOrder,true) {}

    virtual void  EvalBasisSphV(int order, const TA theta, const TA phi,
      TV* ret) const;

    virtual void  EvalBasisCartV(const int order, const TA* v3d, 
      TV* ret) const;

    virtual void  EvalDerivCartV(const int order, const TA* v3d, TV* y,
      TV* dytheta, TV* dyphi) const;

    virtual void  EvalDerivSphV(const int order, const TA theta, 
      const TA phi, TV* y, TV* dytheta, TV* dyphi) const;

    virtual typename CSphBasis<TV,TA>::EImplementation 
    GetImplementation() const { return ETabLegendre; }

    virtual ~CHSHBasisLTab() {};
};


/////////////////////////////////////////////////////////////////////////////
//
//  class CSHBasisTab
//
/////////////////////////////////////////////////////////////////////////////

/// Tabulated Spherical harmonics.
template<class TV,class TA>
class CSHBasisTab : public CSHBasisApprox<TV,TA> {
 
protected: // data

  CSphTabulated<TV,TA>  _tab;

public: // methods
		CSHBasisTab(int maxOrder, int res) : 
        CSHBasisApprox<TV,TA>(maxOrder), _tab(maxOrder,res,GetAnalytic()) {}

    virtual void  EvalBasisCartV(const int order,const TA* v3d,
      TV* dest) const
    { _tab.GetCoeffs4(order,v3d,dest,0,0); }

    virtual void  EvalDerivCartV(const int order, const TA* v3d, TV* y,
      TV* dytheta, TV* dyphi) const
    { _tab.GetCoeffs4(order,v3d,y,dytheta,dyphi); }

    virtual void  EvalBasisCartVInt(const int order, const TA* v3d, 
      const TV** H)  const
    { _tab.GetCoeffsInt4(order,v3d,H); }

    virtual void  EvalDerivCartVInt(const int order, const TA* v3d, 
      const TV** H, const TV** dHdt, const TV** dHdp)  const
    { _tab.GetCoeffsInt4(order,v3d,H,dHdt,dHdp); }

    virtual typename CSphBasis<TV,TA>::EImplementation 
    GetImplementation() const { return ETabulated; }

    virtual ~CSHBasisTab() {};
};

/////////////////////////////////////////////////////////////////////////////
//
//  class CHSHBasisTab
//
/////////////////////////////////////////////////////////////////////////////

/// Tabulated Hemispherical harmonics.
template<class TV,class TA>
class CHSHBasisTab : public CHSHBasisApprox<TV,TA> {

protected: // data

  CSphTabulated<TV,TA>  _tab;
  
public: // methods
		CHSHBasisTab(int maxOrder, int res) : 
        CHSHBasisApprox<TV,TA>(maxOrder), _tab(maxOrder,res,GetAnalytic()) {}

    virtual void  EvalBasisCartV(const int order,const TA* v3d,
      TV* dest) const
    { _tab.GetCoeffs4(order,v3d,dest,0,0); }

    virtual void  EvalDerivCartV(const int order, const TA* v3d, TV* y,
      TV* dytheta, TV* dyphi) const
    { _tab.GetCoeffs4(order,v3d,y,dytheta,dyphi); }

    virtual void  EvalBasisCartVInt(const int order, const TA* v3d, 
      const TV** H)  const
    { _tab.GetCoeffsInt4(order,v3d,H); }

    virtual void  EvalDerivCartVInt(const int order, const TA* v3d, 
      const TV** H, const TV** dHdt, const TV** dHdp)  const
    { _tab.GetCoeffsInt4(order,v3d,H,dHdt,dHdp); }

    virtual typename CSphBasis<TV,TA>::EImplementation 
    GetImplementation() const  { return ETabulated; }

    virtual ~CHSHBasisTab() {};
};

__END_SPHAR_HEADER

#endif // __SHBASIS_H__

