// ===================================================================
// $Id: shbrdf.h,v 1.6 2005/11/29 15:51:50 xkrivanj Exp $
//
// shbrdf.h
//    (Hemi)spherical harmonics representation of BRDFs.
//
// Class: 
//
// Initial coding by Jaroslav Krivanek, December 2003.
// Based on Pascal Gautron's code.

#ifndef __SHBRDF_H__
#define __SHBRDF_H__

// SPHAR headers
#include "shconfig.h"
#include "shlib.h"
#include "shvector.h"

// standard headers

// forward declarations
template<class TV,class TA> class CBRDF;
template<class TV,class TA> class CSphBasisEvalCache;

__BEGIN_SPHAR_HEADER

/////////////////////////////////////////////////////////////////////////////
//
//  class CSphBRDFStatistics
//
/////////////////////////////////////////////////////////////////////////////
template<class TV>
class CSphBRDFStatistics {
  
public:
  /// total number of ray directions for which the BRDF was precomputed
  int numDirections;
  /// number of directions for which the error excedes given max. error
  int numErrorDirections;
  /// maximum order of sh-expansion (can be lower than the maxOrder given as a parameter to the constructor)
  int maxOrder;
  /// minimum SH-order
  int minOrder;
  /// average SH-order
  TV  avgOrder;
  /// order 99%-quantile
  int order99Quantile;
  /// maximum error
  TV  maxError;
  /// minimum error
  TV  minError;
  /// average error
  TV  avgError;
  /// order historgram (length = maxAllowedOrder)
  int *orderHistogram;
  
  CSphBRDFStatistics() 
  {orderHistogram=0;}

  ~CSphBRDFStatistics() 
  { delete [] orderHistogram; }
  
  void SetUp(int maxAllowedOrder);
  
  void Describe(ostream &) const;
  
  int OrderQuantile(TV prob);
};

/////////////////////////////////////////////////////////////////////////////
//
//  class CBRDFRec
//
/////////////////////////////////////////////////////////////////////////////
template<class TV>
class CBRDFRec : public CCoefVec<TV>
{ 
public: // data

  TV *brdfEnergy;
  
  // MAX_{over channels} shEnergy[ch]/brdfEnergy[ch]
  TV error;
  
public: // methods
  
  CBRDFRec(int channels, int order, TV e=0, const TV* brdfe=0);

  ~CBRDFRec();
  
  TV GetError() const { return error; } 

  inline void SetBrdfEnergy(const TV *brdfe);

  const TV* GetBrdfEnergy() const { return brdfEnergy; }

  /// compute BRDF representation error
  /**
     The error is computed from the 'brdfEnergy' and the 
     (H)SH coefficients as a ratio of energy captured
     in the coefficients to the original brdf energy 
     (which is stored in the brdfEnergy array).
   */
  TV ComputeRepError() const;

  /// Update the representation error
  /**
    Calls ComputeRepError() and sets the result to 'error'.
   */
  void UpdateRepError() 
  { error = ComputeRepError(); }

  /// Create a new record with the length adapted to the give max error.
  CBRDFRec* AdaptOrder(int newMaxAllowedOrder,const TV maxAllowedError) const;
};

// --------------------------------------------------------------------------
//  CBRDFRec::SetBrdfEnergy()
// --------------------------------------------------------------------------
template<class TV>
inline void 
CBRDFRec<TV>::SetBrdfEnergy(const TV *brdfe) 
{
  if (!brdfEnergy)
    return; // zero length

  if (brdfe) {
    for(int ch=0; ch<GetNumChannels(); ch++)
      brdfEnergy[ch]=brdfe[ch];
  }
  else {
    for(int ch=0; ch<GetNumChannels(); ch++)
      brdfEnergy[ch]=0;
  }
}

/////////////////////////////////////////////////////////////////////////////
//
//  class CSphBRDF
//
/////////////////////////////////////////////////////////////////////////////
template<class TV,class TA>
class CSphBRDF {
  
public: // methods

  /// Convert a BRDF into (H)SH for a given outgoing direction.
  /**
     @param brdf   [in]  The BRDF to convert
     @param outDir [in]  Fixed outgoing direction
     @param cache  [in]  Pre-evaluated basis.
     @param order  [in]  Order of (H)SH
     @param coeff  [out] (H)SH coefficnets for the BRDF
     @param brdfEnergy [out] energy in the BRDF lobe
    */
  static void BrdfToSH(
    const CBRDF<TV,TA> &brdf, 
    const TA *outDir, 
    const CSphBasisEvalCache<TV,TA>& cache,
    int   order,
    TV    *coeffs,
    TV    &brdfEnergy);
  
  /// loads BRDF representation from file
  CSphBRDF(FILE *file, bool adaptiveDeletion);
  
  /// Compute the BRDF representation for a given BRDF
  /**
    Uses parabolic mapping of the uppor hemisphere to [0,1]x[0,1] to represent
    BRDF from different outgoing (ray) directions. For each ray direction,
    the brdf lobe is represented by a spherical harmonics expansion.
    The BRDF should incorporate the cosine term.
  
    @param brdfSource  (in)
      The BRDF interface exporting method getValueCol() 
      that evaluates the BRDF for given incoming and outgoing 
      direction. This evaluation should incorporate the 
      cosine term.

    @param channels (in)  
      Number of color channels to store the BRDF. 
      CSRDF::getValueCol() is supposed to return this 
      number of channels.

    @param maxOrder  (in) 
      Maximum order of spherical harmonic expansion.

    @param xRes (in) 
      Resulution of the parabolic texture.

    @param yRes   (in) 
      Resulution of the parabolic texture. 

      The actual resolution of the texture is (APhiRes+1)x(AThetaRes+1).
      This is how Pascal wrote it, it works and I'm not going 
      to change it.

    @param maxError  (in) 
      Maximum relative error of the SH-expansion. This is
      used to set adaptively the length of the 
      SH-expansion and to reject vectors that cannnot 
      attain given accuracy (=1-err) even with the order 
      of maxOrder.

    @param adaptiveOrder (in) 
      Set the order of SH expansion adaptively to tightly fullfill
      the error crieterion or set the length of all vectors to 'maxOrder'.

    @param adaptiveDeletion (in) 
      Reject (i.e. set to zero) SH-vector that cannnot attain given 
      accuracy (=1-err) even with the order of maxOrder.
  */
  CSphBRDF(
    const CBRDF<TV,TA> &sourceBRDF, 
    int   channels,
    int   maxOrder, 
    int   xRes, 
    int   yRes,
    TV    maxError,
    bool  adaptiveOrder,
    bool  adaptiveDeletion,
    const CSphBasisEvalCache<TV,TA>& cache);

  /// Just setup the data structures. Do not actually precompute anything.
  CSphBRDF(
    int   channels, 
    int   newMaxAllowedOrder, 
    int   xRes, 
    int   yRes, 
    TV    maxError, 
    bool  adaptiveOrder, 
    bool  adaptiveDeletion);

  /// Return a new adapted BRDF representation
  /**
    The new representation is adapted to a new specified maximum order
    and maximum error.

    @return 0 if data in 'this' are not sufficient for desired max error.
            Otherwise, the adapted BRDF representation is returned.
   */
  CSphBRDF * Adapt(
    int   newMaxAllowedOrder,
    TV    newMaxError, 
    bool  newAdaptiveOrder,
    bool  newAdaptiveDeletion);

  /// Destructor.
  ~CSphBRDF();

  /// Return the u,v parabolic mapping parameters given a direction.
  inline void GetUV(const TA* vect, TA &u, TA &v) const
  {
    const TA minz=4e-2f;
    if (vect[2] < minz) {
      // if vect.z is below minimum, set it to minimum
      // and normalize [x,y,minz]
      TA m = TA(1.0)/sqrtf(vect[0]*vect[0] + vect[1] * vect[1] + minz * minz);
      TA x = vect[0]*m;
      TA y = vect[1]*m;
      TA z = minz*m;
      u=TA(0.5)*(TA(1.0)+(x/(z+TA(1.0))));
      v=TA(0.5)*(TA(1.0)+(y/(z+TA(1.0))));
    }
    else {
      u=TA(0.5)*(TA(1.0)+(vect[0]/(vect[2]+TA(1.0))));
      v=TA(0.5)*(TA(1.0)+(vect[1]/(vect[2]+TA(1.0))));
    }
  }


  /// Return the BRDF order given the u,v, parabolic parameters.
  int GetOrderUV(const TA &u,const TA &v) const;

  /// Returns BRDF coefficients for a given outgoing direction
  const CBRDFRec<TV>* GetCoeffs(const TA* v3d) const {
    TA u,v;
    GetUV(v3d,u,v); 
    return GetCoeffsUV(u,v);
  }

  /// Returns BRDF coefficients for a given outgoing direction
  const CBRDFRec<TV>* GetCoeffsUV(const TA &u,const TA &v) const;

  /// Returns BRDF coefficients for a given texture position
  const CBRDFRec<TV>* GetCoeffsIdx(int x, int y) const
  { return _coeffs[x*(_yRes+1)+y]; }
  
  /// Returns error of the BRDF approximation for a given outgoing direction
  inline TV GetError(const TA* vect) const {
    if (vect[2] <= 0) 0;
    TA u=TA(0.5)*(TA(1.0)+(vect[0]/(vect[2]+TA(1.0))));
    TA v=TA(0.5)*(TA(1.0)+(vect[1]/(vect[2]+TA(1.0))));
    return GetErrorUV(u,v);
  }

  /// Returns error of the BRDF approximation for a given parabolic parameters.
  TV GetErrorUV(const TA& u, const TA &v) const;
  
  /// Saves rep to a file. Return 'true' if error, 'false' if ok.
  bool SaveToFile(FILE *f) const;
  
  bool OK() const { return _ok; }
  
  const CSphBRDFStatistics<TV>& GetStats() const { return _stats; }

  /// Return resolution for theta.
  int GetXRes() const { return _xRes; }
  
  /// Returns resolution for phi.
  int GetYRes()   const { return _yRes;   }
  
  /// Returns the maximum allowed order used to build this BRDF rep.
  TV GetMaxAllowedError() const { return _maxError; }

  /// Returns the 'adaptiveDeletion' flag.
  bool GetAdaptiveDeletion() const { return _adaptiveDeletion; }

protected: // methods
  
  int _numEntries() const { return (_xRes+2)*(_yRes+1); }

  /// fill the statistics once the BRDF rep has been computed
  void _fillStatistics();

protected:  // data

  /// is the BRDF correctly precomputed?
  bool _ok;
  
  /// maximum sh order of this brdf
  int _maxAllowedOrder;
  
  /// number of color channels
  int _channels;
  
  /// resolution of the texture
  int _xRes, _yRes;
  
  /// Maximum relative error of the SH-expansion.
  TV _maxError;
  
  /// Reject SH-vector that cannnot attain given accuracy (=1-err) even with 
  /// the order of maxOrder.
  bool  _adaptiveDeletion;	 
  
  /// Coefficient storage.
  CBRDFRec<TV> **_coeffs;
  
  /// Return value
  CBRDFRec<TV> *_retVal;
  
  /// zero-length vector
  CBRDFRec<TV> _defaultVal;
  
  /// brdf representation statistics
  CSphBRDFStatistics<TV>  _stats;
};


__END_SPHAR_HEADER

#endif // __SPHBRDF_H__

