// ================================================================
// $Id: sheval.h,v 1.10 2005/11/29 15:51:50 xkrivanj Exp $
/// \file sheval.h
/// 
/// \brief Evaluation of spherical harmonic basis functions.
///
/// SPHAR_REPLACEMENT_STRING
///
/// Initial coding by Jaroslav Krivanek, December 2003

#ifndef __SHEVAL_H__
#define __SHEVAL_H__

// SPHAR headers
#include "shconfig.h"
#include "shmath.h"
#include "shlog.h"

__BEGIN_SPHAR_HEADER

#define MAX_FIXED_SH_ORDER 30
#define SQR_MAX_FIXED_SH_ORDER (MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER)

//////////////////////////////////////////////////////////////////
//
//    DECLARATIONS
//
//////////////////////////////////////////////////////////////////

/// Evaluate SH-basis for a given l,m at dir (theta, phi)
double shEvalSph(const int l, const int m, double theta, double phi);

/// Evaluate SH-basis for a given l,m at dir v3d (v3d must be normalized).
double shEvalCart(const int l, const int m, double x, double y, double z);

/// Evaluate HSH-basis for a given l,m at dir (theta, phi)
double hshEvalSph(const int l, const int m, double theta, double phi);

/// Evaluate HSH-basis for a given l,m at dir v3d (v3d must be normalized).
double hshEvalCart(const int l, const int m, double x, double y, double z);

/// Evaluate an Associated Legendre Polynomial P(l,m,x) at x.
double assLegendreEval(const int l,const int m,const double x);

/// numerical computation of associated legendre polynomial derivative.
double dLegNum(int l, int m, double x);

/// Analytic computation of associated legendre polynomial derivative.
double dLeg(int l, int m, double x);

/// Numerical computation of SH derivative wrt theta).
double dSHThetaNum(int l, int m, double theta, double phi);

/// Numerical computation of SH derivative wrt phi.
double dSHPhiNum(int l, int m, double theta, double phi);

/// Analytic derivative of SH wrt theta.
double dSHTheta(const int l, const int m, double theta, double phi);

/// Analytic derivative of SH wrt phi.
double dSHPhi(const int l, const int m, double theta, double phi);

/// Sequential index of a sh function.
inline int shIndex(const int l, const int m) 
{
	return l*(l+1)+m;
}

//////////////////////////////////////////////////////////////////
//
//    LEGENDRE POLYNOMIALS
//
//////////////////////////////////////////////////////////////////

/// Evaluate Associated Legendre Polynomials up to band l at x.
/**
   Store in an array ret, whose format is as for spherical harmonics.
   Fill in only entries for non-negative m <0,..,order-1>.
   The resulting array must be 
       ret[ shIndex(l,m ) ] = P(l,m,x);
	   ret[ shIndex(l,-m) ] = unchanged;

   This function is a helper fuction for (H)SH evaluation.

   Template arguments:
     TV ... type for the result
     TA ... type for the argument
   The evaluation is done in 'TV' precision.
 */
template<class TV,class TA>
void assLegendreEvalV(const int order,const TA x, TV* ret)
{
	assert(order>0);

	int   lmax = order-1;

	TV pmm  = TV(1.0);
	TV somx2 = shm::sqrt((TV(1.0)-x)*(TV(1.0)+x));
	TV fact  = 1.0;

	// loop rule 2 through all m's
	int m, index;
	for( m=0, index=0 ; m < lmax; m++ ) {

		ret[shIndex(m, m)] = pmm;

		// lift l by one
		TV pmmp1 = x * (TV(2.0)*m+TV(1.0)) * pmm;		
		ret[shIndex(m+1, m)] = pmmp1;

		// lift l further
		TV pmmloc = pmm;
		TV pll = 0.0;
		for( int ll=m+2 ; ll<=lmax; ++ll) {
			pll = ( (TV(2.0)*ll-TV(1.0))*x*pmmp1-(ll+m-TV(1.0))*pmmloc )/(ll-m);
			pmmloc = pmmp1;
			pmmp1 = pll;
			
			ret[shIndex(ll, m)] = pll;
		} // for

		pmm *= (-fact) * somx2;
		fact += TV(2.0);
	}

	// m = lmax
	ret[shIndex(m, m)] = pmm;
}


/// analytic computation of associated legendre polynomial derivative
/**
  @param order  (in)  number of bands to evaluate
  @param x      (in)  evaluate dP at x; x in (-1,1); for x=1 and x=-1 the result is undefined.
                      Error might be high for |x| close to 1.
  @param res    (out) place result here
  @param argP   (in)  evaluated legendre polynomial for x
                      if this is null, the function will not fail but
					  call assLegendreEvalV().

  Result is stored the array ret, whose format is as for spherical harmonics.
  Fill in only entries for non-negative m <0,..,order-1>.
  The resulting array is 
       ret[ shIndex(l,m ) ] = P(l,m,x);
	     ret[ shIndex(l,-m) ] = unchanged;
  */
template<class TV,class TA>
void dLegV(int order, const TA x, TV* res, const TV *argP = 0)
{
	if(order<=0) return;
	
	// use the passed array of evaluated P, or evaluate it
	TV locP[MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
	const TV *P = argP;
	TV *p=0;
	if(!P) {
		if(order>MAX_FIXED_SH_ORDER)  p = new TV[order*order];
		else                p = locP;
		assLegendreEvalV(order,x,p);
		P = p;
	}

	// evaluation itself
	res[0] = 0;
	TV ix2 = 1.0/(x*x-1);

	// l=1..order; m=0..l-1
	for(int l=1, index=0; l<order; l++) {
		index+=(l+1);
		TV xl = x*l;
		for(int m=0; m<l; m++, index++) {
			res[index] = ix2 * ( xl*P[index] - (m+l)*P[shIndex(l-1,m)] );
		}
	}

	// l=1..order; m=l
	if(order>1)
		res[shIndex(1,1)] = x/shm::sqrt((1.0-x)*(1.0+x));
	if(order>2)
		res[shIndex(2,2)] = -6.0*x;
	if(order>3) {
		TV somx2 = shm::sqrt((1.0-x)*(1.0+x));
		TV dpmm =  -x * 3.0;
		TV fact = 5.0;
		for(int i=3; i< order; i++) {
			dpmm *= (-fact) * somx2;
			fact += 2.0;
			res[shIndex(i,i)] = dpmm * i;
		}
	}

	if(p && p!=locP)
		delete [] p;
}

//////////////////////////////////////////////////////////////////
//
//    SPHERICAL HAMONICS
//
//////////////////////////////////////////////////////////////////

/// Precomputed normalization for spherical harmonics.
/**
   The normalization factor incorporates the sqrt(2) that the 
   SH are multiplied with for nonzero 'm'.
  */
class CShNormTable {

protected: // data

	float  *_ftable;
	double *_dtable;

public: // methods
	CShNormTable(int maxOrder);
	inline void  GetTable(const float **t) const { *t = _ftable; }
	inline void  GetTable(const double**t) const { *t = _dtable; }
};

extern CShNormTable shNormTab;

/// Evaluate a normalisation factor for a SH basis.
template<class TV>
inline TV 
shNorm(const int l, const int m)
{
  TV t = TV(2*l+1) * TV(shm::factorial(l-m)) /
    ( TV(M_PI * 4.0) * TV(shm::factorial(l+m)) ) ;
	return  shm::sqrt(t);
}

/// Evaluate a normalisation factor for a SH basis for m=0.
template<class TV>
inline TV 
shNorm(const int l)
{
	TV t = TV(2*l+1) * TV(0.25 * M_1_PI);
	return  shm::sqrt(t);
}

/// Fill the sin/cos table of multiple angles.
/**
   Recurrence relations:
    sin(n*x) = 2*sin( (n-1)*x )*cos(x) - sin( (n-2)*x )
    cos(n*x) = 2*cos( (n-1)*x )*cos(x) - cos( (n-2)*x )	
*/
template<class TV>
inline void 
recurrentSinCos(const int order, const TV c, const TV s, TV cosn[], TV sinn[])
{
	cosn[0] = 1;
	sinn[0] = 0;
	cosn[1] = c;
	sinn[1] = s;
	for(int n=2; n<order; n++) {
		sinn[n] = 2.0f*sinn[n-1]*cosn[1] - sinn[n-2];
		cosn[n] = 2.0f*cosn[n-1]*cosn[1] - cosn[n-2];
	}
}

/// Evaluate SH-basis given x in <-1,1>, cos(phi) ('c') and sin(phi) ('s').
/**
   The 'x' = cos(theta) for SH and x = cos(theta)*2-1 for HSH.
   'mul' is a multiplicator for all coefficients. mul=1 for SH and mul=sqrt(2) for HSH
   This is a helper function not meant for direct use.
 */
template<class TV,class TA>
void shEvalHelperV(const int order, TA x, TA c, TA s, TA mul, TV* ret)
{
	TV _cosn[MAX_FIXED_SH_ORDER];  // cos(n*phi) array
	TV _sinn[MAX_FIXED_SH_ORDER];  // sin(n*phi) array
	TV *cosn, *sinn;
	if(order<MAX_FIXED_SH_ORDER) {
		cosn = _cosn;
		sinn = _sinn;
	}
	else {
		FATAL << "Normalisation not implemented yet for SH of order larger than MAX_FIXED_SH_ORDER." << std::endl;
		FATAL_EXIT;
		cosn = new TV[order];
		sinn = new TV[order];
	}

	// evaluate sin(n*phi) and cos(n*phi)
	recurrentSinCos<TV>(order,c,s,cosn,sinn);
	for(int i=0; i<order; i++) { cosn[i]*=mul; sinn[i]*=mul; }

	// evaluate associated legendre polynomial P_l^m(x)
	assLegendreEvalV(order,x,ret);

	const TV* normTable;
	shNormTab.GetTable(&normTable);

	for(int l=0, index=0; l<order; l++, index+=2*l) {
		ret[index] *= normTable[index] * mul;
		// m = 1..l  (&& -1..-l at the same time)
		for(int m=1; m<=l; ++m) {
			ret[index-m]  = ret[index+m] * sinn[m] * normTable[index+m];
			ret[index+m] *=                cosn[m] * normTable[index+m];
		}
	}

	if(sinn != _sinn) {
		delete [] cosn;
		delete [] sinn;
	}
}

/// Evaluate SH-bases up to a given order (exclusive) at dir (theta,phi).
template<class TV,class TA>
inline void shEvalSphV(const int order, const TA theta, const TA phi, TV* ret)
{
	TA x = shm::cos(theta);
	if (x >  1) x = 1;
	if (x < -1) x = -1;
	TA c = shm::cos(phi);
	TA s = shm::sin(phi);

  shEvalHelperV<TV,TA>(order,x,c,s,1.0,ret);
}

/// Evaluate SH-bases up to order at dir v3d (v3d must be normalized).
/** TV is for vector type, TA for argument type */
template<class TV, class TA>
inline void shEvalCartV(const int order, const TA* v3d, TV* ret)
{
	TA c,s;        // cos(phi), sin(phi)
	TA x = v3d[2]; // cos(theta)
	if      (x >=  1) { x =  1; c = 1; s = 0; }
	else if (x <= -1) { x = -1; c = 1; s = 0; }
	else {
		TA ist = TA(1.0)/shm::sqrt(TA(1.0)-x*x); // 1/sin(theta) != inf
		c = v3d[0] * ist;
		s = v3d[1] * ist;
	}

  shEvalHelperV<TV,TA>(order,x,c,s,1,ret);
}


/// Evaluate SH base and their derivatives wrt to theta and phi.
template<class TV,class TA>
void 
dSHEvalHelperV(const int order,
               const TA x,
               TA sin_theta, 
               const TA c,
               const TA s,
               const TA mul, 
               TV* y,
               TV* dytheta=0,
               TV *dyphi=0)
{
	assert(y);

	TV _cosn[MAX_FIXED_SH_ORDER];  // cos(n*phi) array
	TV _sinn[MAX_FIXED_SH_ORDER];  // sin(n*phi) array
	TV *cosn, *sinn;
	if(order<MAX_FIXED_SH_ORDER) {
		cosn = _cosn;
		sinn = _sinn;
	}
	else {
		FATAL << "Normalisation not implemented yet for SH of order larger than MAX_FIXED_SH_ORDER." << std::endl;
		FATAL_EXIT;
		cosn = new TV[order];
		sinn = new TV[order];
	}

	// evaluate sin(n*phi) and cos(n*phi)
	recurrentSinCos<TV>(order,c,s,cosn,sinn);
	for(int i=0; i<order; i++) { cosn[i]*=mul; sinn[i]*=mul; }

	// get the normalization values
	const TV* normTable;
	shNormTab.GetTable(&normTable);

	// 1) evaluate associated legendre polynomial P_l^m(x)
  float  P[MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
  float dP[MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
	assLegendreEvalV(order,x,P);
  // float *P = y;

	if(dytheta) {
		assert(dyphi);
		/// @@@ yes, this is a dirty hack : dHSH/dtheta = sqrt(2) * 2 * dSH/dtheta
		/// sqrt(2) ... normalization; 2 ... because of polynomial shifting; mul=sqrt(2) for HSH
		sin_theta *= shm::sqr2(mul);

		// 2) evaluate legendre derivative
    dLegV(order,x,dP,P);
		
		// 3) evaluate sh derivative wrt theta
		for(int l=0, index=0; l<order; l++, index+=2*l) {
			dytheta[index] = -sin_theta * normTable[index] * dP[index] * mul;
			for(int m=1; m<=l; ++m) {
				dytheta[index+m] = 
          -sin_theta * normTable[index+m] * dP[index+m] * cosn[m];
				dytheta[index-m] = 
          -sin_theta * normTable[index+m] * dP[index+m] * sinn[m];
			}
		}
	}

	if(dyphi) {
		for(int l=0, index=0; l<order; l++, index+=2*l) {
			y[index] = P[index] * normTable[index] * mul;
			dyphi[index] = 0;
			// m = 1..l  (&& -1..-l at the same time)
			for(int m=1; m<=l; ++m) {
				y[index-m]  = P[index+m] * sinn[m] * normTable[index+m];
				y[index+m]  = P[index+m] * cosn[m] * normTable[index+m];
				dyphi[index+m] = -m * y[index-m];
				dyphi[index-m] =  m * y[index+m];
			}
		}
	}
	else {
		for(int l=0, index=0; l<order; l++, index+=2*l) {
			y[index] = P[index] * normTable[index] * mul;
			// m = 1..l  (&& -1..-l at the same time)
			for(int m=1; m<=l; ++m) {
				y[index-m]  = P[index+m] * sinn[m] * normTable[index+m];
				y[index+m]  = P[index+m] * cosn[m] * normTable[index+m];
			}
		}		
	}

	if(sinn != _sinn) {
		delete [] cosn;
		delete [] sinn;
	}
}

/// Evaluate SH bases and their derivatives wrt theta and phi.
template<class TV,class TA>
inline void dSHEvalCartV(const int order, 
                         const TA* v3d, 
                         TV* y, 
                         TV* dytheta, 
                         TV* dyphi)
{
	TA c,s;        // cos(phi), sin(phi)
	TA sin_theta;
	TA x = v3d[2]; // cos(theta)
	// if      (x >=  1) { x =  1; sin_theta = 0; c = 1; s = 0; }
	// else if (x <= -1) { x = -1; sin_theta = 0; c = 1; s = 0; }
	// some of derivatives are undefined in 1, -1
	const TA limit=TA(0.9999);
	if      (x >=  limit) { x =  limit; sin_theta = 0; c = 1; s = 0; }
	else if (x <= -limit) { x = -limit; sin_theta = 0; c = 1; s = 0; }
	else {
		TA ist = TA(1.0)/shm::sqrt(TA(1.0)-x*x); // 1/sin(theta) != inf
		c = v3d[0] * ist;
		s = v3d[1] * ist;
		sin_theta = shm::sqrt(TV(1)-x*x);
	}
  dSHEvalHelperV<TV,TA>(order,x,sin_theta,c,s,1,y,dytheta,dyphi);
}

/// Evaluate SH bases and their derivatives wrt theta and phi.
template<class TV,class TA>
inline void dSHEvalSphV(const int order, 
						const TA theta, 
						const TA phi, 
						TV* y, 
						TV* dytheta, 
						TV* dyphi)
{
	TA sin_theta = shm::sin(theta);
	TA x         = shm::cos(theta);
	const TA limit=0.9999f;
	if (x >= limit) { x = limit;}
	if (x <= -limit) { x = -limit;}
	TA c = shm::cos(phi);
	TA s = shm::sin(phi);

  dSHEvalHelperV<TV,TA>(order,x,sin_theta,c,s,1,y,dytheta,dyphi);
}

//////////////////////////////////////////////////////////////////
//
//    HEMISPHERICAL HAMONICS
//
//////////////////////////////////////////////////////////////////

/// Evaluate a normalisation constant for a SH function.
template<class TV>
inline TV 
hshNorm(const int l, const int m)
{
	return shNorm<TV>(l,m) * sqrt(2.0);
}

/// Evaluate a normalisation constant for a SH function for m=0.
template<class TV>
inline TV 
hshNorm(const int l)
{
	// according to all logic, the ration between hsh and sh
	// normalisation should be 2; however only sqrt(2) works correctly
	return shNorm<TV>(l) * sqrt(2.0);
}

/// Evaluate HSH-bases up to a given order (exclusive) at dir (theta,phi).
template<class TV,class TA>
inline void hshEvalSphV(const int order, const TA theta, const TA phi, TV* ret)
{
	// Polynomial shifting
	TA x = shm::cos(theta) * 2 - 1;
	if (x < -1) x = -1;
	if (x > 1) x = 1;
	TA c = shm::cos(phi);
	TA s = shm::sin(phi);

  shEvalHelperV<TV,TA>(order,x,c,s,TA(M_SQRT2),ret);
}

/// Evaluate HSH-bases up to order at dir v3d (v3d must be normalized).
template<class TV, class TA>
inline void hshEvalCartV(const int order, const TA* v3d, TV* ret)
{
	TA c,s;        // cos(phi), sin(phi)
	TA x = v3d[2] * 2 - 1;
	if (x >  1) { x =  1; c = 1; s = 0; }
	else {
		if (x < -1) x = -1;
		TA ist = TA(1.0)/shm::sqrt(TA(1.0)-shm::sqr2(v3d[2])); // 1/sin(theta) != inf
		c = v3d[0] * ist;
		s = v3d[1] * ist;
	}

  shEvalHelperV<TV,TA>(order,x,c,s,TA(M_SQRT2),ret);
}

/// Evaluate HSH bases and their derivatives wrt theta and phi.
template<class TV, class TA>
void  dHSHEvalCartV(const int order, const TA* v3d, TV* y, 
                    TV* dytheta, TV* dyphi)
{
	TA c,s;        // cos(phi), sin(phi)
	TA sin_theta;
	TA x = v3d[2] * 2 - 1;
	const TA limit=TA(0.9999);
	if (x >  limit) { x = limit; c = 1; s = 0; sin_theta=0; }
	else {
		if (x < -limit) x = -limit;
		sin_theta = shm::sqrt(TA(1)-shm::sqr2(v3d[2]));
		if(sin_theta<0.0001) { c = 1; s = 0; }
		else {
			TA ist = 1.0f/sin_theta;
			c = v3d[0] * ist;
			s = v3d[1] * ist;
		}
	}
  dSHEvalHelperV<TV,TA>(order,x,sin_theta,c,s,TA(M_SQRT2),y,dytheta,dyphi);
}

/// Evaluate HSH bases and their derivatives wrt theta and phi.
template<class TV,class TA>
void  dHSHEvalSphV(const int order, const TA theta, const TA phi,
                   TV* y, TV* dytheta, TV* dyphi)
{
	TA   sin_theta = shm::sin(theta);
	TA   x = shm::cos(theta) * 2 - 1;
	const TA limit=0.9999;
	if      (x >  limit) { x =  limit;}
	else if (x < -limit) { x = -limit;}
	TA c = shm::cos(phi);
	TA s = shm::sin(phi);
  dSHEvalHelperV<TV,TA>(order,x,sin_theta,c,s,M_SQRT2,y,dytheta,dyphi);
}

__END_SPHAR_HEADER

#endif // __shm_H__
