// ===================================================================
// $Id: shlib.h,v 1.8 2005/11/29 15:51:50 xkrivanj Exp $
//
/// \file shlib.h
/// \brief Various auxiliary classes for (hemi)spherical harmonics.
///
/// SPHAR_REPLACEMENT_STRING
///
/// Initial coding by Jaroslav Krivanek, December 2003.

#ifndef __SHLIB_H__
#define __SHLIB_H__

// SPHAR headers
#include "shconfig.h"
#include "sheval.h"
#include "shvec.h"

// standard headers

__BEGIN_SPHAR_HEADER

// forward declarations
template<class TV,class TA> class CSphBasis;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//  class CTabLegendre
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
/// Tabulated associated legendre polynomials for faster calculations
template<class TV,class TA>
class CTabLegendre {
  
protected: // data

  /// precomputed data array
  TV *_array;
  /// precomputed derivatives
  TV *_darray;
  /// maximum order
  int _maxOrder;
  /// array stride is multiply of four greater or equar to order^2
  int _stride;
  /// resolution
  int _res;
  /// (TA)(resolution-1)
  TA _res_min_1;
  
public: // methods
  
  /// Constructor. 'res' = resolution; 'maxOrder' = maximum order.
  CTabLegendre(int res, int maxOrder, bool derivative);
  /// Destructor.
  ~CTabLegendre();
  /// Evaluate ALP with nearest neigbor interpolation.
  inline TV EvalNN       (const int l,const int m,const TA x)  const;
  /// Evaluate ALP with linear interpolation.
  inline TV EvalLin      (const int l,const int m,const TA x)  const;
  /// Evaluate all ALP's up to given order with nearest neigbor interpolation.
  inline void  EvalNNV      (const int order,const TA x, TV* ret) const;
  /// Evaluate all ALP's up to given order with linear interpolation.
  void  EvalLinV     (const int order,const TA x, TV* ret) const;
  /// Return all ALP's up to given order - nearest neigbor interpolation.
  inline void EvalNNV(const int order,const TA x, const TV**p, const TV**dp) const;
};

// --------------------------------------------------------------------------
//  CTabLegendre::EvalNN()
// --------------------------------------------------------------------------
template<class TV,class TA>
inline TV 
CTabLegendre<TV,TA>::EvalNN(const int l,const int m,const TA x)  const
{
  int   idx = (int)((x+TA(1.0)) * TA(0.5) * _res_min_1);
  return _array[idx * _stride + shIndex(l,m)];
}

// --------------------------------------------------------------------------
//  CTabLegendre::EvalLin()
// --------------------------------------------------------------------------
template<class TV,class TA>
inline TV 
CTabLegendre<TV,TA>::EvalLin(const int l,const int m,const TA x)  const
{
  TA f   = (x+TA(1.0)) * TA(0.5) * _res_min_1;
  TA fi  = floor(f);
  TA du  = f-fi;
  int   idx = (int)fi;
  
  if(idx>=_res-1)
    return _array[(_res-1)*_stride+shIndex(l,m)];
  else 
    return 
    (TA(1.0)-du) * _array[ idx   *_stride + shIndex(l,m) ] +
    (        du) * _array[(idx+1)*_stride + shIndex(l,m) ];
}

// --------------------------------------------------------------------------
//  CTabLegendre::EvalNNV()
// --------------------------------------------------------------------------
template<class TV,class TA>
inline void  
CTabLegendre<TV,TA>::EvalNNV(const int order,const TA x, TV* ret) const
{
  int   idx = (int)((x+TA(1.0)) * TA(0.5) * _res_min_1);
  shv<TV>::movvv(order*order, ret, _array + (idx*_stride));
}

// --------------------------------------------------------------------------
//  CTabLegendre::EvalNNV()
// --------------------------------------------------------------------------
template<class TV,class TA>
inline void 
CTabLegendre<TV,TA>::EvalNNV(const int order,const TA x,
                      const TV**p, const TV**dp) const
{
  int   idx = (int)((x+TA(1.0)) * TA(0.5) * _res_min_1);
  assert(idx>=0);
  assert(idx<_res);
  if(p)  *p  = _array  + (idx * _stride);
  if(dp) *dp = _darray + (idx * _stride);
}


// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//  class CSphBasisEvalRec
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
/// Container for a single (hemi)spherical evaluation.
template<class TV,class TA>
class CSphBasisEvalRec {
  
public: // data
  /// direction in cartesian coordinates
  TA v[3];
  /// theta angle of the direction (measured from normal)
  TA theta;
  /// phi angle of the direction (measured from the x-axis)
  TA phi;
  /// vector of (H)SH coefficients for the direction
  TV *coeffs;
};

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
//  class CSphBasisEvalCache
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
/// Storage for pre-calculated evaluated basis functions.
template<class TV,class TA>
class CSphBasisEvalCache {
  
protected: // data
  /// Array of precalculated evaluation records.
  CSphBasisEvalRec<TV,TA> *_evalRecs;
  /// Normalisation value = 2*Pi/numSamples (upper hemisphere only!!!).
  TV _normValue;
  /// Number of samples taken.
  int _numSamples;
  /// Order of the basis.
  int _order;
  /// (Hemi)spherical basis for which we precompute.
  const CSphBasis<TV,TA>* _sphBasis;
  
public: // methods
  
  CSphBasisEvalCache(int numSamples,int order,
    const CSphBasis<TV,TA>* shBasis);
  
  ~CSphBasisEvalCache();
  
  inline int GetNumSamples() const { return _numSamples; }
  
  inline int GetOrder()      const { return _order; }
  
  inline const CSphBasisEvalRec<TV,TA>& operator[] (int i) const 
  { return _evalRecs[i]; }
  
  inline const TV& GetNormalisation() const { return _normValue; }
};

__END_SPHAR_HEADER

#endif // __SHLIB_H__
