 // =========================================================================
 // $Id: shmath.h,v 1.7 2005/11/29 15:51:50 xkrivanj Exp $
/// \file shmath.h
/// 
/// \brief Supporting Sphar math class encapsulating <math.h> and adding more.
///
///  #define __HAS_LONG_DOUBLE_MATH  if your math library has 
///   special calls for long double arguments (line pow->powl, sin->sinl etc.).
///
/// SPHAR_REPLACEMENT_STRING
///
/// Initial coding by Jaroslav Krivanek, December 2000.

#ifndef __SHMATH_H__
#define __SHMATH_H__

// SPHAR headers
#include "shconfig.h"

// standard headers
#include <cmath>

__BEGIN_SPHAR_HEADER

// For better partability. Please keep it here.
#ifndef M_PI
#define M_PI        3.14159265358979323846
#endif
#ifndef M_1_PI
#define M_1_PI	    0.31830988618379067154
#endif
#ifndef M_SQRT2
#define M_SQRT2     1.41421356237309504880
#endif

/// Spherical harmonic math.
/**
  This class serves as an interface for math fuctions from the "cmath" header,
  without being obliged to pay atention to call the right math function
  according to a type of argument (double, float).
*/
class shm
{
  public: // member functions
  
  static inline int    iabs(int a) { return ::abs(a); }
  
  static inline float  acos (float a) { return ::acosf(a); }
  static inline float  asin (float a) { return ::asinf(a); }
  static inline float  atan (float a) { return ::atanf(a); } 
  static inline float  atan2(float a1,float a2) { return  ::atan2f(a1,a2); }
  static inline float  ceil (float a) { return  ::ceilf(a); }
  static inline float  cos  (float a) { return  ::cosf(a);  }
  static inline float  cosh (float a) { return  ::coshf(a); }
  static inline float  exp  (float a) { return  ::expf(a);  }
  static inline float  fabs (float a) { return  ::fabsf(a); }
  static inline float  floor(float a) { return  ::floorf(a);}
  static inline float  fmod (float a1, float a2) { return  ::fmodf(a1, a2); }
  static inline float  frexp(float a1, int* a2) { return  ::frexpf(a1,a2); }
  // static inline float  hypot(float a1, float a2) { return  ::_hypotf(a1, a2); }
  // static inline float  j0   (float a) { return  ::_j0f(a); }
  // static inline float  j1   (float a) { return  ::_j1f(a); }
  // static inline float  jn   (int a1, float a2) { return  ::_jnf(a1, a2); }
  static inline float  ldexp(float a1,int a2) { return  ::ldexpf(a1, a2); }
  static inline float  logl (float a) { return  ::logf(a); }
  static inline float  log10(float a) { return  ::log10f(a); }
  static inline float  modf (float a1, float* a2) { return  ::modff(a1, a2); }
  static inline float  pow  (float a1, float a2) { return  ::powf(a1,a2); }
  static inline float  sin  (float a) { return  ::sinf(a); }
  static inline float  sinh (float a) { return  ::sinhf(a); }
  static inline float  sqrt (float a) { return  ::sqrtf(a); }
  static inline float  tan  (float a) { return  ::tanf(a); }
  static inline float  tanh (float a) { return  ::tanhf(a); }
  // static inline float  y0   (float a) { return  ::_y0f(a); }
  // static inline float  y1   (float a) { return  ::_y1f(a); }
  // static inline float  yn   (int a1, float a2) { return  ::_ynf(a1, a2); }
  
  static inline double  acos (double a) { return ::acos(a); }
  static inline double  asin (double a) { return ::asin(a); }
  static inline double  atan (double a) { return ::atan(a); } 
  static inline double  atan2(double a1,double a2) { return  ::atan2(a1,a2); }
  static inline double  ceil (double a) { return  ::ceil(a); }
  static inline double  cos  (double a) { return  ::cos(a);  }
  static inline double  cosh (double a) { return  ::cosh(a); }
  static inline double  exp  (double a) { return  ::exp(a);  }
  static inline double  fabs (double a) { return  ::fabs(a); }
  static inline double  floor(double a) { return  ::floor(a);}
  static inline double  fmod (double a1, double a2) { return  ::fmod(a1, a2); }
  static inline double  frexp(double a1, int* a2) { return  ::frexp(a1,a2); }
#ifdef _MSC_VER
  static inline double  hypot(double a1, double a2) { return  ::_hypot(a1, a2); }
  static inline double  j0   (double a) { return  ::_j0(a); }
  static inline double  j1   (double a) { return  ::_j1(a); }
  static inline double  jn   (int a1, double a2) { return  ::_jn(a1, a2); }
  static inline double  y0   (double a) { return  ::_y0(a); }
  static inline double  y1   (double a) { return  ::_y1(a); }
  static inline double  yn   (int a1, double a2) { return  ::_yn(a1, a2); }
#else // _MSC_VER
  static inline double  hypot(double a1, double a2) { return  ::hypot(a1, a2); }
  static inline double  j0   (double a) { return  ::j0(a); }
  static inline double  j1   (double a) { return  ::j1(a); }
  static inline double  jn   (int a1, double a2) { return  ::jn(a1, a2); }
  static inline double  y0   (double a) { return  ::y0(a); }
  static inline double  y1   (double a) { return  ::y1(a); }
  static inline double  yn   (int a1, double a2) { return  ::yn(a1, a2); }
#endif // _MSC_VER
  static inline double  ldexp(double a1,int a2) { return  ::ldexp(a1, a2); }
  static inline double  logl (double a) { return  ::log(a); }
  static inline double  log10(double a) { return  ::log10(a); }
  static inline double  modf (double a1, double* a2) { return  ::modf(a1, a2); }
  static inline double  pow  (double a1, double a2) { return  ::pow(a1,a2); }
  static inline double  sin  (double a) { return  ::sin(a); }
  static inline double  sinh (double a) { return  ::sinh(a); }
  static inline double  sqrt (double a) { return  ::sqrt(a); }
  static inline double  tan  (double a) { return  ::tan(a); }
  static inline double  tanh (double a) { return  ::tanh(a); }
  
#ifdef __HAS_LONG_DOUBLE_MATH
  static inline long double  acos (long double a) { return ::acosl(a); }
  static inline long double  asin (long double a) { return ::asinl(a); }
  static inline long double  atan (long double a) { return ::atanl(a); } 
  static inline long double  atan2(long double a1,long double a2) {
    return  ::atan2l(a1,a2); }
  static inline long double  ceil (long double a) { return  ::ceill(a); }
  static inline long double  cos  (long double a) { return  ::cosl(a);  }
  static inline long double  cosh (long double a) { return  ::coshl(a); }
  static inline long double  exp  (long double a) { return  ::expl(a);  }
  static inline long double  fabs (long double a) { return  ::fabsl(a); }
  static inline long double  floor(long double a) { return  ::floorl(a);}
  static inline long double  fmod (long double a1, long double a2) {
    return  ::fmodl(a1, a2); }
  static inline long double  frexp(long double a1, int* a2) {
    return  ::frexpl(a1,a2); }
  // static inline long double  hypot(long double a1, long double a2) {
  // return  ::_hypotl(a1, a2); }
  // static inline long double  j0   (long double a) { return  ::_j0l(a); }
  // static inline long double  j1   (long double a) { return  ::_j1l(a); }
  // static inline long double  jn   (int a1, long double a2) {
  //    return  ::_jnl(a1, a2); }
  static inline long double  ldexp(long double a1,int a2) {
    return  ::ldexpl(a1, a2); }
  static inline long double  logl (long double a) { return  ::logl(a); }
  static inline long double  log10(long double a) { return  ::log10l(a); }
  static inline long double  modf (long double a1, long double* a2) {
    return  ::modfl(a1, a2); }
  static inline long double  pow  (long double a1, long double a2) {
    return  ::powl(a1,a2); }
  static inline long double  sin  (long double a) { return  ::sinl(a); }
  static inline long double  sinh (long double a) { return  ::sinhl(a); }
  static inline long double  sqrt (long double a) { return  ::sqrtl(a); }
  static inline long double  tan  (long double a) { return  ::tanl(a); }
  static inline long double  tanh (long double a) { return  ::tanhl(a); }
  // static inline long double  y0   (long double a) { return  ::_y0l(a); }
  // static inline long double  y1   (long double a) { return  ::_y1l(a); }
  // static inline long double  yn   (int a1, long double a2) {
  //   return  ::_ynl(a1, a2); }
#else
  static inline long double  acos (long double a) { return ::acos(a); }
  static inline long double  asin (long double a) { return ::asin(a); }
  static inline long double  atan (long double a) { return ::atan(a); } 
  static inline long double  atan2(long double a1,long double a2) {
    return  ::atan2(a1,a2); }
  static inline long double  ceil (long double a) { return  ::ceil(a); }
  static inline long double  cos  (long double a) { return  ::cos(a);  }
  static inline long double  cosh (long double a) { return  ::cosh(a); }
  static inline long double  exp  (long double a) { return  ::exp(a);  }
  static inline long double  fabs (long double a) { return  ::fabs(a); }
  static inline long double  floor(long double a) { return  ::floor(a);}
  static inline long double  fmod (long double a1, long double a2) {
    return  ::fmod(a1, a2); }
  static inline long double  frexp(long double a1, int* a2) {
    return  ::frexp(a1,a2); }
  // static inline long double  hypot(long double a1, long double a2) {
  // return  ::_hypotl(a1, a2); }
  // static inline long double  j0   (long double a) { return  ::_j0l(a); }
  // static inline long double  j1   (long double a) { return  ::_j1l(a); }
  // static inline long double  jn   (int a1, long double a2) {
  //    return  ::_jnl(a1, a2); }
  static inline long double  ldexp(long double a1,int a2) {
    return  ::ldexp(a1, a2); }
  static inline long double  logl (long double a) { return  ::log(a); }
  static inline long double  log10(long double a) { return  ::log10(a); }
  static inline long double  pow  (long double a1, long double a2) {
    return  ::pow(a1,a2); }
  static inline long double  sin  (long double a) { return  ::sin(a); }
  static inline long double  sinh (long double a) { return  ::sinh(a); }
  static inline long double  sqrt (long double a) { return  ::sqrt(a); }
  static inline long double  tan  (long double a) { return  ::tan(a); }
  static inline long double  tanh (long double a) { return  ::tanh(a); }
  // static inline long double  y0   (long double a) { return  ::_y0(a); }
  // static inline long double  y1   (long double a) { return  ::_y1(a); }
  // static inline long double  yn   (int a1, long double a2) {
  //   return  ::_yn(a1, a2); }
#endif // __HAS_LONG_DOUBLE_MATH

  /// Returns 'a'-squared.
  template<class T>
  static inline T sqr2(T a) { return (a) * (a); }

  /// Kronecker delta (=1 if n1==n2, 0 otherwise).
  template<class T>
  static inline T delta(T n1, T n2) { return ( (n1==n2) ? 1 : 0); }

  /// Size of the precomputed factorial array.
  static const int FACT_ARRAY_SIZE = 35;

  /// Precomputed factorial array.
  static const double SFactArray[FACT_ARRAY_SIZE];

  /// Return n-factorial as a double (table-based).
  static inline double factorial(int n);

  static double binom(int n,int k) 
  { 
    return factorial(n) / ( factorial(n-k) * factorial(k) );
  }
};

// --------------------------------------------------------------------------
//  shm::factorial
// --------------------------------------------------------------------------
inline double shm::factorial(int n)
{
  assert(n >= 0);

  if (n < FACT_ARRAY_SIZE)
    // take the value from a table
    return SFactArray[n];

  // compute the value traditionally starting from the
  // maximum value in the array
  double fact = SFactArray[FACT_ARRAY_SIZE-1];
  double mul = FACT_ARRAY_SIZE;
  for(int i = FACT_ARRAY_SIZE; i < n; i++, mul += 1.0 )
    fact *= mul;
  return fact;
}

__END_SPHAR_HEADER

#endif // __GMATH_H__
