// ===================================================================
// $Id: shrot.h,v 1.14 2005/11/29 15:51:50 xkrivanj Exp $
/// \file shrot.h
/// \brief Spherical harmonic rotation code.
///
/// SPHAR_REPLACEMENT_STRING
///
/// December 2003    Jaroslav Krivanek     Initial coding in shlib.h
/// May 2004         Jaroslav Krivanek     Code moved from shlib.h to shrot.h.

#ifndef __SHROT_H__
#define __SHROT_H__

// SPHAR headers
#include "shconfig.h"
#include "sheval.h"
#include "shrotmat.h"
#include "shvec.h"

__BEGIN_SPHAR_HEADER

class CSHRotMatrix;


/// Generate C++ code for SH rotation around Z-axis.
void SHRotZGetCode(ostream& o, 
                   const char *out, const char* in, 
                   const char *cosn, const char* sinn,
                   int order);

/// Rotate about Z-axis. Works for SH and HSH.
template<class T>
inline void 
shRotZ(T* out, const T* in, const T* cosn, const T* sinn, int order)
{
	// rotate
	for(int l=0, index=0; l<order; l++, index+=2*l) {
		out[index] = in[index];
		// m = 1..l  (&& -1..-l at the same time)
		for(int m=1; m<=l; ++m) {
			out[index-m] = cosn[m] * in[index-m] - sinn[m] * in[index+m];
			out[index+m] = sinn[m] * in[index-m] + cosn[m] * in[index+m];
		}
	}
}

/// Rotate about Z-axis. Works for SH and HSH.
/**
    @param order (in) number of bands
    @param c  (in) cosine of the rotation angle
	  @param s  (in) sine of the rotation angle
	  @param in (in) input vector
	  @param out (out) rotated vector
  */
template<class TReal>
void shRotZ(int order, const TReal c, const TReal s, const TReal* in, TReal* out)
{
	TReal _cosn[MAX_FIXED_SH_ORDER], _sinn[MAX_FIXED_SH_ORDER];
	TReal *cosn, *sinn;
	if(order<MAX_FIXED_SH_ORDER) { cosn = _cosn;            sinn = _sinn; }
	else               { cosn = new TReal[order]; sinn = new TReal[order]; }
	// evaluate sin(n*phi) and cos(n*phi)
	recurrentSinCos<TReal>(order,c,s,cosn,sinn);
	// rotate
  shRotZ<TReal>(order,cosn,sinn,in,out);

	if(sinn != _sinn) {
		delete [] cosn;
		delete [] sinn;
	}
}

/// ZYZ Euler decomposition of the given 3x3 rotation matrix.
/**
  The rotation matrix is specified by three 3D vectors U,V,N
  (each vector is one row of the matrix).
  The results are stored as sines and cosines of the rotation angles.

  @param U [in] First row of the rotation matrix.
  @param V [in] Second row of the rotation matrix.
  @param N [in] Third row of the rotation matrix.
  @param cz1 [out] cos(alpha)   ... Z-rotation
  @param sz1 [out] sin(alpha)   ... Z-rotation
  @param cy  [out] cos(beta)    ... Y-rotation
  @param sy  [out] sin(beta)    ... Y-rotation
  @param cz2 [out] cos(gamma)   ... Z-rotation
  @param sz2 [out] sin(gamma)   ... Z-rotation

  @return 
  If the z-component of N (i.e matrix element m22) if greater than
  'threshold' the function returns 'true' indicating that the decomposition
  could not be performed.
  Otherwise, 'false' is returned, indicating the decomposition is ok.
 */
template <class T>
inline bool 
EulerAnglesZYZ(const T *U, const T *V, const T *N,
               T &cz1, T &sz1, T &cy, T &sy, T &cz2, T &sz2,
               const T threshold = T(0.9999))
{
  cy = N[2];  // m22
  if( fabs(cy) > threshold ) return true; // error in decomposition

  sy  = sqrtf(1-cy*cy);
  float isy = 1.0f/sy;
  cz1 =  U[2]*isy;        // m02
  sz1 =  V[2]*isy;        // m12
  cz2 = -N[0]*isy;        // m20
  sz2 =  N[1]*isy;        // m21

  return false;
 }

/// Rotate band l=1 of a SH vector.
template <class T>
inline void 
RotateBand1(T *dest, const T *src, const T *U, const T *V, const T *N)
{
  dest[1] =  src[1]*V[1] - src[2]*N[1] + src[3]*U[1];
  dest[2] = -src[1]*V[2] + src[2]*N[2] - src[3]*U[2];
  dest[3] =  src[1]*V[0] - src[2]*N[0] + src[3]*U[0];
}

#define MAX_DERIVATIVE 10

class CSHRotGenerator {

private: // data

	/// pointers to currently built matrices
  CSHRotMatrix *_dR[MAX_DERIVATIVE+1];

protected: // types

	struct SPreRec { double u,v,w; unsigned int nz; };

protected: // data

	/// order for which the generic rotation data are precomputed
	int _order;

	/// precomputed generic rotation data
	SPreRec **_precomputed;

  /// Rotation matrix aroun X by +/-90 degrees. For ZXZXZ decomposition.
  CSHRotMatrix  _Rxp90, _Rxm90;

  /// derivative of the rotation matrix around Y in y=0
  CSHRotMatrix _dyMatrix;

  /// subdiagonal of _dyMatrix
  float *_dySubDiag;

  /// second derivative of the rotation matrix around Y in y=0
  CSHRotMatrix _ddyMatrix;

  /// diagonal of _ddyMatrix
  float *_ddyDiag;

  /// sub-sub diagonal of _ddyMatrix
  float *_ddySSubDiag;

protected: // methods

	static double little_u(int l, int m1, int m2);
	static double little_v(int l, int m1, int m2);
	static double little_w(int l, int m1, int m2);

	double P(int AIndex, int l, int m1, int m2);
	inline double U(int l, int m1, int m2);
        inline double V(int l, int m1, int m2);
	inline double W(int l, int m1, int m2);

	// specilized for derivative of rotation around Y
	double dPY(int AIndex, int l, int m1, int m2, int nth);
	inline double dUY(int l, int m1, int m2, int nth);
	inline double dVY(int l, int m1, int m2, int nth);
	inline double dWY(int l, int m1, int m2, int nth);
  inline double dTerm(int fl, int fm1, int fm2,int gl, int gm1, int gm2, int nth);

	inline const SPreRec& _getGenericValue(int l, int m1, int m2) const 
	{ 
		return _precomputed[shIndex(l,m1)][l+m2];
	}

	void _deletePrecomputed();

	void _setNewOrder(int order);

public:

    CSHRotGenerator(int order);

	~CSHRotGenerator();

  /// Updates the precomputed structures to work up to give order.
  void LiftOrder(int order) {if(order>_order)_setNewOrder(order);}

  /// Return the rotation matrix around X by +90 degrees.
  const CSHRotMatrix& GetRxp90() const {return _Rxp90; }
  /// Return the rotation matrix around X by -90 degrees.
  const CSHRotMatrix& GetRxm90() const {return _Rxm90; }

  /// Rotates a SH coefficient vector around X-axis by 90 degrees 
  /// (for ZXZXZ decomposition).
  void TransformVectorXp90(float *out, const float* in, int order) const
  { _Rxp90.TransformVectorXp90(out,in,order);  }

  /// Rotates a SH coefficient vector around X-axis by -90 degrees 
  /// (for ZXZXZ decomposition).
  void TransformVectorXm90(float *out, const float* in, int order) const
  { _Rxm90.TransformVectorXm90(out,in,order); } 

	/// Builds a SH rotation matrix given a 3x3 rotation matrix.
	void BuildMatrix(CSHRotMatrix &shmat, int order,
                     double r00, double r01, double r02,
					 double r10, double r11, double r12,
					 double r20, double r21, double r22);

	/// Builds a SH rotation matrix given three Euler angles (ZYZ).
	void BuildMatrixZYZ(CSHRotMatrix &shmat, int order,
		                double AZ1, double AY, double AZ2);

	/// Builds a rotation matrix for given angle of rotation around Y.
	void BuildMatrixY(
    CSHRotMatrix &shmat, 
    int           order, 
    double        cos_beta, 
    double        sin_beta);

	/// Builds a rotation matrix for given angle of rotation around a given axis.
  /**
     Angle given by its sine and cosine.

     @param c  (in) cosine of the rotation angle.
     @param s  (in) sine of the rotation angle.
    */
  void BuildMatrixAxisCosSin(
    CSHRotMatrix &shmat, 
    int    order, 
    const  float* axis, 
    double c,
    double s);

	/// Builds a rotation matrix for given angle of rotation around a given axis.
  inline void BuildMatrixAxisAngle(
    CSHRotMatrix &R, 
    int order, 
    const float* axis, 
    double angle)
  {
    BuildMatrixAxisCosSin(R,order,axis,cos(angle),sin(angle));
  }


	/// Builds derivatives of the SH rotation matrix around Y.
  /**
    Derivatives of order 0, 1, ... nth are built. 
    
    @param (in,out) dR   Array of (nth + 1) pointer to SH rotation matrice
                         into which the derivatives will be stored.
   */
  void BuildMatrix_dY(
    CSHRotMatrix **dR, 
    int           nth,
    int           order,
    double        cos_beta, 
    double        sin_beta);

	/// Builds derivative of rotation around Y at 0.
	void BuildMatrix_dY0(CSHRotMatrix &dmat, int order);

	/// Builds the second derivative of rotation around Y at 0 
  void BuildMatrix_ddY0(CSHRotMatrix &ddR, int order);

	/// Builds the second derivative of rotation around Y at 0 
  /// (numerical differentiation)
	void BuildMatrixNum_ddY0(CSHRotMatrix &dmat, int order);

	/// Builds the third derivative of rotation around Y at 0 
  /// (numerical differentiation)
  void BuildMatrixNum_dddY0(CSHRotMatrix &dddR, int order);

  /// Derivative of rotation around given axis at 0.
  void BuildMatrixNum_dAxis0(CSHRotMatrix &dR, int order, const  float* axis);

  /// Second derivative of rotation around given axis at 0.
  void BuildMatrixNum_ddAxis0(CSHRotMatrix &ddR, int order, const  float* axis);

	/// Construct rotation matrix from local frame to global frame.
	inline void ToGlobalFrame(
		CSHRotMatrix &shm,
		int order, 
		const float* U,
		const float* V,
		const float* N)
	{
		BuildMatrix(shm,order,
			U[0], U[1], U[2], 
			V[0], V[1], V[2], 
			N[0], N[1], N[2]);
	}

	/// Construct rotation matrix from global frame to local frame.
	inline void ToLocalFrame(
		CSHRotMatrix &shm,
		int order, 
		const float* U,
		const float* V,
		const float* N)
	{
		BuildMatrix(shm,order,
			U[0], V[0], N[0], 
			U[1], V[1], N[1], 
			U[2], V[2], N[2]);
	}

  /// Rotate around Y using first and partial second order Taylor expansion.
  inline void RotYdiff1(
    int order,
    float* dest, 
    const float* src, 
    float y);

  /// Rotate around Y using full second order Taylor expansion.
  inline void RotYdiff2(
    int order,
    float* dest, 
    const float* src, 
    float y);

  /// Generate unrolled C code for RotYdiff1().
  void RotYdiff1GenCode(
    ostream &o,
    int order,
    const char* dest, 
    const char* src);
};

// --------------------------------------------------------------------------
//  RotYdiff1()
// --------------------------------------------------------------------------
inline void 
CSHRotGenerator::RotYdiff1(int order,
                           float* dest, 
                           const float* src, 
                           float y)
{
  float yy=0.5*y*y;
  dest[0] = src[0];
  int i;
  for(i=1; i<order*order-1; i++) {
    dest[i] = src[i] * (1.0f + yy*_ddyDiag[i]) +
      y * (_dySubDiag[i]*src[i-1] - _dySubDiag[i+1]*src[i+1]);
  }
  dest[i] = src[i] * (1.0f + yy*_ddyDiag[i]) +
    y * _dySubDiag[i] * src[i-1];
}

// --------------------------------------------------------------------------
//  RotYdiff2()
// --------------------------------------------------------------------------
inline void 
CSHRotGenerator::RotYdiff2(int order,
                           float* dest, 
                           const float* src, 
                           float y)
{
  float yy=0.5*y*y;

#if 0
  float tmpvec1[MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];
  float tmpvec2[MAX_FIXED_SH_ORDER*MAX_FIXED_SH_ORDER];

  shvf::movvv(order*order,dest,src);

  _dyMatrix.TransformVector(tmpvec1,src,order);
  shvf::madvsip(order*order,dest,tmpvec1,y);

  _ddyMatrix.TransformVector(tmpvec2,src,order);
  shvf::madvsip(order*order,dest,tmpvec2,yy);
#else
  dest[0] = src[0];
  dest[1] = src[1];
  int i;
  for(i = 2; i < order * order - 2; i++) {
    dest[i] = src[i] * (1.0f + yy * _ddyDiag[i]);
    dest[i] +=  y * (_dySubDiag[i]   * src[i-1] - _dySubDiag[i+1] * src[i+1]);
    dest[i] += yy * (_ddySSubDiag[i] * src[i-2] + _ddySSubDiag[i+2] * src[i+2]);
  }
  dest[i] = src[i] * (1.0f + yy*_ddyDiag[i]);
  dest[i] +=  y * (_dySubDiag[i]  *src[i-1] - _dySubDiag[i+1]*src[i+1]);
  dest[i] += yy * (_ddySSubDiag[i]*src[i-2]);
  i++;
  dest[i] = src[i] * (1.0f + yy*_ddyDiag[i]) + y * _dySubDiag[i] * src[i-1];
#endif
}


/// Precomputed data for a SH rotation (ZXZXZ decomposition)
/**
  Usage:
    Upon creation, maximum SH order must be specified.
    Order can be changed with SetNewOrder().

    For a desired rotation, call ToGlobalFrame() to build the rotation 
    and then TransformVector() for each SH vector to be rotated.
  */
class CZXZXZRotationObject {

protected: // data
  /// The allocated storage fits SH rotations up to this order.
  int _order;
  /// Cosines and sines of the Z, Y, Z rotation angles in the Euler ZYZ 
  /// decomposition.
  float *_cnz1, *_cnz2, *_cny, *_snz1, *_snz2, *_sny;
  /// Temporary vectors for the rotations.
  float *_tmpvec1, *_tmpvec2;
  /// Axes of the local frame wrt global frame
  float _U[3], _V[3], _N[3];

protected: // methods

  void _deleteData();

public:

  CZXZXZRotationObject(int order);

  ~CZXZXZRotationObject();

  void SetNewOrder(int order);

  void LiftOrder(int order) {
    if(order>_order) SetNewOrder(order); 
  }

  /// Build the SH rotation for a given 3x3 rotation matrix.
  /**
     U, V, N are the rows of the rotation matrix.
     They are passed to 'EulerAnglesZYZ()' for the decomposition.
   */
  template<class TA>
  void ToGlobalFrame(
    const TA *U, 
    const TA *V, 
    const TA *N, 
    int order);

  /// Rotate vector.
  void TransformVector(
    float *out, 
    const float *in, 
    int order,
    const CSHRotGenerator& rotGenerator);

  void  TransformVectorUnroll(
    float *out, 
    const float *in, 
    int order,
    const CSHRotGenerator& rotGenerator);

  /*
  void  TransformVectorUnrollSSE(
    float *out, 
    const float *in, 
    int order,
    const CSHRotGenerator& rotGenerator);
  */
};

// --------------------------------------------------------------------------
//  CZXZXZRotationObject::TransformVector()
// --------------------------------------------------------------------------
inline void 
CZXZXZRotationObject::TransformVector(float *out, 
                                      const float *in, 
                                      int order,
                                      const CSHRotGenerator& rotGenerator)
{
  // Z (alpha)
  shRotZ<float>(_tmpvec1,in,_cnz1,_snz1,order);
  // X (-90)
  rotGenerator.TransformVectorXm90(_tmpvec2,_tmpvec1,order);
  // Z (beta)
  shRotZ<float>(_tmpvec1,_tmpvec2,_cny,_sny,order);
  // X (90)
  rotGenerator.TransformVectorXp90(_tmpvec2,_tmpvec1,order);
  // Z(gamma)
  shRotZ<float>(out,_tmpvec2,_cnz2,_snz2,order);
}

/// Globally shared SH rotation matrix generator. It is not thread safe to use it.
extern CSHRotGenerator      rotGenerator;
extern CZXZXZRotationObject zxzxzRotObject;

__END_SPHAR_HEADER

#endif // __SHROT_H__
