// ===================================================================
// $Id: shrotmat.h,v 1.5 2005/11/29 15:51:50 xkrivanj Exp $
/// \file shrotmat.h
/// \brief Spherical harmonic rotation matrix.
///
/// SPHAR_REPLACEMENT_STRING
///
/// December 2003    Jaroslav Krivanek     Initial coding in shlib.h
/// May 2004         Jaroslav Krivanek     Code moved from shlib.h to shrot.h.
/// September 2004   Jaroslav Krivanek     Code moved from shrot.h to shrotmat.h.

#ifndef __SHROTMAT_H__
#define __SHROTMAT_H__

// SPHAR headers
#include "shconfig.h"
#include "sheval.h"

__BEGIN_SPHAR_HEADER

/// Spherical harmonics rotation matrix.
class CSHRotMatrix {

protected: // data

	int _order;

public:

	double **_matrix;

	CSHRotMatrix(int order);

	~CSHRotMatrix();

  /// Resize the matrix to a given order.
  void Resize(int newOrder);

	inline int GetOrder() const { return _order; }

	inline const double& GetElement(int l, int m1, int m2) const  
	{ 
		assert(l<_order); assert(m1<=l && -m1<=l); assert(m2<=l && -m2<=l);
		return _matrix[shIndex(l,m1)][l+m2]; 
	}

	inline void SetElement(int l, int m1, int m2, const double& value) 
	{ 
		assert(l<_order); assert(m1<=l && -m1<=l); assert(m2<=l && -m2<=l);
		_matrix[shIndex(l,m1)][l+m2] = value; 
	}

	inline const double& operator() (int l, int m1, int m2) const 
	{ 
		assert(l<_order); assert(m1<=l && -m1<=l); assert(m2<=l && -m2<=l);
		return _matrix[shIndex(l,m1)][l+m2]; 
	}

	inline double& operator()       (int l, int m1, int m2) 
	{ 
		assert(l<_order); assert(m1<=l && -m1<=l); assert(m2<=l && -m2<=l);
		return _matrix[shIndex(l,m1)][l+m2]; 
	}

	const CSHRotMatrix& mul(const double a, int order=-1);

	const CSHRotMatrix& div(const double a, int order=-1)
	{ return mul(1.0f/a,order); }

  // element-by-element matrix multiplication
  const CSHRotMatrix& mul(const CSHRotMatrix& m, int order=-1);

	const CSHRotMatrix& sub(const CSHRotMatrix& m, int order=-1);

	const CSHRotMatrix& add(const CSHRotMatrix& m, int order=-1);

	const CSHRotMatrix& assign(const CSHRotMatrix& m, int order=-1);

	const CSHRotMatrix& setz(int order=-1);

	const CSHRotMatrix& operator*=(const double a)
	{ return mul(a); }

	const CSHRotMatrix& operator/=(const double a)
	{ return div(a); }

	const CSHRotMatrix& operator-=(const CSHRotMatrix& m)
	{ return sub(m); }

	const CSHRotMatrix& operator+=(const CSHRotMatrix& m)
	{ return add(m); }

	const CSHRotMatrix& operator=(const CSHRotMatrix& m)
	{ return assign(m); }

	friend bool EpsilonEqual(const CSHRotMatrix& m1, const CSHRotMatrix& m2, double eps, int order=-1);

	void Describe(ostream& o, int order=-1) const;

	static void SetIdentity(CSHRotMatrix& m, int order=-1);

	/// Rotate the sh-vector.
	/**
    @param out   (out) Transformed vector.
    @param in    (in)  The vector to transform.
	  @param order (in)  Order of the SH vector. It can be lower than the matrix order.
	                     If order is negative (default) the matrix order is used.                      
	*/
	void TransformVector(float *out, const float* in, int order=-1) const;

  /// Generate C code for transformation by a given matrix.
  void TransformVectorGenCode(ostream& o, const char *out, const char* in, int order) const;

  /// Specialization of 'TransformVector' assuming that the matrix is rotation 
  /// around X by -90 degrees.
  void TransformVectorXp90(float *out, const float* in, int order) const;

  /// Specialization of 'TransformVector' assuming that the matrix is rotation 
  /// around X by -90 degrees.
  void TransformVectorXm90(float *out, const float* in, int order) const;

	/// Approximate Y rotation by its derivative.
	/**
	  This function is quivalent to taking (*this)+Identity matrix and multiplying with 
	  given vector, with the assumption that nonzeros are only on the minor and major 
	  diagonal. This function is hacked to take all this into account to deliver optimal performance.
	  All those assumptions are fullfilled by a derivative of rotation around Y in zero,
	  and you don't really want to use it for anything else.
	  To build the derivative of Y rotation matrix  use CSHRotGenerator::BuildMatrixYdY()
	  or BuildMatrixYdY0().
	  This function gives reasonable approximation of rotation around Y for 
	  angle up to 10 degrees.

    'this' matrix is supposed to be a derivative of Y rotation in zero.

	   @param out   Transformed vector.
	   @param in    Vector to be transformed.
     @param ddyDiag  Diagonal of the second derivative matrix.
	   @param angle Angle of rotation in radians.
	   @param order Order of the SH representation.
	*/
	void TransformVectorYdY(
    float* out, 
    const float *in,
    const float *ddyDiag,
    float angle,
    int order=-1) const;
};

__END_SPHAR_HEADER

#endif // __SHROTMAT_H__
