// ===================================================================
// $Id: shtab.h,v 1.6 2005/11/29 15:51:50 xkrivanj Exp $
/// \file shtab.h
/// \brief Storage for tabulated (hemi)spherical basis.
///
/// SPHAR_REPLACEMENT_STRING
///
/// Initial coding by Jaroslav Krivanek, June 2004.

#ifndef __SHTAB_H__
#define __SHTAB_H__

// SPHAR headers
#include "shconfig.h"
#include "sheval.h"

// standard headers

__BEGIN_SPHAR_HEADER

// forward declarations
template<class TV,class TA> class CSphBasis;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// class CSphTabulated
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

/// Pre-evaluated tabulated (hemi)spherical basis.
/**
  Constructor gets a pointer to a (hemi)spherical basis. The basis
  is pre-evaluated and stored in and internal table.
  The basis can be looked up later for fast basis evaluating.
  Nearest neigbor and bilinear interpolation is supported.
  However, bilinear interpolation is slower than actually 
  evaluating the basis (at least for (hemi) spherical harmonics).

  The mapping from a hemisphere to a unit disc is the parabolic mapping
  of [Heindrich and Seidel 1999] - the same mapping that we use to
  represent BRDFs (see shbrdf.(h|cpp)).
  */
template<class TV,class TA>
class CSphTabulated {

protected:

  /// upper hemisphere
  TV **_upper;
  TV **_upper_dtheta;
  TV **_upper_dphi;

  /// lower hemisphere
  TV **_lower;
  TV **_lower_dtheta;
  TV **_lower_dphi;

  /// all zeros
  TV *_defaultVal;

  /// maximum order
  int _maxOrder;

  /// maximum length of coeff vector (=PAD_FOUR(sqr(maxOrder))
  int _maxVecLen;

  /// resolution of the texture
  int _xRes,_yRes;

protected: // methods
  
  int _numEntries() const { return (_xRes+2)*(_yRes+1); }

public: // methods

  /// Pre-evaluates the given basis and stores it in and internal table.
  /**
    @param maxOrder (in) Evaluate the basis up to the order of 'maxOrder'.
    @param res      (in) Resolution of the texture used to represent the basis.
    @param shBasis  (in) Basis to pre-evaluate.
    */
  CSphTabulated(int maxOrder, int res, const CSphBasis<TV,TA>* shBasis);

  ~CSphTabulated();

  /// Evaluate the basis functions and their derivatives w.r.t theta and phi.
  /**
    The data are copied to the arrays 'dest', 'dest_dtheta' and 'dest_dphi'.
    'dest_dtheta' and 'dest_dphi' can be NULL. In this case the derivatives are 
    not evaluated.

    Note!!! This is actually slower than evaluating the basis function 
    analytically.
   */
  void GetCoeffs4(
    int order, 
    const TA *vect, 
    TV *dest,
    TV *dest_dtheta, 
    TV *dest_dphi) const;

  /// Evaluate the basis functions and their derivative w.r.t. theta and phi
  /// in the direction of 'vect'. 
  /**
     Nearest neigbor interpolation. No data is copied, only pointer to
     the internal structure is returned.
     None of dest, dest_dtheta or dest_dphi is allowed to be NULL.
    */
  inline void GetCoeffsInt4(
    int order, 
    const TA *vect, 
    const TV **dest,
    const TV **dest_dtheta, 
    const TV **dest_dphi) const;

  /// Evaluate the basis functions in the direction of 'vect'. 
  /**
     Nearest neigbor interpolation. No data is copied, only pointer to
     the internal structure is returned.
     'dest' is not allowed to be NULL.
    */
  inline void GetCoeffsInt4(
    int order, 
    const TA *vect, 
    const TV **dest) const;
};


// --------------------------------------------------------------------------
//  CSphTabulated::GetCoeffsInt4()
// --------------------------------------------------------------------------
template<class TV,class TA>
inline void
CSphTabulated<TV,TA>::GetCoeffsInt4(int order, 
                                    const TA *vect, 
                                    const TV **dest,
                                    const TV **dest_dtheta, 
                                    const TV **dest_dphi) const
{
  assert(order>0);
  assert(order<=_maxOrder);
  assert(dest);
  assert(dest_dtheta);
  assert(dest_dphi);

  TA u=TA(0.5)*(TA(1.0)+(vect[0]/(TA(1.0)+fabs(vect[2]))));
  TA v=TA(0.5)*(TA(1.0)+(vect[1]/(TA(1.0)+fabs(vect[2]))));

  assert(u>=0.0);
  assert(u<=1.0);
  assert(v>=0.0);
  assert(v<=1.0);
  // if (u<0.0 || u>1 || v<0 || v>1) { FATAL_EXIT; }
  
  int uindex = (int)(u * _xRes + TA(0.4999));
  int vindex = (int)(v * _yRes + TA(0.4999));
  int index  = uindex*(_yRes+1)+vindex;

  if(vect[2] < 0) {
    *dest = _lower[index];
    *dest_dtheta = _lower_dtheta[index];
    *dest_dphi   = _lower_dphi[index];
  }
  else {
    *dest = _upper[index];
    *dest_dtheta = _upper_dtheta[index];
    *dest_dphi   = _upper_dphi[index];
  }

  return;
}


// --------------------------------------------------------------------------
//  CSphTabulated::GetCoeffsInt4()
// --------------------------------------------------------------------------
template<class TV,class TA>
inline void
CSphTabulated<TV,TA>::GetCoeffsInt4(int order,
                                    const TA *vect,
                                    const TV **dest)const
{
  assert(order>0);
  assert(order<=_maxOrder);
  assert(dest);

  TA u=TA(0.5)*(TA(1.0)+(vect[0]/(TA(1.0)+fabs(vect[2]))));
  TA v=TA(0.5)*(TA(1.0)+(vect[1]/(TA(1.0)+fabs(vect[2]))));

  assert(u>=0.0);
  assert(u<=1.0);
  assert(v>=0.0);
  assert(v<=1.0);
  // if (u<0.0 || u>1 || v<0 || v>1) { FATAL_EXIT; }
  
  int uindex = (int)(u * _xRes + TA(0.4999));
  int vindex = (int)(v * _yRes + TA(0.4999));
  int index  = uindex*(_yRes+1)+vindex;

  if(vect[2] < 0)
    *dest = _lower[index];
  else
    *dest = _upper[index];

  return;
}

__END_SPHAR_HEADER

#endif
