// ================================================================
// $Id: shvec.h,v 1.14 2005/11/29 15:51:50 xkrivanj Exp $
/// \file shvec.h
/// \brief Vector operation on arrays of numbers.
///
/// SPHAR_REPLACEMENT_STRING
///
/// December 2003   Jaroslav Krivanek    Initial coding
/// May 2004        Jaroslav Krivanek    Added SSE intrinsic support

#ifndef __SHVEC_H__
#define __SHVEC_H__

// Use SSE instruction if possible
#define SSE_INTRINSIC

#ifndef __SSE__
#undef SSE_INTRINSIC
#endif // __SSE__

// SPHAR headers
#include "shconfig.h"

// standard headers
#include <cmath>
#ifdef SSE_INTRINSIC
#include <xmmintrin.h>
#endif

__BEGIN_SPHAR_HEADER

// #pragma vector aligned

/// returns a multiply of four greater than or equal to 'h' 
inline int mulFour(int h) 
{
  if(h & 3)  return (h & (int)~3)+4;
  else       return h;
}


#ifdef SSE_INTRINSIC
#define ALIGN16             __declspec(align(16)) 

#ifdef _MSC_VER
#define NEW_ALIGN16(type,n)  ((type*)_aligned_malloc((n)*sizeof(type),16))
#define FREE_ALIGN16(array)  if(array){_aligned_free(array);(array)=0;}
#else
// <VH> --- ????? Not sure ??? _aligned_malloc,free only under MS !
#define NEW_ALIGN16(type,n)  ((type*)malloc((n)*sizeof(type)))
#define FREE_ALIGN16(array)  if (array) { free(array);(array)=0;}
#endif // _MSC_VC

#define PAD_FOUR(h)           mulFour(h)
union extract_m128
{
  __m128 m;
  float f[4];
};
#else
#define NEW_ALIGN16(type,n)  ((type*)malloc((n)*sizeof(type)))
#define ALIGN16              
#define FREE_ALIGN16(array)  if(array){free(array);(array)=0;}
#define PAD_FOUR(h)           (h)
#endif


/// Aritmetic operations on vectors.
/**
Implementation of these operations can be optimized using
processor specific instructions.

PARAMETERS

1) always the vector length. 
2) operation destination, can serve as the first source at the same time
3..n) operation sources 

If the return value is a scalar (e.g. dot product, sum) then it is returned
from the function (there is no destination parameter).

NAMING

xxx(vs|vv)[ip][4]

"xxx"...name of the operation (mul, add, div, ...)
"vs"...first operand is vector, second is scalar
"vv"...both operands are vectors

"ip"...'in place' operation (destination = first source),
no suffix means that destination is the last argument.

"4"...instruction implementation can assume that the length of the vector 
a multiply of 4.
*/

template<class  T>
class shv {

public:
  /// set vector to zero
  static inline void setz( int n, T *dest);
  static inline void setz4( int n, T *dest)
  { setz(n,dest); }

  /// move vector to vector
  static inline void movvv( int n, T* dest, const T* src);
  static inline void movvv4( int n, T* dest, const T* src)
  { movvv(n,dest,src); }

  /// move scalar to vector
  static inline void movvs( int n, T* dest, const T src);
  static inline void movvs4( int n, T* dest, const T src)
  { movvs(n,dest,src); }
  static inline void set( int n, T* dest, const T src)
  { movvs(n,dest,src); }

  /// add vector to another vector
  static inline void addvvip ( int n, T* src1_dest, const T* src2);
  static inline void addvvip4( int n, T* src1_dest, const T* src2)
  { addvvip(n,src1_dest,src2); }

  /// multiply vector by scalar in place
  static inline void mulvsip( int n, T *src_dest, const T c );
  static inline void mulvsip4( int n, T *src_dest, const T c ) 
  { mulvsip(n,src_dest,c); }

  /// multiply vector by scalar
  static inline void mulvs ( int n, T* dest, const T *src, const T c );
  static inline void mulvs4( int n, T* dest, const T *src, const T c )
  { mulvs(n,dest,src,c); }

  /// multiply vector by vector
  static inline void mulvv ( int n, T* dest, const T *src1, const T *src2 );
  static inline void mulvv4( int n, T* dest, const T *src1, const T *src2 )
  { mulvv(n,dest,src1,src2); }

  /// multiply vector by vector in place
  static inline void mulvvip( int n, T* src1_dest, const T *src2 )
  { for(int i=0; i<n; i++) src1_dest[i] *= src2[i]; }

  /// multiply vector by a scalar and add to destination
  static inline void madvsip ( int n, T *src1_dest, const T* src2, const T src3);
  static inline void madvsip4 ( int n, T *src1_dest, const T* src2, const T src3)
  { madvsip (n,src1_dest,src2,src3); }

  /// multiply vector by a vector and add to destination
  static inline void madvvip ( int n, T *src1_dest, const T* src2, const T* src3);
  static inline void madvvip4 ( int n, T *src1_dest, const T* src2, const T* src3)
  { madvvip (n,src1_dest,src2,src3); }

  /// linear interpolation of four vectors
  static inline void lerpvv(
    int n, T *dest,
    const T *src1, const T w1,
    const T *src2, const T w2);

  static inline void lerpvv4(
    int n, T *dest,
    const T *src1, const T w1,
    const T *src2, const T w2)
  { lerpvv(n,dest,src1,w1,src2,w2); }

  /// bilinear interpolation of four vectors
  static inline void bili(
    int n, T *dest,
    const T *src1, const T w1,
    const T *src2, const T w2,
    const T *src3, const T w3,
    const T *src4, const T w4);
  static inline void bili4(
    int n, T *dest,
    const T *src1, const T w1,
    const T *src2, const T w2,
    const T *src3, const T w3,
    const T *src4, const T w4)
  { bili(n,dest,src1,w1,src2,w2,src3,w3,src4,w4); }

  /// dot product of two vectors
  static inline T dot( int n, const T *src1, const T *src2);
  static inline T dot4( int n, const T *src1, const T *src2)
  { return dot(n,src1,src2); }

  /// L2-norm of a vector (sum of squares)
  static inline T L2( int n, const T *src);
  static inline T L24( int n, const T *src)
  { return L2(n,src); }

  /// squared distance between the two vectors
  static inline T sqrdist( int n, const T *src1, const T *src2 );
  static inline T sqrdist4( int n, const T *src1, const T *src2 )
  { sqrdist(n,src1,src2 ); }

  /// sum a vector
  static inline T sum( int n, const T *src1);
  static inline T sum4( int n, const T *src1)
  { sum(n,src1); }

  /// dest can be equal to src
  static inline void rcp( int n, T *dest, const T *src);
  static inline void rcp4( int n, T *dest, const T *src)
  { rcp(n, dest, src); }

};

typedef shv<float>  shvf;
typedef shv<double> shvd;


// --------------------------------------------------------------------------
//  shv::setz()
// --------------------------------------------------------------------------
template<class  T>
inline void shv<T>::setz( int n, T *dest)
{ 
  memset(dest, 0, n*sizeof(T) ); 
}

// --------------------------------------------------------------------------
//  shv::movvv()
// --------------------------------------------------------------------------
template<class  T>
inline void shv<T>::movvv( int n, T* dest, const T* src)
{ 
  memcpy(dest,src,n*sizeof(T)); 
}

// --------------------------------------------------------------------------
//  shv::movvs()
// --------------------------------------------------------------------------
template<class  T>
inline void shv<T>::movvs( int n, T* dest, const T src)
{
  for(int i=0; i<n; i++) dest[i]=src;
}


// --------------------------------------------------------------------------
//  shv::addvvip()
// --------------------------------------------------------------------------
template<class  T>
inline void shv<T>::addvvip( int n, T* src1_dest, const T* src2)
{
  for(int i=0; i<n; i++)
    src1_dest[i] += src2[i];
}

#ifdef SSE_INTRINSIC
template<>
inline void shv<float>::addvvip4( int n, float *src1_dest, const float* src2 ) 
{ 
  // SSE intrinsic version
  assert( (n&3) == 0 );
  n /= 4;

  __m128* p_src1_dest = (__m128*) src1_dest;
  __m128* p_src2      = (__m128*) src2;

  for ( int i = 0; i<n; i++)
    p_src1_dest[i] = _mm_add_ps(p_src1_dest[i], p_src2[i]);  
}
#endif

// --------------------------------------------------------------------------
//  shv::mulvsip()
// --------------------------------------------------------------------------
template<class  T>
inline void shv<T>::mulvsip( int n, T *src_dest, const T c ) 
{ 
  for(int i=0; i<n; i++) 
    src_dest[i] *= c; 
}


#ifdef SSE_INTRINSIC
template<>
inline void shv<float>::mulvsip4( int n, float *src1_dest, const float src2 ) 
{ 
  // SSE intrinsic version
  assert( (n&3) == 0 );
  n /= 4;

  __m128* p_src1_dest = (__m128*) src1_dest;
  __m128  m_src2      = _mm_set_ps1(src2);  // m_src2[0, 1, 2, 3] = src2

  for ( int i = 0; i<n; i++)
    p_src1_dest[i] = _mm_mul_ps(p_src1_dest[i],  m_src2);  
}
#endif

// --------------------------------------------------------------------------
//  shv::mulvs()
// --------------------------------------------------------------------------
template<class  T>
inline void shv<T>::mulvs( int n, T *dest, const T *src, const T c ) 
{ 
  for(int i=0; i<n; i++) 
    dest[i] = src[i] * c; 
}


#ifdef SSE_INTRINSIC
template<>
inline void 
shv<float>::mulvs4(int n, float *dest, const float *src, const float src2 )
{ 
  // SSE intrinsic version
  assert( (n&3) == 0 );
  n /= 4;

  __m128* p_dest  = (__m128*) dest;
  __m128* p_src1  = (__m128*) src;
  __m128  m_src2  = _mm_set_ps1(src2);  // m_src2[0, 1, 2, 3] = src2

  for ( int i = 0; i<n; i++)
    p_dest[i] = _mm_mul_ps(p_src1[i],  m_src2);  
}
#endif

// --------------------------------------------------------------------------
//  shv::mulvv()
// --------------------------------------------------------------------------
template<class  T>
inline void shv<T>::mulvv( int n, T *dest, const T *src1, const T *src2 ) 
{ 
  for(int i=0; i<n; i++) 
    dest[i] = src1[i] * src2[i]; 
}

#ifdef SSE_INTRINSIC
template<>
inline void 
shv<float>::mulvv4(int n, float *dest, const float *src1, const float *src2 )
{ 
  // SSE intrinsic version
  assert( (n&3) == 0 );
  n /= 4;

  __m128* p_dest  = (__m128*) dest;
  const __m128* p_src1  = (const __m128*) src1;
  const __m128* p_src2  = (const __m128*) src2;

  for ( int i = 0; i<n; i++)
    p_dest[i] = _mm_mul_ps(p_src1[i],  p_src2[i]);
}
#endif

// --------------------------------------------------------------------------
//  shv::madvsip()
// --------------------------------------------------------------------------
template<class  T>
inline void shv<T>::madvsip (int      n, 
                             T       *src1_dest,
                             const T *src2, 
                             const T  src3)
{ 
  for(int i=0; i<n; i++) 
    src1_dest[i] += (src2[i] * src3); 
}


#ifdef SSE_INTRINSIC
template<>
inline void shv<float>::madvsip4(int          n, 
                                 float       *src1_dest,
                                 const float *src2, 
                                 const float  src3)
{  
  // SSE intrinsic version
  assert( (n&3) == 0 );
  n /= 4;

  __m128* p_src1_dest = (__m128*) src1_dest;
  __m128* p_src2      = (__m128*) src2;
  __m128  m_src3      = _mm_set_ps1(src3);  // m_src3[0, 1, 2, 3] = src3

  for ( int i = 0; i < n; i++)
    p_src1_dest[i] = _mm_add_ps(p_src1_dest[i],_mm_mul_ps(p_src2[i],m_src3));
}
#endif


// --------------------------------------------------------------------------
//  shv::madvvip()
// --------------------------------------------------------------------------
template<class  T>
inline void shv<T>::madvvip (int      n, 
                             T       *src1_dest,
                             const T *src2, 
                             const T *src3)
{ 
  for(int i=0; i<n; i++) 
    src1_dest[i] += (src2[i] * src3[i]); 
}


#ifdef SSE_INTRINSIC
template<>
inline void shv<float>::madvvip4(int          n, 
                                 float       *src1_dest,
                                 const float *src2, 
                                 const float *src3)
{ 
  // SSE intrinsic version
  assert( (n&3) == 0 );

  n /= 4;

  __m128* p_src1_dest = (__m128*) src1_dest;
  __m128* p_src2      = (__m128*) src2;
  __m128* p_src3      = (__m128*) src3;;

  for ( int i = 0; i < n; i++)
    p_src1_dest[i] = _mm_add_ps(p_src1_dest[i],_mm_mul_ps(p_src2[i],p_src3[i]));
}
#endif


// --------------------------------------------------------------------------
//  shv::bili()
// --------------------------------------------------------------------------
template<class  T>
inline void
shv<T>::bili(
             int n, T *dest,
             const T *src1, const T w1,
             const T *src2, const T w2,
             const T *src3, const T w3,
             const T *src4, const T w4)
{
  for(int i=0; i<n; i++)
    dest[i] = src1[i]*w1 + src2[i]*w2 + src3[i]*w3 + src4[i]*w4;
}


#ifdef SSE_INTRINSIC
template<>
inline void
shv<float>::bili4(
                  int n, float *dest,
                  const float *src1, const float w1,
                  const float *src2, const float w2,
                  const float *src3, const float w3,
                  const float *src4, const float w4)
{
  // SSE intrinsic version
  assert( (n&3) == 0 );
  n /= 4;

  __m128  m1, m2;

  __m128* p_dest = (__m128*) dest;

  __m128* p_src1 = (__m128*) src1;
  __m128* p_src2 = (__m128*) src2;
  __m128* p_src3 = (__m128*) src3;
  __m128* p_src4 = (__m128*) src4;

  __m128  m_w1   = _mm_set_ps1(w1);
  __m128  m_w2   = _mm_set_ps1(w2);
  __m128  m_w3   = _mm_set_ps1(w3);
  __m128  m_w4   = _mm_set_ps1(w4);

  for ( int i = 0; i < n; i++, p_dest++,p_src1++,p_src2++,p_src3++,p_src4++)
  {
    m2 = _mm_mul_ps(*p_src1,m_w1);
    m1 = _mm_mul_ps(*p_src2,m_w2);
    m2 = _mm_add_ps(m2,m1);
    m1 = _mm_mul_ps(*p_src3,m_w3);
    m2 = _mm_add_ps(m2,m1);
    m1 = _mm_mul_ps(*p_src4,m_w4);
    *p_dest = _mm_add_ps(m2,m1);
  }
}
#endif

// --------------------------------------------------------------------------
//  shv::lerpvv()
// --------------------------------------------------------------------------
template<class  T>
inline void
shv<T>::lerpvv(
               int n, T *dest,
               const T *src1, const T w1,
               const T *src2, const T w2)
{
  for(int i=0; i<n; i++)
    dest[i] = src1[i]*w1 + src2[i]*w2;
}


#ifdef SSE_INTRINSIC
template<>
inline void
shv<float>::lerpvv4(
                    int n, float *dest,
                    const float *src1, const float w1,
                    const float *src2, const float w2)
{
  // SSE intrinsic version
  assert( (n&3) == 0 );
  n /= 4;

  __m128  m1, m2;

  __m128* p_dest = (__m128*) dest;

  __m128* p_src1 = (__m128*) src1;
  __m128* p_src2 = (__m128*) src2;

  __m128  m_w1   = _mm_set_ps1(w1);
  __m128  m_w2   = _mm_set_ps1(w2);

  for ( int i = 0; i < n; i++, p_dest++,p_src1++,p_src2++)
  {
    m2 = _mm_mul_ps(*p_src1,m_w1);
    m1 = _mm_mul_ps(*p_src2,m_w2);
    *p_dest = _mm_add_ps(m2,m1);
  }
}
#endif

// --------------------------------------------------------------------------
//  shv::dot()
// --------------------------------------------------------------------------
template<class  T>
inline T shv<T>::dot( int n, const T *src1, const T *src2)
{ 
  T s=0; 
  for(int i=0; i<n; i++) { 
    s += src1[i] * src2[i]; 
  } 
  return s; 
}


#ifdef SSE_INTRINSIC
template<>
inline float shv<float>::dot4( int n, const float *src1, const float *src2)
{ 
  // SSE intrinsic version
  assert( (n&3) == 0 );

  n /= 4;

  __m128  m_sum       = _mm_setzero_ps();
  __m128* p_src1      = (__m128*) src1;
  __m128* p_src2      = (__m128*) src2;

  for ( int i = 0; i < n; i++)
    m_sum = _mm_add_ps(m_sum,_mm_mul_ps(p_src1[i],p_src2[i]));

  extract_m128 a;
  a.m=m_sum;
  return a.f[0]+a.f[1]+a.f[2]+a.f[3];
}
#endif


// --------------------------------------------------------------------------
//  shv::L2()
// --------------------------------------------------------------------------
template<class  T>
inline T shv<T>::L2( int n, const T *src)
{ 
  T s=0; 
  for(int i=0; i<n; i++) { 
    s += src[i] * src[i]; 
  } 
  return s; 
}

// --------------------------------------------------------------------------
//  shv::sqrdist()
// --------------------------------------------------------------------------
template<class  T>
inline T shv<T>::sqrdist( int n, const T *src1, const T *src2 )
{ 
  T s=0; 
  for(int i=0; i<n; i++) {
    T tmp = (src1[i] - src2[i]); tmp = tmp * tmp;
    s += tmp;
  } 
  return s; 
}

// --------------------------------------------------------------------------
//  shv::sum()
// --------------------------------------------------------------------------
template<class  T>
inline T shv<T>::sum( int n, const T *src1)
{ 
  T s=0; 
  for(int i=0; i<n; i++) 
    s += src1[i]; 
  return s;
}


// --------------------------------------------------------------------------
//  shv::rcp()
// --------------------------------------------------------------------------
template<class  T>
inline void shv<T>::rcp(int n, T *dest, const T *src) 
{ 
  for(int i=0; i<n; i++) 
    dest[i] = T(1)/src[i]; 
}


#ifdef SSE_INTRINSIC
template<>
inline void 
shv<float>::rcp4(int n, float *dest, const float *src)
{ 
  // SSE intrinsic version
  assert( (n&3) == 0 );
  n /= 4;

  __m128* p_dest  = (__m128*) dest;
  __m128* p_src   = (__m128*) src;

  for ( int i=0; i<n; i++)
    p_dest[i] = _mm_rcp_ps(p_src[i]);
}
#endif

__END_SPHAR_HEADER

#endif // __GVECT_H__
