// ===================================================================
// $Id: shvector.h,v 1.7 2005/11/29 15:51:50 xkrivanj Exp $
/// \file shvector.h
/// \brief Multichannel vector of SH coefficients.
///
/// SPHAR_REPLACEMENT_STRING
///
/// December 2003    Jaroslav Krivanek    Initial coding (in shlib.h)
/// May 2004         Jaroslav Krivanek    Moved from shlib.h to shvec.h
///                                       Added 16-alignment and 4-padding.

#ifndef __SHVECTOR_H__
#define __SHVECTOR_H__

// SPHAR headers
#include "shconfig.h"
#include "shvec.h"

__BEGIN_SPHAR_HEADER

/// Multichannel vector of SH coefficients. Zero length is supported.
template<class TV>
class CCoefVec
{
protected: // data
  
  /// coefficient array
  TV** _array;
    
  /// (H)SH order
  int _order;
  
  /// physical vector length (padded to multiple of four if SSE is used)
  int _vecLen;
  
  /// number of color channels
  int _channels;
    
public: // methods
  
  CCoefVec(int channels, int order);
    
  ~CCoefVec();
  
  inline int GetOrder() const 
  { return _order; }

  /// return the physical vector length (possibly padded to mul of four)
  inline int GetVecLen() const
  { return _vecLen; }
  
  /// return the number of channels
  inline int GetNumChannels() const 
  { return _channels; }
  
    
  TV* GetChannel(int c)       
  { 
    assert(c>=0); 
    assert(c<_channels);
    return _array[c]; 
  }
  
  const TV* GetChannel(int c) const 
  { 
    assert(c>=0); 
    assert(c<_channels);
    return _array[c]; 
  }
  
  void  SetCoeff(int c, int index, TV val) 
  { 
    assert(c>=0); 
    assert(c<_channels);
    _array[c][index] = val; 
  }
  
  const TV GetCoeff(int c, int index) const 
  { 
    assert(c>=0); 
    assert(c<_channels);
    return _array[c][index]; 
  }

  /// Zero-out the padding entries.
  void ZeroPadding()
  {
    int o2 = _order * _order;
    if(_vecLen <= o2)  return;
    for(int ch=0; ch<_channels; ch++) {
      TV *dch = GetChannel(ch);
      for(int i=o2; i<_vecLen; i++)
        dch[i]=0;
    }
  }
  
  /// multiply vector 'v' by the scalar 's' and add to this
  inline CCoefVec<TV>& madvsip(const CCoefVec<TV>&v, TV s);
  
  /// set data from 'v'
  inline void SetData(const CCoefVec<TV>&v);
  
  /// set the padded data to zero
  inline void ZeroPadded();
  
  // set all data to zeros
  inline void setz();
  
  // multiply the whole vector by a scalar
  inline void mul(TV s);

protected:

  inline void _setOrder(int order)
  {
    assert(order>=0);
    _order  = order;
    _vecLen = PAD_FOUR(order*order);
  }

  // friends
  friend class CRCache;
  friend class CAdaptiveCaching;
  template<class _TV,class _TA> friend class CSphBRDF;
};


// --------------------------------------------------------------------------
//  CCoefVec::CCoefVec
// --------------------------------------------------------------------------
template<class TV>
CCoefVec<TV>::CCoefVec(int channels, int order) 
{
  assert(channels>=0);
  assert(order>=0);
  
  if(order<0) order=0;
  if(channels<0) channels=0;
  
  _channels = channels;
  _order    = order;
  _vecLen   = PAD_FOUR( order * order );

  if (_channels>0) {
    _array = NEW_ALIGN16(TV*, _channels);
    assert(_array);
    if(_order>0) {
      for(int i=0;i<_channels;i++) {
        _array[i] = NEW_ALIGN16(TV,_vecLen);
        assert(_array[i]);
      }
    }
    else {
      shv<TV*>::setz(_channels,_array);
    }
  }
  else
    _array = 0;

  setz();
}


// --------------------------------------------------------------------------
//  CCoefVec::~CCoefVec()
// --------------------------------------------------------------------------
template<class TV>
CCoefVec<TV>::~CCoefVec() 
{ 
  if(_array) {
    for(int i=0;i<_channels;i++)
      FREE_ALIGN16(_array[i]);
    FREE_ALIGN16(_array);
  }
}

// --------------------------------------------------------------------------
//  CCoefVec::operator+=()
// --------------------------------------------------------------------------
/*
template<class TV>
inline CCoefVec<TV>& 
CCoefVec<TV>::operator+= (const CCoefVec<TV> &v)
{
  int channels = Min( _channels, v._channels );
  int n        = Min( _sqrOrder, v._sqrOrder );
  for(int i=0; i<channels; i++)
  shv<TV>::addvvip4 ( n, GetChannel(i), v.GetChannel(i) );
  return *this;
}
*/

// --------------------------------------------------------------------------
//  CCoefVec::madvsip()
// --------------------------------------------------------------------------
template<class TV>
inline CCoefVec<TV>& 
CCoefVec<TV>::madvsip(const CCoefVec<TV>&v, TV s)
{
  int channels = MIN( _channels, v._channels );
  int n        = MIN( GetVecLen(), v.GetVecLen() );

  for (int ch=0; ch<channels; ch++)
    shv<TV>::madvsip4 ( n, GetChannel(ch), v.GetChannel(ch), s );

  return *this;
}

// --------------------------------------------------------------------------
//  CCoefVec::SetData()
// --------------------------------------------------------------------------
template<class TV>
inline void
CCoefVec<TV>::SetData(const CCoefVec<TV>&v)
{
  int channels = MIN( _channels, v._channels );
  int n        = MIN( GetVecLen(), v.GetVecLen() );

  for (int ch=0; ch<channels; ch++)
    shv<TV>::movvv(n,GetChannel(ch),v.GetChannel(ch));
}

// --------------------------------------------------------------------------
//  CCoefVec::ZeroPadded()
// --------------------------------------------------------------------------
template<class TV>
inline void
CCoefVec<TV>::ZeroPadded()
{
  int vecLen   = GetVecLen();
  // int sqrOrder = shm::sqr2(_order);
  int sqrOrder = _order * _order;
  int len      = vecLen - sqrOrder;

  if (len > 0) {
    for(int ch = 0; ch < _channels; ch++)
      shv<TV>::setz(len,GetChannel(ch) + sqrOrder); 
  }
}

// --------------------------------------------------------------------------
//  CCoefVec::setz()
// --------------------------------------------------------------------------
template<class TV>
inline void 
CCoefVec<TV>::setz()
{ 
  for(int i=0;i<_channels;i++) 
    shv<TV>::setz4( GetVecLen(), GetChannel(i) ); 
}

// --------------------------------------------------------------------------
//  CCoefVec::mul()
// --------------------------------------------------------------------------
template<class TV>
inline void 
CCoefVec<TV>::mul(TV s) 
{ 
  for(int i = 0; i < _channels; i++) 
    shv<TV>::mulvsip4( GetVecLen(), GetChannel(i), s);
}

__END_SPHAR_HEADER

#endif // __SHVEC_H__

