// ===================================================================
// $Id: shbasis.cpp,v 1.7 2005/11/29 15:51:50 xkrivanj Exp $
/// \file shbasis.cpp
///
/// \brief Abstraction of (hemi)spherical basis and various implementations
/// of spherical and hemispherical harmonics.
///
/// SPHAR_REPLACEMENT_STRING
///
/// Initial coding by Jaroslav Krivanek, June 2004.

// in case the source is included somewhere
#ifndef __SHBASIS_CPP__
#define __SHBASIS_CPP__

// SPHAR headers
#include "shconfig.h"
#include "shbasis.h"

__BEGIN_SPHAR_SOURCE

/////////////////////////////////////////////////////////////////////////
//
//  class CSHBasisLTab
//
//////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  shEvalHelperTabbedV()
// --------------------------------------------------------------------------
// copy&paste&modify of the function from sheval.c
template<class TV,class TA>
void shEvalHelperTabbedV(const int order,
                         const TA x,
                         const TA c,
                         const TA s, 
                         const TA mul,
                         const CTabLegendre<TV,TA>& leg,
                         TV* ret)
{
  TV _cosn[MAX_FIXED_SH_ORDER];  // cos(n*phi) array
  TV _sinn[MAX_FIXED_SH_ORDER];  // sin(n*phi) array
  TV *cosn, *sinn;
  if(order<MAX_FIXED_SH_ORDER) {
    cosn = _cosn;
    sinn = _sinn;
  }
  else {
    FATAL << "Normalisation not implemented yet for SH of order larger than"
      "MAX_FIXED_SH_ORDER." << std::endl;
    FATAL_EXIT;
    cosn = new TV[order];
    sinn = new TV[order];
  }
  
  // evaluate sin(n*phi) and cos(n*phi)
  recurrentSinCos<TV>(order,c,s,cosn,sinn);
  for(int i=0; i<order; i++) { cosn[i]*=mul; sinn[i]*=mul; }
  
  // evaluate associated legendre polynomial P_l^m(x)
  // assLegendreEvalV(order,x,ret);
  const TV *P;
  leg.EvalNNV(order,x,&P,0);
  
  const TV* normTable;
  shNormTab.GetTable(&normTable);
  
  for(int l=0, index=0; l<order; l++, index+=2*l) {
    ret[index] = P[index] * normTable[index] * mul;
    // m = 1..l  (&& -1..-l at the same time)
    for(int m=1; m<=l; ++m) {
      float f = P[index+m] * normTable[index+m];
      ret[index-m]  = f * sinn[m];
      ret[index+m]  = f * cosn[m];
    }
  }
  
  if(sinn != _sinn) {
    delete [] cosn;
    delete [] sinn;
  }
}

// --------------------------------------------------------------------------
//  dSHEvalHelperTabbedV()
// --------------------------------------------------------------------------
// copy&paste&modify of the function from sheval.c
template<class TV,class TA>
void dSHEvalHelperTabbedV(const int order, 
                          const TA x,
                          TA sin_theta,
                          const TA c, 
                          const TA s,
                          const TA mul,
                          const CTabLegendre<TV,TA>& leg, 
                          TV* y,
                          TV* dytheta=0,
                          TV *dyphi=0)
{
	assert(y);

	TV _cosn[MAX_FIXED_SH_ORDER];  // cos(n*phi) array
	TV _sinn[MAX_FIXED_SH_ORDER];  // sin(n*phi) array
	TV *cosn, *sinn;
	if(order<MAX_FIXED_SH_ORDER) {
		cosn = _cosn;
		sinn = _sinn;
	}
	else {
		FATAL << "Normalisation not implemented yet for SH of order larger than MAX_FIXED_SH_ORDER." << std::endl;
		FATAL_EXIT;
		cosn = new TV[order];
		sinn = new TV[order];
	}

	// evaluate sin(n*phi) and cos(n*phi)
	recurrentSinCos<TV>(order,c,s,cosn,sinn);
	for(int i=0; i<order; i++) { cosn[i]*=mul; sinn[i]*=mul; }

	// get the normalization values
	const TV* normTable;
	shNormTab.GetTable(&normTable);

	// 1) evaluate associated legendre polynomial P_l^m(x)
  const TV *P, *_dP;
  leg.EvalNNV(order,x,&P,&_dP);

	if(dytheta) {
		assert(dyphi);
		/// @@@ yes, this is a dirty hack : dHSH/dtheta = sqrt(2) * 2 * dSH/dtheta
		/// sqrt(2) ... normalization; 2 ... because of polynomial shifting; mul=sqrt(2) for HSH
    sin_theta *= shm::sqr2(mul);
	
    // 2) evaluate legendre derivative
#if 0
    // from some reason, tabbed legendre derivative does not work with 
    // HSH, but i haven't succeded in figuring out why
    dLegV(order,x,dyphi,P);
    const TV *dP = dyphi;
#else
    const TV *dP = _dP;
#endif

		// 3) evaluate sh derivative wrt theta
		for(int l=0, index=0; l<order; l++, index+=2*l) {
			dytheta[index] = -sin_theta * normTable[index] * dP[index] * mul;
			for(int m=1; m<=l; ++m) {
				dytheta[index+m] = 
          -sin_theta * normTable[index+m] * dP[index+m] * cosn[m];
				dytheta[index-m] = 
          -sin_theta * normTable[index+m] * dP[index+m] * sinn[m];
			}
		}
	}

	if(dyphi) {
		for(int l=0, index=0; l<order; l++, index+=2*l) {
			y[index] = P[index] * normTable[index] * mul;
			dyphi[index] = 0;
			// m = 1..l  (&& -1..-l at the same time)
			for(int m=1; m<=l; ++m) {
				y[index-m]  = P[index+m] * sinn[m] * normTable[index+m];
				y[index+m]  = P[index+m] * cosn[m] * normTable[index+m];
				dyphi[index+m] = -m * y[index-m];
				dyphi[index-m] =  m * y[index+m];
			}
		}
	}
	else {
		for(int l=0, index=0; l<order; l++, index+=2*l) {
			y[index] = P[index] * normTable[index] * mul;
			// m = 1..l  (&& -1..-l at the same time)
			for(int m=1; m<=l; ++m) {
				y[index-m]  = P[index+m] * sinn[m] * normTable[index+m];
				y[index+m]  = P[index+m] * cosn[m] * normTable[index+m];
			}
		}		
	}

	if(sinn != _sinn) {
		delete [] cosn;
		delete [] sinn;
	}

}

/////////////////////////////////////////////////////////////////////////////
//
//  CSHBasisLTab
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CSHBasisLTab::EvalBasisSphV()
// --------------------------------------------------------------------------
template<class TV,class TA>
void  
CSHBasisLTab<TV,TA>::EvalBasisSphV(int order,
                                   const TA theta, 
                                   const TA phi,
                                   TV* ret) const
{ 
  TA  x = shm::cos(theta);
  if (x >  1) x = 1;
  if (x < -1) x = -1;
  TA c = shm::cos(phi);
  TA s = shm::sin(phi);
  
  shEvalHelperTabbedV<TV,TA>(order,x,c,s,1.0,leg,ret);
}

// --------------------------------------------------------------------------
//  CSHBasisLTab::EvalBasisCartV()
// --------------------------------------------------------------------------
template<class TV,class TA>
void  
CSHBasisLTab<TV,TA>::EvalBasisCartV(const int order, const TA* v3d, 
                             TV* ret) const
{ 
  TA c,s;        // cos(phi), sin(phi)
  TA x = v3d[2]; // cos(theta)
  if      (x >=  1) { x =  1; c = 1; s = 0; }
  else if (x <= -1) { x = -1; c = 1; s = 0; }
  else {
    TA ist = 1.0/sqrt(1.0-x*x); // 1/sin(theta) != inf
    c = v3d[0] * ist;
    s = v3d[1] * ist;
  }
  shEvalHelperTabbedV<TV,TA>(order,x,c,s,1.0,leg,ret);
}

// --------------------------------------------------------------------------
//  CSHBasisLTab::EvalDerivCartV()
// --------------------------------------------------------------------------
template<class TV,class TA>
void  
CSHBasisLTab<TV,TA>::EvalDerivCartV(const int order, const TA* v3d, TV* y, 
                             TV* dytheta, TV* dyphi) const
{
  TV c,s;        // cos(phi), sin(phi)
  TV sin_theta;
  TV x = v3d[2]; // cos(theta)
  const TV limit=0.9999f;
  if      (x >=  limit) { x =  limit; sin_theta = 0; c = 1; s = 0; }
  else if (x <= -limit) { x = -limit; sin_theta = 0; c = 1; s = 0; }
  else {
    sin_theta = shm::sqrt(1.0-x*x);
    TV ist = 1.0/sin_theta;
    c = v3d[0] * ist;
    s = v3d[1] * ist;
  }
  dSHEvalHelperTabbedV<TV,TA>(order,x,sin_theta,c,s,1.0,leg,y,dytheta,dyphi);
}

// --------------------------------------------------------------------------
//  CSHBasisLTab::EvalDerivSphV()
// --------------------------------------------------------------------------
template<class TV,class TA>
void  
CSHBasisLTab<TV,TA>::EvalDerivSphV(const int order, const TA theta, 
                            const TA phi, TV* y, TV* dytheta, 
                            TV* dyphi) const
{
  TA   sin_theta = shm::sin(theta);
  TA   x         = shm::cos(theta);
  const TA limit=0.9999f;
  if      (x >= limit)  { x = limit;  }
  else if (x <= -limit) { x = -limit; }
  TA c = shm::cos(phi);
  TA s = shm::sin(phi);
  
  dSHEvalHelperTabbedV<TV,TA>(order,x,sin_theta,c,s,1.0,leg,y,dytheta,dyphi);
}

/////////////////////////////////////////////////////////////////////////////
//
//  CHSHBasisLTab
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CHSHBasisLTab::EvalBasisSphV()
// --------------------------------------------------------------------------
template<class TV,class TA>
void  
CHSHBasisLTab<TV,TA>::EvalBasisSphV(int order, const TA theta, const TA phi, 
                             TV* ret) const
{
  // Polynomial shifting
  TA x = shm::cos(theta) * 2 - 1;
  if (x < -1) x = -1;
  if (x > 1)  x = 1;
  TA c = shm::cos(phi);
  TA s = shm::sin(phi);
  
  shEvalHelperTabbedV<TV,TA>(order,x,c,s,TA(M_SQRT2),leg,ret);
}

// --------------------------------------------------------------------------
//  CHSHBasisLTab::EvalBasisCartV()
// --------------------------------------------------------------------------
template<class TV,class TA>
void
CHSHBasisLTab<TV,TA>::EvalBasisCartV(const int order, const TA* v3d, 
                                     TV* ret) const
{
  TA c,s;        // cos(phi), sin(phi)
  TA x = v3d[2] * 2 - 1;
  if (x >  1) { x =  1; c = 1; s = 0; }
  else {
    if (x < -1) x = -1;
    TA ist = 1.0/shm::sqrt(1.0 - shm::sqr2(v3d[2]));// 1/sin(theta) != inf
    c = v3d[0] * ist;
    s = v3d[1] * ist;
  }
  
  shEvalHelperTabbedV<TV,TA>(order,x,c,s,TA(M_SQRT2),leg,ret);
}

// --------------------------------------------------------------------------
//  CHSHBasisLTab::EvalDerivCartV()
// --------------------------------------------------------------------------
template<class TV,class TA>
void  
CHSHBasisLTab<TV,TA>::EvalDerivCartV(const int order, const TA* v3d, TV* y, 
                                     TV* dytheta, TV* dyphi) const
{
  TA c,s;        // cos(phi), sin(phi)
  TA sin_theta;
  TA x = v3d[2] * 2 - 1;
  const TA limit=0.9999f;
  if (x >  limit) { x = limit; c = 1; s = 0; sin_theta=0; }
  else {
    if (x < -limit) x = -limit;
    sin_theta = shm::sqrt(TA(1)-shm::sqr2(v3d[2]));
    if(sin_theta<0.0001) { c = 1; s = 0; }
    else {
      TA ist = 1.0f/sin_theta;
      c = v3d[0] * ist;
      s = v3d[1] * ist;
    }
  }
  // dSHEvalHelperV(order,x,sin_theta,c,s,float(M_SQRT2),y,dytheta,dyphi);
  dSHEvalHelperTabbedV<TV,TA>(order,x,sin_theta,c,s,TA(M_SQRT2),leg,
    y,dytheta,dyphi);
}

// --------------------------------------------------------------------------
//  CHSHBasisLTab::EvalDerivSphV()
// --------------------------------------------------------------------------
template<class TV,class TA>
void
CHSHBasisLTab<TV,TA>::EvalDerivSphV(const int order, const TA theta, 
                             const TA phi, TV* y, TV* dytheta, 
                             TV* dyphi) const
{
  TA   sin_theta = shm::sin(theta);
  TA   x = shm::cos(theta) * 2 - 1;
  const TA limit=0.9999f;
  if      (x >  limit) { x =  limit;}
  else if (x < -limit) { x = -limit;}
  float c = shm::cos(phi);
  float s = shm::sin(phi);
  dSHEvalHelperTabbedV<TV,TA>(order,x,sin_theta,c,s,TA(M_SQRT2),leg,y,dytheta,
    dyphi);
}


// explicit instatiation
template class CSHBasis<float,float>;
template class CSHBasis<float,double>;
template class CSHBasis<double,float>;
template class CSHBasis<double,double>;

template class CHSHBasis<float,float>;
template class CHSHBasis<float,double>;
template class CHSHBasis<double,float>;
template class CHSHBasis<double,double>;

template class CSHBasisLTab<float,float>;
template class CSHBasisLTab<float,double>;
template class CSHBasisLTab<double,float>;
template class CSHBasisLTab<double,double>;

template class CHSHBasisLTab<float,float>;
template class CHSHBasisLTab<float,double>;
template class CHSHBasisLTab<double,float>;
template class CHSHBasisLTab<double,double>;

template class CSHBasisTab<float,float>;
template class CSHBasisTab<float,double>;
template class CSHBasisTab<double,float>;
template class CSHBasisTab<double,double>;

template class CHSHBasisTab<float,float>;
template class CHSHBasisTab<float,double>;
template class CHSHBasisTab<double,float>;
template class CHSHBasisTab<double,double>;

__END_SPHAR_SOURCE

#endif // __SHBASIS_CPP__
