// ===================================================================
// $Id: shbrdf.cpp,v 1.6 2005/11/29 15:51:50 xkrivanj Exp $
//
// shbrdf.h
//    (Hemi)spherical harmonics representation of BRDFs.
//
// Initial coding by Jaroslav Krivanek, December 2003.
// Based on Pascal Gautron's code.

// in case the source is included somewhere
#ifndef __SHBRDF_CPP__
#define __SHBRDF_CPP__

// #define DOTS

// GOLEM headers
#ifdef DOTS
#include "dotcount.h"
#endif

// SPHAR headers
#include "shconfig.h"
#include "shmath.h"
#include "shvec.h"
#include "shbrdf.h"

#include "brdf.h"

__BEGIN_SPHAR_SOURCE

/////////////////////////////////////////////////////////////////////////////
//
//  class CBRDFRec
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CBRDFRec::CBRDFRec()
// --------------------------------------------------------------------------
template<class TV>
CBRDFRec<TV>::CBRDFRec(int channels, int order, TV e, const TV* brdfe)
: CCoefVec<TV>(channels, order), error(e)
{
  // allocate the BRDF energy array (one float per channel)
  if(channels>0) {
    brdfEnergy=new TV[channels];
    assert(brdfEnergy);
    SetBrdfEnergy(brdfe);
  }
  else
    brdfEnergy=0;
}

// --------------------------------------------------------------------------
//  CBRDFRec::~CBRDFRec()
// --------------------------------------------------------------------------
template<class TV>
CBRDFRec<TV>::~CBRDFRec() 
{
  delete [] brdfEnergy; 
}

// --------------------------------------------------------------------------
//  CBRDFRec::ComputeRepError()
// --------------------------------------------------------------------------
template<class TV>
TV
CBRDFRec<TV>::ComputeRepError() const
{
  TV maxErr = 0;
  for(int ch=0; ch< GetNumChannels(); ch++) {
    TV l2 = shv<TV>::L24(GetVecLen(), GetChannel(ch));
    TV e  = (brdfEnergy[ch] - l2) / (brdfEnergy[ch] + 1e-3);
    if(e > maxErr)
      maxErr = e;
  }
  return maxErr;
}

// --------------------------------------------------------------------------
//  CBRDFRec::AdaptOrder()
// --------------------------------------------------------------------------
template<class TV>
CBRDFRec<TV>*
CBRDFRec<TV>::AdaptOrder(int newMaxAllowedOrder, const TV maxAllowedError) const
{
  // ------------------------------------
  //  1. Find the minimum required order
  // ------------------------------------
  TV    aError=0;
  int   aOrder=0;
  int   l;

  int maxOrder = MIN(GetOrder(), newMaxAllowedOrder);

  // find the minimum length for each channel separately
  for(int ch=0; ch<_channels; ch++) {
    if( brdfEnergy[ch] < 0.00001 )
      continue;
    // error of this channel
    TV e = 0, l2 = 0;
    const TV* thisChannel = GetChannel(ch);

    for( l=0; l < newMaxAllowedOrder; l++) {
      // update L2 norm of the sh-expansion
      for(int m = -l; m<=l; m++)
        l2 += shm::sqr2( thisChannel[shIndex(l,m)] );
      // update error estimate
      e = fabsf(brdfEnergy[ch] - l2) / (brdfEnergy[ch] + 1e-3);
      // is the error criterion is fullfilled?
      if(e <= maxAllowedError) {
        l++; break;
      }
    }
    // is the required order for this channels greater than for previous?
    if( l > aOrder) {
      aOrder = l;
      aError = e;
    }
  } // for ( channels )

  // ------------------------------------
  //  2. Alloc the new record
  // ------------------------------------
  CBRDFRec* newRec = new CBRDFRec(GetNumChannels(), aOrder, aError, brdfEnergy);
  newRec->SetData(*this);
  newRec->ZeroPadded();

  return newRec;
}

/////////////////////////////////////////////////////////////////////////////
//
//  Helper classes for CSphBRDF
//
/////////////////////////////////////////////////////////////////////////////

template<class T>
class cvect3d {
public: 
  T x,y,z; 
  cvect3d(){}
  cvect3d(const T xx,const T yy,const T zz):x(xx),y(yy),z(zz){}
  operator const T*() const { return (const T*) this; }
  cvect3d operator=(const cvect3d& v) { x=v.x; y=v.y; z=v.z; return *this;}
};
template<class T>
inline cvect3d<T> Normalize(const cvect3d<T>& v) {
  T is = T(1.0)/sqrtf(v.x*v.x + v.y*v.y + v.z*v.z);
  return cvect3d<T>(v.x*is,v.y*is,v.z*is);
}

/////////////////////////////////////////////////////////////////////////////
//
//  class CSphBRDFStatistics
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CSphBRDFStatistics::SetUp()
// --------------------------------------------------------------------------
template<class TV>
void 
CSphBRDFStatistics<TV>::SetUp(int maxAllowedOrder) 
{
  numDirections = 0;
  numErrorDirections = 0;
  maxOrder = 0;
  minOrder = maxAllowedOrder;
  avgOrder = 0;
  maxError = 0;
  minError = 1;
  avgError = 0;
  delete [] orderHistogram; // just in case
  orderHistogram = new int[maxAllowedOrder+1];
  assert(orderHistogram);
  shv<int>::setz( maxAllowedOrder+1, orderHistogram );
}

// --------------------------------------------------------------------------
//  CSphBRDFStatistics::Describe()
// --------------------------------------------------------------------------
template<class TV>
void 
CSphBRDFStatistics<TV>::Describe(ostream &o) const
{
  o << "CSphBRDFStatistics{\n";
  o << "numDirections        = " << numDirections << '\n';
  o << "numErrorDirections   = " << numErrorDirections << '\n';
  o << "maxOrder             = " << maxOrder << '\n';
  o << "minOrder             = " << minOrder << '\n';
  o << "avgOrder             = " << avgOrder << '\n';
  o << "order99Quantile      = " << order99Quantile << '\n';
  o << "maxError             = " << maxError*100 << "%\n";
  o << "minError             = " << minError*100 << "%\n";
  o << "avgError             = " << avgError*100 << "%\n";
  o << "orderHistogram       = {" << orderHistogram[0];
  for(int i=1; i<=maxOrder; i++)
    o  << /*","*/ " " << orderHistogram[i];
  o << "}\n";
  o << "}\n";
}

// --------------------------------------------------------------------------
//  CSphBRDFStatistics::OrderQuantile()
// --------------------------------------------------------------------------
template<class TV>
int 
CSphBRDFStatistics<TV>::OrderQuantile(TV prob)
{
  if(prob<0) prob = 0;
  if(prob>1) prob = 1;
  
  int limit = (int)(((TV)numDirections)*prob);
  
  int s=0;
  
  for(int i=0; i<=maxOrder; i++) {
    s += orderHistogram[i];
    if( s>= limit)
      return i;
  }
 
  FATAL 
    << "Unreachable code reached in CSphBRDFStatistics<TV>::OrderQuantile()."
    << std::endl;
  FATAL_EXIT; // this should never happen
}

/////////////////////////////////////////////////////////////////////////////
//
//  class CSphBRDF
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CSphBRDF::BrdfToSH()
// --------------------------------------------------------------------------
template<class TV,class TA>
void 
CSphBRDF<TV,TA>::BrdfToSH(const CBRDF<TV,TA> &brdf, 
                   const TA *outDir, 
                   const CSphBasisEvalCache<TV,TA>& cache,
                   int   order,
                   TV    *coeffs,
                   TV    &brdfEnergy)
{
  // coefficient vector length
  int vecLen = order*order;
  // accumulated BRDF energy
  TV  brdfEn = 0;
  // set coefficient to zero
  shv<TV>::setz(vecLen, coeffs);
  // multiplier for result normalization
  TV     norm            = cache.GetNormalisation();
  // number of directions for Monte Carlo sampling
  int    numCacheSamples = cache.GetNumSamples();

  // for all directions
  for(int in_index=0; in_index < numCacheSamples; in_index++) {            
    const TV *H = cache[in_index].coeffs;
    const TA *v = cache[in_index].v;
    // evaluate the brdf
    TV brdfval = brdf.Eval(v, outDir);
    // update BRDF energy
    brdfEn += brdfval*brdfval;
    // update the coeff vector
    shv<TV>::madvsip(vecLen, coeffs, H, brdfval);
  }
  // normalize result
  shv<TV>::mulvsip(vecLen, coeffs, norm);
  brdfEnergy = brdfEn * norm;
  // get energy in the SH representation
  TV shEnergy = shv<TV>::L2( vecLen, coeffs );
}

// --------------------------------------------------------------------------
//  CSphBRDF::CSphBRDF()
// --------------------------------------------------------------------------
template<class TV,class TA>
CSphBRDF<TV,TA>*
CSphBRDF<TV,TA>::Adapt(int   newMaxAllowedOrder,
                       TV    newMaxError, 
                       bool  newAdaptiveOrder,
                       bool  newAdaptiveDeletion)
{
  CSphBRDF *adapted =
    new CSphBRDF( _channels, newMaxAllowedOrder, _xRes, _yRes, 
		  newMaxError, newAdaptiveOrder, newAdaptiveDeletion);

  int   yRes1  = _yRes + 1;
  TA    deltax = 2.0/_xRes;
  TA    deltay = 2.0/_yRes;
  int x, y;
  cvect3d<TA> vect;
  for(x=0; x <= _xRes; x++) {
    vect.x = x*deltax-1;
    for(y=0; y <= _yRes; y++) {
      vect.y =  y*deltay-1;			
      TA sqrDist = shm::sqr2(vect.x) + shm::sqr2(vect.y);
      if (sqrDist > TA(1.0) )
        continue;

      const CBRDFRec<TV> *oldRec        = this->_coeffs[x*yRes1+y];
      TV                 oldError      = oldRec->GetError();
      int                oldOrder      = oldRec->GetOrder();
      const TV           *oldBrdfEnergy = oldRec->GetBrdfEnergy();

      if( !newAdaptiveOrder ) {
        // the easier part -> no adaptive representation
        if( newMaxAllowedOrder > oldOrder && oldError!=0 ) {
          goto impossible_to_adapt; // higher order is required than is available
        }

        // create new record for this direction
        CBRDFRec<TV>* newRec = 
          new CBRDFRec<TV>(_channels,newMaxAllowedOrder,oldError,oldBrdfEnergy);
        assert(newRec);
        // set the record to the texture
        adapted->_coeffs[x*yRes1+y] = newRec;
        // copy coefficients from the old vector
        newRec->SetData(*oldRec);
        // error must be updated because some coeffs might have been chopped off
        newRec->UpdateRepError();
      }
      else {
        // the more tricky part -> adaptive representation
        CBRDFRec<TV>* newRec = oldRec->AdaptOrder(newMaxAllowedOrder,newMaxError);
        assert(newRec);
        // set the record to the texture
        adapted->_coeffs[x*yRes1+y] = newRec;
        // could we possibly do better by increasing the order?
        if( newRec->GetError() > newMaxError && newMaxAllowedOrder > oldOrder) {
          goto impossible_to_adapt;
        }
      }
    }
  }

  // build the stats
  adapted->_fillStatistics();
  adapted->_stats.Describe(DEBUGW);

  return adapted;

impossible_to_adapt:
  delete adapted;
  return 0;
}

// --------------------------------------------------------------------------
//  CSphBRDF::CSphBRDF()
// --------------------------------------------------------------------------
template<class TV,class TA>
CSphBRDF<TV,TA>::CSphBRDF(FILE* file, bool adaptiveDeletion)
: _defaultVal(0,0),
_adaptiveDeletion(adaptiveDeletion)
{
  _ok = false;
  _retVal = 0;
  _coeffs = 0;

  double v1=0;

  if( fscanf(file, " %d %d %d %d %lg ", 
    &_channels, &_maxAllowedOrder, &_xRes, &_yRes, &v1) != 5) 
  {
    FATAL << "Can't read BRDF header from file." << std::endl;
    return;
  }
  _maxError = v1;

  if (_maxAllowedOrder < 1)
  {
    FATAL << "Invalid (H)SH order in file." << std::endl;
    return;
  }
  if (_channels < 1) {
    FATAL << "Invalid number of channels." << std::endl;
    return;
  }
  
  int entries = _numEntries();
  
  _coeffs = new CBRDFRec<TV>*[entries];
  assert(_coeffs);
 
  int i;
  // zero all entries
  for(i=0; i<entries; i++) _coeffs[i] = 0;
  
  for(i=0; i<entries; i++)
  {
    int order;
    TV  error;
    if ( fscanf(file, "order: %d error: %lg\n", &order, &v1) != 2) {
      FATAL << "Can't read sh order." << std::endl;
      return; // error
    }
    error = v1;
    if(order==-1) {
      _coeffs[i] = 0;
      continue;
    }
    if(order<0 || order>_maxAllowedOrder) {
      FATAL << "Invalid order of SH." << std::endl;
      return;
    }
    if(order==0) {
      _coeffs[i] = &_defaultVal;
      continue;
    }
    if(error<0 || error>1) {
      FATAL << "Invalid error of SH expansion." << std::endl;
      return;
    }

    // read the coefficients
    _coeffs[i] = new CBRDFRec<TV>(_channels,order,error);
    assert(_coeffs[i]);
    for(int c=0; c<_channels; c++) 
    {
      // read the BRDF energy
      if( fscanf(file, "brdfen: %lg \ncoeffs: ", &v1) != 1 ) {
        FATAL << "Could not read BRDF energy from file." << std::endl;
        return;
      }
      _coeffs[i]->brdfEnergy[c] = v1;

      for(int index=0; index< order*order; index++) 
      {
        if ( fscanf(file, "%lg ", &v1) != 1) 
        {
          FATAL << "Could not read coefficient from file." << std::endl;
          return;
        }
        _coeffs[i]->SetCoeff(c,index,v1);
      }
    }
  } 

  _retVal = new CBRDFRec<TV>(_channels,_maxAllowedOrder);
  
  // build the stats
  _fillStatistics();
  _stats.Describe(DEBUGW);
  
  // it's OK now
  _ok = true;
}
// --------------------------------------------------------------------------
//  CSphBRDF::SaveToFile()
// --------------------------------------------------------------------------
template<class TV,class TA>
bool
CSphBRDF<TV,TA>::SaveToFile(FILE *file) const
{
  int i;
  int entries = _numEntries();
  
  double v1;
  v1 = _maxError;
  if( fprintf(file, "%d %d %d %d %lg\n", _channels, _maxAllowedOrder, _xRes, _yRes, v1) == EOF )
    return true; // error
  
  for(i=0; i<entries; i++)
  {
    if(_coeffs[i]==0) {
      // -1 as order of NULL's
      if( fprintf(file, "-1 0 \n") == EOF )
        return true; // error
      continue;
    }
    
    int   order = _coeffs[i]->GetOrder(); 
    TV    error = _coeffs[i]->GetError(); 
    v1 = error;
    // write order of this vector
    if( fprintf(file, "order: %d error: %lg\n", order, v1 ) == EOF )
      return true; // error
   
    const TV *brdfEnergy = _coeffs[i]->GetBrdfEnergy();
    // write the coefficients
    if(order>0) {
      for(int c=0; c < _channels; c++) {
        // write brdf energy
        v1 = brdfEnergy ? brdfEnergy[c] : 0.0;
        if( fprintf(file, "brdfen: %lg \ncoeffs: ", v1) == EOF )
          return true; // error
        for(int index=0; index< order*order; index++) {
          v1 = _coeffs[i]->GetCoeff(c, index);
          if ( fprintf(file, "%lg ", v1 ) == EOF)
            return true;
        }
        if ( fputc('\n', file) == EOF) return true;
      }
    }
  }
  
  return false; // no error
}


// --------------------------------------------------------------------------
//  CSphBRDF::CSphBRDF()
// --------------------------------------------------------------------------
template<class TV,class TA>
CSphBRDF<TV,TA>::CSphBRDF(int   channels, 
                          int   maxAllowedOrder, 
                          int   xRes, 
                          int   yRes, 
                          TV    maxError, 
                          bool  adaptiveOrder, 
                          bool  adaptiveDeletion)
                          : _defaultVal(0,0)
{
  assert(xRes>0);
  assert(yRes>0);
  assert(channels>0);
  assert(maxAllowedOrder>0);
  
  _xRes             = xRes;
  _yRes             = yRes;
  _channels         = channels;
  _maxAllowedOrder  = maxAllowedOrder;
  _maxError         = maxError;
  _adaptiveDeletion = adaptiveDeletion;
  
  _coeffs = new CBRDFRec<TV>*[_numEntries()];
  assert(_coeffs);
  // zero all entries
  for(int i=0; i<_numEntries(); i++) _coeffs[i] = &_defaultVal;
  _retVal = new CBRDFRec<TV>(channels,_maxAllowedOrder);
  assert(_retVal);
  _retVal->setz();
}


// --------------------------------------------------------------------------
//  CSphBRDF::CSphBRDF()
// --------------------------------------------------------------------------
template<class TV,class TA>
CSphBRDF<TV,TA>::CSphBRDF(const CBRDF<TV,TA> &sourceBRDF, 
                          int   channels, 
                          int   maxAllowedOrder, 
                          int   xRes,
                          int   yRes,
                          TV    maxError,
                          bool  adaptiveOrder,
                          bool  adaptiveDeletion,
                          const CSphBasisEvalCache<TV,TA>& cache)
                          : _defaultVal(0,0)
{
  assert(xRes>0);
  assert(yRes>0);
  assert(channels>0);
  assert(maxAllowedOrder>0);
  
  int i, in_index;
  _xRes             = xRes;
  _yRes             = yRes;
  _channels         = channels;
  _maxAllowedOrder  = maxAllowedOrder;
  _maxError         = maxError;
  _adaptiveDeletion = adaptiveDeletion;
  
  _coeffs = new CBRDFRec<TV>*[_numEntries()];
  assert(_coeffs);
  // zero all entries
  for(i=0; i<_numEntries(); i++) _coeffs[i] = &_defaultVal;
  _retVal = new CBRDFRec<TV>(channels,_maxAllowedOrder);
  assert(_retVal);
  _retVal->setz();
  
  int   xRes1  = xRes + 1;
  int   yRes1  = yRes + 1;
  TA    deltax = 2.0/xRes;
  TA    deltay = 2.0/yRes;
  
  TV    norm            = cache.GetNormalisation();
  int   numCacheSamples = cache.GetNumSamples();

  cvect3d<TA>   vect, vect1;

#ifdef DOTS
  CDotsPrintCounter dots(xRes1*yRes1, 10, STATUSW);
#endif
  
  TV *brdfEval = new TV[numCacheSamples * _channels];
   
  int maxVectorLen = PAD_FOUR ( _maxAllowedOrder*_maxAllowedOrder );

  // allocate new BRDF record
  CBRDFRec<TV> *coeffs = new CBRDFRec<TV>(channels,_maxAllowedOrder);
  assert(coeffs);
  coeffs->setz();

  int x, y;
  for(x=0; x <= xRes; x++)
  {
    vect.x = x*deltax-1;
    for(y=0; y <= yRes; y++)
    {
      vect.y =  y*deltay-1;
      TA sqrDist = shm::sqr2(vect.x) + shm::sqr2(vect.y);				
      if (sqrDist > 1.0)
        continue;

      vect.z = 0.5 - 0.5 * sqrDist;
      vect1 = Normalize(vect);
          
      // somehow it is much more efficient to first evaluate the brdf for all 
      // directions and then integrate from the resulting array
      TV *brdf = brdfEval;
      shv<TV>::setz(channels*numCacheSamples, brdf);
      for(in_index=0; in_index < numCacheSamples; in_index++, brdf+=channels) {
        const TA *v = cache[in_index].v;
        sourceBRDF.EvalCol(brdf, v, vect1);
      }
     
      // set coefficients to zero
      coeffs->setz();
      // set the brdf energy to zero
      TV *brdfEnergy = coeffs->brdfEnergy;
      shv<TV>::setz(_channels, brdfEnergy);

      // numerical quadrature
      for(in_index=0, brdf = brdfEval; in_index < numCacheSamples; 
        in_index++, brdf+=channels) 
      {
        const TV* H = cache[in_index].coeffs;
        for(int ch=0; ch<channels; ch++) {
          brdfEnergy[ch] += shm::sqr2(brdf[ch]);
          shv<TV>::madvsip4(maxVectorLen, coeffs->GetChannel(ch),H,brdf[ch]);
          if(brdf[ch]>10e+10) {
            PRINTVAR(brdf[ch]);
            PRINTVAR(shm::sqr2(brdf[ch]));
            PRINTVAR(brdfEnergy[ch]);
          }
        }
      }
      // normalize coefficients
      coeffs->mul(norm);
      // normalize brdf energy
      shv<TV>::mulvsip(_channels, brdfEnergy, norm);

      // -------------------------
      //  Adaptive order handling 
      // -------------------------
      if(adaptiveOrder) {
        // create new record, adapted in order
        _coeffs[x*yRes1+y] = coeffs->AdaptOrder(_maxAllowedOrder,_maxError);
        assert(_coeffs[x*yRes1+y]);
      }
      else {
        // use the original record
        coeffs->UpdateRepError();
        _coeffs[x*yRes1+y] = coeffs;
        // and allocate a new one
        coeffs = new CBRDFRec<TV>(channels,_maxAllowedOrder);
        assert(coeffs);
        coeffs->setz();
      }
#ifdef DOTS
      dots.Inc();
#endif
    }
  }

  // delete the superfluos coefficient vector
  delete coeffs;
  
  delete [] brdfEval;

  // compute the stats
  _fillStatistics();

  _stats.Describe(DEBUGW);
}

// --------------------------------------------------------------------------
//  CSphBRDF::_fillStatistics()
// --------------------------------------------------------------------------
template<class TV,class TA>
void 
CSphBRDF<TV,TA>::_fillStatistics()
{
  _stats.SetUp(_maxAllowedOrder);

  /// fill in the statistics
  int   yRes1  = _yRes + 1;
  TA deltax = 2.0/_xRes;
  TA deltay = 2.0/_yRes;
  int x, y;
  cvect3d<TA> vect;
  for(x=0; x <= _xRes; x++) {
    vect.x = x*deltax-1;
    for(y=0; y <= _yRes; y++) {
      vect.y =  y*deltay-1;			
      TA sqrDist = shm::sqr2(vect.x) + shm::sqr2(vect.y);
      if (sqrDist > 1.0f)
        continue;
      
      TA  error = _coeffs[x*yRes1+y]->GetError();
      int order = _coeffs[x*yRes1+y]->GetOrder();
     
      _stats.numDirections ++;
      _stats.avgError += error;
      _stats.maxError = MAX(_stats.maxError, error);
      _stats.minError = MIN(_stats.minError, error);
      _stats.maxOrder = MAX(_stats.maxOrder, order);
      _stats.minOrder = MIN(_stats.minOrder, order);
      _stats.avgOrder += order;
      _stats.numErrorDirections += (error > _maxError);
      _stats.orderHistogram[order] ++;
    }
  }
  _stats.avgError /= (TV)_stats.numDirections;
  _stats.avgOrder /= (TV)_stats.numDirections;
  _stats.order99Quantile = _stats.OrderQuantile(0.99f);
}

// --------------------------------------------------------------------------
//  CSphBRDF::~CSphBRDF()
// --------------------------------------------------------------------------
template<class TV,class TA>
CSphBRDF<TV,TA>::~CSphBRDF()
{
  delete _retVal;
  
  for(int i=0; i< _numEntries(); i++) {
    if (_coeffs[i] != &_defaultVal) 
      delete _coeffs[i];
  }
    
  delete [] _coeffs;
}

// --------------------------------------------------------------------------
//  CSphBRDF::GetOrder()
// --------------------------------------------------------------------------
template<class TV,class TA>
int
CSphBRDF<TV,TA>::GetOrderUV(const TA& u,const TA& v) const
{
  if (u<0.0 || u>1 || v<0 || v>1)
    return 0;

  //BILINEAR INTERPOLATION
  int uindex = (int)(u * _xRes);
  int vindex = (int)(v * _yRes);

  // get the interpolated vectors
  const CBRDFRec<TV> &s00 = *_coeffs[ uindex    *(_yRes+1)+vindex  ] ;
  const CBRDFRec<TV> &s10 = *_coeffs[ (uindex+1)*(_yRes+1)+vindex  ] ;
  const CBRDFRec<TV> &s01 = *_coeffs[ (uindex)  *(_yRes+1)+vindex+1] ;
  const CBRDFRec<TV> &s11 = *_coeffs[ (uindex+1)*(_yRes+1)+vindex+1] ;

  // determine the valid output vector length
  int resorder = s00.GetOrder();
  if( s01.GetOrder() > resorder ) resorder = s01.GetOrder();
  if( s10.GetOrder() > resorder ) resorder = s10.GetOrder();
  if( s11.GetOrder() > resorder ) resorder = s11.GetOrder();

  return resorder;
}

// --------------------------------------------------------------------------
//  CSphBRDF::GetCoeffs()
// --------------------------------------------------------------------------
template<class TV,class TA>
const CBRDFRec<TV>*
CSphBRDF<TV,TA>::GetCoeffsUV(const TA &u,const TA &v) const
{
  // if (vect[2] < 0) return &_defaultVal;
  if (u<0.0 || u>1 || v<0 || v>1)
    return &_defaultVal;

#if 0
  // NEAREST NEIGHBOR INTERPOLATION 
  int uindex = (int)(u * _xRes + TA(0.4999));
  int vindex = (int)(v * _yRes + TA(0.4999));
  if(_coeffs[ uindex*(_yRes+1)+vindex]->GetError() > _maxError)
    return 0;
  else
    return _coeffs[ uindex*(_yRes+1)+vindex];
#else

  //BILINEAR INTERPOLATION
  int uindex = (int)(u * _xRes);
  int vindex = (int)(v * _yRes);

  // get the interpolated vectors
  const CBRDFRec<TV> &s00 = *_coeffs[ uindex    *(_yRes+1)+vindex  ] ;
  const CBRDFRec<TV> &s10 = *_coeffs[ (uindex+1)*(_yRes+1)+vindex  ] ;
  const CBRDFRec<TV> &s01 = *_coeffs[ (uindex)  *(_yRes+1)+vindex+1] ;
  const CBRDFRec<TV> &s11 = *_coeffs[ (uindex+1)*(_yRes+1)+vindex+1] ;
  
  // compute the interpolation weights
  TV du = fabs(uindex-u*_xRes);
  TV dv = fabs(vindex-v*_yRes);
  TV dudv = du*dv;
  TV s11w = dudv;
  TV s00w = 1.0f - du - dv + dudv;
  TV s10w = du - dudv;
  TV s01w = dv - dudv;

  // interpolate error
  _retVal->error = 
    s00.GetError() * s00w + s10.GetError() * s10w +
    s01.GetError() * s01w + s11.GetError() * s11w;
  
  if(_adaptiveDeletion) {		
    // if error is too high, don't use the (H)SH representation
    if(_retVal->error > _maxError)
      return 0;
  }
  
  // determine the valid output vector length
  int resorder = s00.GetOrder();
  if( s01.GetOrder() > resorder ) resorder = s01.GetOrder();
  if( s10.GetOrder() > resorder ) resorder = s10.GetOrder();
  if( s11.GetOrder() > resorder ) resorder = s11.GetOrder();
  
  // set the vector length to the output value
  _retVal -> _setOrder(resorder);
  
  // perform the interpolation for all channels
  _retVal -> setz();
  _retVal -> madvsip( s00, s00w );
  _retVal -> madvsip( s10, s10w );
  _retVal -> madvsip( s01, s01w );
  _retVal -> madvsip( s11, s11w );

  return _retVal;
#endif
}
  
// --------------------------------------------------------------------------
//  CSphBRDF::GetError()
// --------------------------------------------------------------------------
template<class TV,class TA>
TV 
CSphBRDF<TV,TA>::GetErrorUV(const TA& u, const TA &v) const
{
  if (u<0 || u>1 || v<0 || v>1) return 0;

  int uindex = (int)floor(u * _xRes);
  int vindex = (int)floor(v * _yRes);
  
  TV du = fabs(uindex-u*_xRes);
  TV dv = fabs(vindex-v*_yRes);
  
  //BILINEAR INTERPOLATION OF ERROR
  return 
    _coeffs[ uindex    *(_yRes+1)+vindex  ]->GetError() * (1-du)*(1-dv) +
    _coeffs[ (uindex+1)*(_yRes+1)+vindex  ]->GetError() * (  du)*(1-dv) +
    _coeffs[ (uindex)  *(_yRes+1)+vindex+1]->GetError() * (1-du)*(  dv) +
    _coeffs[ (uindex+1)*(_yRes+1)+vindex+1]->GetError() * (  du)*(  dv);
}

// --------------------------------------------------------------------------
//  CBRDFPhongLobe::Eval()
// --------------------------------------------------------------------------
template<class TV,class TA>
inline TV 
CBRDFPhongLobe<TV,TA>::Eval(const TA *L, const TA *V) const
{
  // L ... direction towards light source
  // V ... direction toward viewer
  if(L[2]<0 || V[2]<0) 
    return 0;
  // reflected viewing direction ( for Normal = (0,0,1) )
  TA R[3] = {-V[0], -V[1], V[2]};
  // dot(R,L)
  TV cosR_L = R[0]*L[0] + R[1]*L[1] + R[2]*L[2];
  TV spec_term = pow(cosR_L, _n) * (_n + 2.0) * (0.5*M_1_PI);
  //       |   specular term    |  |cos term|
  //       vvvvvvvvvvvvvvvvvvvvvv   vvvvvvvv
  return        spec_term         *   L[2];
}


// explicit instantiation
template class CSphBRDFStatistics<float>;
template class CSphBRDFStatistics<double>;

template class CBRDFRec<float>;
template class CBRDFRec<double>;

template class CSphBRDF<float,float>;
template class CSphBRDF<float,double>;
template class CSphBRDF<double,float>;
template class CSphBRDF<double,double>;

__END_SPHAR_SOURCE

#endif
