// ===================================================================
// $Id: sheval.cpp,v 1.6 2005/11/29 15:51:50 xkrivanj Exp $
/// \file sheval.cpp
///
/// \brief Evaluation of spherical harmonic basis functions.
///
/// SPHAR_REPLACEMENT_STRING
///
/// Initial coding by Jaroslav Krivanek, December 2003.

// in case the source is included somewhere
#ifndef __SHEVAL_CPP__
#define __SHEVAL_CPP__

// SPHAR headers
#include "shconfig.h"
#include "sheval.h"

__BEGIN_SPHAR_SOURCE

// --------------------------------------------------------------------------
//  shEvalSph()
// --------------------------------------------------------------------------
double shEvalSph(const int l, const int m, double theta, double phi)
{
	double   x = cos(theta);
	if (x >  1) x = 1;
	if (x < -1) x = -1;

	double legendre = assLegendreEval(l, shm::iabs(m), x);

	if (m == 0) return legendre * shNorm<double>( l );
	if (m > 0)  return cos(double(m) * phi) * legendre * sqrt(2.0) * shNorm<double>(l,m);
	else        return sin(double(-m)* phi) * legendre * sqrt(2.0) * shNorm<double>(l,-m);
}

// --------------------------------------------------------------------------
//  shEvalCart()
// --------------------------------------------------------------------------
double shEvalCart(const int l, const int m, double x, double y, double z)
{
	double phi = atan2(y,x);
	if(phi<0) phi += 2.0*M_PI;
	if (z >  1) z = 1;
	if (z < -1) z = -1;

	double legendre = assLegendreEval(l, shm::iabs(m), z);

	if      (m == 0)  return legendre * shNorm<double>( l );
	else if (m > 0)   return shm::cos(double(m) * phi) * legendre * sqrt(2.0) * shNorm<double>(l,m);
	else              return shm::sin(double(-m)* phi) * legendre * sqrt(2.0) * shNorm<double>(l,-m);
}

// --------------------------------------------------------------------------
//  hshEvalSph()
// --------------------------------------------------------------------------
double hshEvalSph(const int l, const int m, double theta, double phi)
{
	// Polynomial shifting
	double x = shm::cos(theta) * 2.0 - 1.0;
	if (x < -1) x = -1;
	if (x > 1) x = 1;

	double legendre = assLegendreEval(l, shm::iabs(m), x);

	if (m == 0)  return legendre * hshNorm<double>( l );
	if (m > 0)   return shm::cos(double(m) * phi) * legendre * sqrt(2.0) * hshNorm<double>(l,m);
	else         return shm::sin(double(-m)* phi) * legendre * sqrt(2.0) * hshNorm<double>(l,-m);
}

// --------------------------------------------------------------------------
//  hshEvalCart()
// --------------------------------------------------------------------------
double hshEvalCart(const int l, const int m, double x, double y, double z)
{
	double phi = atan2(y, x);
	if(phi<0) phi += 2.0*M_PI;
	z = z * 2 - 1;
	if (z >  1) z = 1;
	if (z < -1) z = -1;

	double legendre = assLegendreEval(l, abs(m), z);

	if (m == 0)  return legendre * hshNorm<double>( l );
	if (m > 0)   return shm::cos(double(m) * phi) * legendre * sqrt(2.0) * hshNorm<double>(l,m);
	else         return shm::sin(double(-m)* phi) * legendre * sqrt(2.0) * hshNorm<double>(l,-m);
}

// --------------------------------------------------------------------------
//  assLegendreEval()
// --------------------------------------------------------------------------
double assLegendreEval(const int l,const int m,const double x)
{
	// evaluate an Associated Legendre Polynomial P(l,m,x) at x
	double pmm = double(1.0);
	if(m>0) {
		double somx2 = shm::sqrt((double(1.0)-x)*(double(1.0)+x));
		double fact = 1.0;
		for(int i=1; i<=m; i++) {
			pmm *= (-fact) * somx2;
			fact += double(2.0);
		}
	}

	if(l==m) return pmm;

	double pmmp1 = x * (double(2.0)*m+double(1.0)) * pmm;
	if(l==m+1) return pmmp1;

	double pll = 0.0;
	for(int ll=m+2; ll<=l; ++ll) {
		pll = ( (double(2.0)*ll-double(1.0))*x*pmmp1-(ll+m-double(1.0))*pmm ) / (ll-m);
		pmm = pmmp1;
		pmmp1 = pll;
	}
	return pll;
}

// --------------------------------------------------------------------------
//  dLegNum()
// --------------------------------------------------------------------------
double dLegNum(int l, int m, double x) 
{
	double dx = 0.0000001;
	if(x+dx>1)	dx = -dx;
	double fx   = assLegendreEval(l,m,x);
	double fxdx = assLegendreEval(l,m,x+dx);
	return (fxdx - fx) / dx;
}

// --------------------------------------------------------------------------
//  dLeg()
// --------------------------------------------------------------------------
double dLeg(int l, int m, double x)
{
	if(l<=0) return 0;
	if(l==m) { 		
		if(m<=1)
			return x/shm::sqrt((1.0-x)*(1.0+x));
		else if(m==2) 
			return -6.0*x;
		else if(m>2) {
			double somx2 = shm::sqrt((1.0-x)*(1.0+x));
			double dpmm =  -m * x * 3.0;
			double fact = 5.0;
			for(int i=3; i<=m; i++) {
				dpmm *= (-fact) * somx2;
				fact += 2.0;
			}
			return dpmm;
		}
	}

	// m<l
	double P_lmx  = assLegendreEval(l,m,x);
	double P_l1mx = assLegendreEval(l-1,m,x);
	return 1.0/(x*x-1) * (x*l*P_lmx - (m+l)*P_l1mx);
}

// --------------------------------------------------------------------------
//  dSHThetaNum()
// --------------------------------------------------------------------------
double dSHThetaNum(int l, int m, double theta, double phi) 
{
	double dx = 0.0000001;
	if( theta+dx > M_PI ) dx = -dx;
	double fx   = shEvalSph(l,m,theta,phi);
	double fxdx = shEvalSph(l,m,theta+dx,phi);
	return (fxdx - fx) / dx;
}

// --------------------------------------------------------------------------
//  dSHPhiNum()
// --------------------------------------------------------------------------
double dSHPhiNum(int l, int m, double theta, double phi) 
{
	double dx = 0.0000001;
	if( theta+dx > M_PI*2.0 ) dx = -dx;
	double fx   = shEvalSph(l,m,theta,phi);
	double fxdx = shEvalSph(l,m,theta,phi+dx);
	return (fxdx - fx) / dx;
}

// --------------------------------------------------------------------------
//  dSHTheta()
// --------------------------------------------------------------------------
double dSHTheta(const int l, const int m, double theta, double phi)
{
	double x = shm::cos(theta);
	if (x >  1) x =  1;
	if (x < -1) x = -1;
	double msin_theta = -sqrt(1-x*x);

	double dlegendre = dLeg(l,shm::iabs(m),x);

	if (m == 0)return             shNorm<double>(l)    * dlegendre * msin_theta;
	if (m > 0) return sqrt(2.0) * shNorm<double>(l, m) * dlegendre * msin_theta * cos( phi*m);
	else       return sqrt(2.0) * shNorm<double>(l,-m) * dlegendre * msin_theta * sin(-phi*m);
}

// --------------------------------------------------------------------------
//  dSHPhi()
// --------------------------------------------------------------------------
double dSHPhi(const int l, const int m, double theta, double phi)
{
	double x = shm::cos(theta);
	if (x >  1) x =  1;
	if (x < -1) x = -1;

	double legendre = assLegendreEval(l, shm::iabs(m), x);

	if (m == 0)  return  0;
	if (m > 0)   return (-m) * sqrt(2.0) * shNorm<double>(l, m) * legendre * shm::sin(phi*m);
	else         return (-m) * sqrt(2.0) * shNorm<double>(l,-m) * legendre * shm::cos(phi*m);

	/// the following is also valid for m!=0:
	///    return shEvalSph<double>(l,-m,theta,phi) * (-m);
	/*
	  SH   m>0:  Y = cos( m*phi) * legendre * sqrt(2.0) * N(l,|m|);
	       m<0:  Y = sin(-m*phi) * legendre * sqrt(2.0) * N(l,|m|);
	
	  dSH  m>0: dY = (-m) * sin(phi*m) * legendre * sqrt(2.0) * N(l,|m|);
	       m<0: dY = (-m) * cos(phi*m) * legendre * sqrt(2.0) * N(l,|m|);	  
		     +----------------------------+
		===> | dSH(l,m) =  -m * SH(l,-m); |
		     +----------------------------+
	*/
}


// --------------------------------------------------------------------------
//  _fillNormTable()
// --------------------------------------------------------------------------
template<class TV>
void _fillNormTable(const int order, TV* ret)
{
	for(int l=0, index=0; l<order; l++, index+=2*l ) {
		TV norm = shNorm<TV>( l );
		ret[index] = norm;
		norm *= TV(M_SQRT2);
		// m = 1..l  (&& -1..-l at the same time)
		for(int m=1; m<=l; ++m ) {
			// recurrent normalization factor computation
			norm /= shm::sqrt( TV((l-m+1) * (l+m)) );
			ret[index-m] = norm;
			ret[index+m] = norm;
		}
	}
}
template void _fillNormTable<float>(const int order, float* ret);
template void _fillNormTable<double>(const int order, double* ret);

// --------------------------------------------------------------------------
//  CShNormTable::CShNormTable()
// --------------------------------------------------------------------------
CShNormTable::CShNormTable(int maxOrder) 
{
	_ftable = new float[maxOrder*maxOrder];
	_dtable = new double[maxOrder*maxOrder];
	assert(_ftable);
	assert(_dtable);
	_fillNormTable<float> (maxOrder, _ftable);
	_fillNormTable<double>(maxOrder, _dtable);
}

CShNormTable shNormTab(MAX_FIXED_SH_ORDER);

__END_SPHAR_SOURCE

#endif
