// ===================================================================
// $Id: shlib.cpp,v 1.8 2005/11/29 15:51:50 xkrivanj Exp $
///  \file shlib.cpp
///
/// \brief Various auxiliary classes for (hemi)spherical harmonics.
///
/// SPHAR_REPLACEMENT_STRING
///
/// Initial coding by Jaroslav Krivanek, December 2003.

// in case the source is included somewhere
#ifndef __SHLIB_CPP__
#define __SHLIB_CPP__

// SPHAR headers
#include "shconfig.h"
#include "shlog.h"
#include "shvec.h"
#include "shlib.h"
#include "shbasis.h"

// standard headers


__BEGIN_SPHAR_SOURCE

/////////////////////////////////////////////////////////////////////////
//
//  class CTabLegendre
//
//////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//   CTabLegendre::CTabLegendre()
// --------------------------------------------------------------------------
template<class TV,class TA>
CTabLegendre<TV,TA>::CTabLegendre(int res, int maxOrder, bool derivative)
{
  if(maxOrder<1 || res<2) {
    _array = 0;
    _darray = 0;
    _maxOrder = 0;
    _stride = 0;
    _res = 0;
    _res_min_1 = 0;
    return;
  }
  
  // DEBUG  << " - filling precomputed legendre...";
  
  _stride = mulFour(maxOrder*maxOrder);
  _res    = res;
  _res_min_1 = res-1;
  
  _array = new TV[_stride*res];
  assert(_array);
  shv<TV>::setz(_stride*res, _array);
  if(derivative) {
    _darray = new TV[_stride*res];
    assert(_darray);
    shv<TV>::setz(_stride*res, _darray);
  }
  else {
    _darray=0;
    FATAL_EXIT;
  }
  
  double *tmparray  = new double[_stride];
  assert(tmparray);
  double *tmpdarray = new double[_stride];
  assert(tmpdarray);
  
#define THR 0.999

  TV *c = _array;
  TV *dc = _darray;
  for(int i=0; i<res; i++ , c+=_stride, dc+=_stride) {
    double x = (double)i/(double)(res-1);
    x = x * 2.0 - 1.0;
    assLegendreEvalV<double,double>(maxOrder,x,tmparray);
    for(int j=0;j<maxOrder*maxOrder; j++)
      c[j] = tmparray[j];
    if(derivative) {
      if(x<-THR) x = -THR;
      if(x>THR) x = THR;
      dLegV<double>(maxOrder,x,tmpdarray,tmparray);
      for(int j=0;j<maxOrder*maxOrder; j++)
        dc[j] = tmpdarray[j];
    }
  }
  
  SAFE_DELETE( tmparray );
  SAFE_DELETE( tmpdarray );
  
  // DEBUGW << "done." << std::endl;
}

// --------------------------------------------------------------------------
//   CTabLegendre::~CTabLegendre()
// --------------------------------------------------------------------------
template<class TV,class TA>
CTabLegendre<TV,TA>::~CTabLegendre()
{
  delete _array;
  delete _darray;
}

// --------------------------------------------------------------------------
//   CTabLegendre::EvalLinV()
// --------------------------------------------------------------------------
template<class TV,class TA>
void  
CTabLegendre<TV,TA>::EvalLinV(const int order,const TA x, TV* ret) const
{
  TA f    = (x+TA(1.0)) * TA(0.5) * _res_min_1;
  TA fi   = floor(f);
  TV du   = TV(f-fi);
  int idx = (int)fi;
  
  if(idx>=_res-1)
    shv<TV>::movvv(order*order, ret, _array + ((_res-1)*_stride));
  else {
    shv<TV>::setz(order*order, ret);
    shv<TV>::madvsip(order*order, ret, _array + idx*_stride, (1.0-du));
    shv<TV>::madvsip(order*order, ret, _array + (idx+1)*_stride, du);
  }
}


/////////////////////////////////////////////////////////////////////////////
//
//  class CSphBasisEvalCache
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CSphBasisEvalCache::CSphBasisEvalCache()
// --------------------------------------------------------------------------
template<class TV,class TA>
CSphBasisEvalCache<TV,TA>::CSphBasisEvalCache(int numSamples, 
                                              int order, 
                                              const CSphBasis<TV,TA>* shBasis)
{
  if(numSamples<1 || order<1 || !shBasis) {
    _evalRecs = 0;
    _normValue = 0;
    _numSamples = 0;
    _order = 0;
    _sphBasis = 0;
    return;
  }

  _order    = order;
  _sphBasis = shBasis;
  
  // Compute number of divisions in theta (M) and phi (N).
  // N should be approx Pi*M
  int M, N;
  TA invM, invN;
  if (numSamples==1) {
    M = N  = 1;
    invM = invN = 1.0f;
    _numSamples = 1;
  }
  else {
    M = (int)(sqrt((TA)numSamples*M_1_PI));
    if (M < 1) M = 1;
    N = (int)((TA)numSamples/(TA)M);
    invN = 1.0f / N;
    invM = 1.0f / M;
    _numSamples = M*N;
  }
  
  _normValue = M_PI * 2.0 / (TV)_numSamples;
  
  int vecLen = PAD_FOUR(order*order);
  
  _evalRecs  = new CSphBasisEvalRec<TV,TA>[_numSamples];
  
  CSphBasisEvalRec<TV,TA>* rec = _evalRecs;
  
  TA r1 = 0.0, r2 = 0.0;		
  for(int ir1 = N; ir1-- ; r1 += invN, r2 = 0 ) {
    for(int ir2 = M ; ir2-- ; r2 += invM, rec++ ) {
      // generate jittered tuple
      // CVector2D v(r1 + RND::rand0bf(invN),r2 + RND::rand0bf(invM));
      TA vxx = r1 + 0.5*invN;
      TA vyy = r2 + 0.5*invM;
      // direction in spherical coordinates
      // RND::hemisph_const(c,v);
      TA sin_theta = sqrtf(1.0f - vxx * vxx);
      TA phi       = 2.0 * M_PI * vyy;
      rec->phi    = phi;
      rec->theta  = acos(vxx);
      rec->v[0]   = cos(phi)*sin_theta;
      rec->v[1]   = sin(phi)*sin_theta;
      rec->v[2]   = vxx;
      rec->coeffs = NEW_ALIGN16(TV, vecLen);
      assert(rec->coeffs);
      shv<TV>::setz4(vecLen,rec->coeffs);
      // evaluate the basis
      shBasis->EvalBasisCartV(order, rec->v, rec->coeffs);
    }
  }
}

// --------------------------------------------------------------------------
//  CSphBasisEvalCache::~CSphBasisEvalCache()
// --------------------------------------------------------------------------
template<class TV,class TA>
CSphBasisEvalCache<TV,TA>::~CSphBasisEvalCache() 
{ 
  for(int i=0; i<_numSamples; i++)
  {
    FREE_ALIGN16(_evalRecs[i].coeffs);
  }
  delete _evalRecs;
}

// explicit instatiation
template class CTabLegendre<float,float>;
template class CTabLegendre<float,double>;
template class CTabLegendre<double,float>;
template class CTabLegendre<double,double>;

template class CSphBasisEvalRec<float,float>;
template class CSphBasisEvalRec<float,double>;
template class CSphBasisEvalRec<double,float>;
template class CSphBasisEvalRec<double,double>;

template class CSphBasisEvalCache<float,float>;
template class CSphBasisEvalCache<float,double>;
template class CSphBasisEvalCache<double,float>;
template class CSphBasisEvalCache<double,double>;

__END_SPHAR_SOURCE

#endif
