// ===================================================================
// $Id: shrot.cpp,v 1.13 2005/11/29 15:51:50 xkrivanj Exp $
/// \file shrot.cpp
///
/// \brief Spherical harmonic rotations.
///
/// SPHAR_REPLACEMENT_STRING
///
/// January 2004    Jaroslav Krivanek    Initial coding in shlib.cpp
/// May     2004    Jaroslav Krivanek    Code moved from shlib.cpp to shrot.cpp.

// in case the source is included somewhere
#ifndef __SHROT_CPP__
#define __SHROT_CPP__

// SPHAR headers
#include "shconfig.h"
#include "shrot.h"
#include "shrotmat.h"
#include "shvec.h"

// standard headers
#include <iomanip>

__BEGIN_SPHAR_SOURCE

CSHRotGenerator      rotGenerator(0);
CZXZXZRotationObject zxzxzRotObject(0);


// Generate C++ code for SH rotation around Z-axis.
void 
SHRotZGetCode(ostream& o, 
              const char *out, const char* in, 
              const char *cosn, const char* sinn,
              int order)
{
	// rotate
	for(int l=0, index=0; l<order; l++, index+=2*l) {
    // o << "if(order==" << l << ") return;" << std::endl;
    // out[index] = in[index];
		o << out << "[" << index << "] = " << in << "[" << index << "];\n";
		// m = 1..l  (&& -1..-l at the same time)
		for(int m=1; m<=l; ++m) {
			o 
        << out << "[" << index-m << "] = " << cosn << "[" << m << "]*"
        << in  << "[" << index-m << "] - " << sinn << "[" << m << "]*"
        << in  << "[" << index+m << "];\n";
      o 
        << out << "[" << index+m << "] = " << sinn << "[" << m << "]*"
        << in  << "[" << index-m << "] + " << cosn << "[" << m << "]*"
        << in  << "[" << index+m << "];\n";
		}
	}
}


/////////////////////////////////////////////////////////////////////////////  
//
//  class CSHRotGenerator
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//   CSHRotGenerator::CSHRotGenerator()
// --------------------------------------------------------------------------
CSHRotGenerator::CSHRotGenerator(int order):
_Rxp90(order), 
_Rxm90(order),
_dyMatrix(order),
_ddyMatrix(order)
{
  _precomputed = 0;
  _setNewOrder(order);

  _dySubDiag   = 0;
  _ddyDiag     = 0;
  _ddySSubDiag = 0;
}

// --------------------------------------------------------------------------
//   CSHRotGenerator::~CSHRotGenerator()
// --------------------------------------------------------------------------
CSHRotGenerator::~CSHRotGenerator()
{
  _deletePrecomputed();

  FREE_ALIGN16(_dySubDiag);
  FREE_ALIGN16(_ddyDiag);
  FREE_ALIGN16(_ddySSubDiag);
}

// --------------------------------------------------------------------------
//   CSHRotGenerator::_deletePrecomputed()
// --------------------------------------------------------------------------
void CSHRotGenerator::_deletePrecomputed()
{
  if(_precomputed) {
    for(int i=0; i<_order*_order; i++)
      SAFE_DELETE_ARRAY(_precomputed[i]);
    SAFE_DELETE_ARRAY(_precomputed);
  }
}


// --------------------------------------------------------------------------
//   CSHRotGenerator::_setNewOrder()
// --------------------------------------------------------------------------
void CSHRotGenerator::_setNewOrder(int order)
{
  _deletePrecomputed();
  _order = order;

  if(_order<=0) return;

  _precomputed = new SPreRec*[_order*_order];

  for(int l=0; l < order; l++) {
    for(int m1=-l; m1 <=l; m1++) {
      SPreRec* ppp = new SPreRec[2*l+1];
      _precomputed[l*(l+1)+m1] = ppp;
      for(int m2=-l; m2 <= l; m2++) {
        ppp[l+m2].u = little_u(l, m1, m2);
        ppp[l+m2].v = little_v(l, m1, m2);
        ppp[l+m2].w = little_w(l, m1, m2);
        ppp[l+m2].nz=0;
        if(ppp[l+m2].u != 0) ppp[l+m2].nz |= (unsigned)1;
        if(ppp[l+m2].v != 0) ppp[l+m2].nz |= (unsigned)2;
        if(ppp[l+m2].w != 0) ppp[l+m2].nz |= (unsigned)4;
      }
    }
  }

  ////////////////////////////
  // differential Y-rotation
  ////////////////////////////
  // build the first and second derivative of the Y-rotation matrix
  CSHRotMatrix R(order);
  _dyMatrix.Resize(order);
  _ddyMatrix.Resize(order);
  CSHRotMatrix *dR[3] = {&R,&_dyMatrix,&_ddyMatrix};
  rotGenerator.BuildMatrix_dY(dR,2,order,cos(0.0),sin(0.0));

  // extract sub-diagonal from the first derivative matrix
  // extract diagonal and sub-sub-diagonal from the second deriv. matrix
  FREE_ALIGN16(_dySubDiag);
  FREE_ALIGN16(_ddyDiag);
  FREE_ALIGN16(_ddySSubDiag);

  int vecLen = PAD_FOUR(order*order);

  if(order>0) {
    _dySubDiag = NEW_ALIGN16(float, vecLen);
    assert(_dySubDiag);
    shvf::setz4(vecLen,_dySubDiag);

    _ddyDiag = NEW_ALIGN16(float, vecLen);
    assert(_ddyDiag);
    shvf::setz4(vecLen,_ddyDiag);

    _ddySSubDiag = NEW_ALIGN16(float, vecLen);
    assert(_ddySSubDiag);
    shvf::setz4(vecLen,_ddySSubDiag);

    for(int l=0; l<order; l++) {
      for(int m=-l; m<=l; m++) {
        if(m>-l) _dySubDiag[shIndex(l,m)] = _dyMatrix.GetElement(l,m,m-1);
        if(m>-l+1) _ddySSubDiag[shIndex(l,m)] = _ddyMatrix.GetElement(l,m,m-2);
        _ddyDiag[shIndex(l,m)] = _ddyMatrix.GetElement(l,m,m);
      }
    }
  }
  else {
    _dySubDiag = 0;
    _ddyDiag = 0;
    _ddySSubDiag = 0;
  }

  /////////////////////
  // X(+-90) rotation
  /////////////////////
  // resize the precomputed X-rotation matrices
  _Rxp90.Resize(order);
  _Rxm90.Resize(order);
  float X[3] = {1,0,0};
  BuildMatrixAxisAngle(_Rxp90,order,X,M_PI/2);
  BuildMatrixAxisAngle(_Rxm90,order,X,-M_PI/2);
}


// @@@ use this one
void uvw(const int l, const int m, const int n, double& u, double& v, double& w)
{
  // Pre-calculate simple reusable terms
  double d = shm::delta(m, 0);
  int abs_m = abs(m);

  // Only calculate the required denominator once
  double denom;
  if (abs(n) == l)
    denom = (2 * l) * (2 * l - 1);
  else
    denom = (l + n) * (l - n);

  // Now just calculate the scalars
  u = sqrt((l + m) * (l - m) / denom);
  v = 0.5f * sqrt((1 + d) * (l + abs_m - 1) * (l + abs_m) / denom) * (1 - 2 * d);
  w = -0.5f * sqrt((l - abs_m - 1) * (l - abs_m) / denom) * (1 - d);
}


// --------------------------------------------------------------------------
//   CSHRotGenerator::little_u()
// --------------------------------------------------------------------------
double CSHRotGenerator::little_u(int l, int m1, int m2)
{
  if (abs(m2) < l) 
    return sqrt( ((double)(l+m1)*(l-m1))/((l+m2)*(l-m2)) );
  else
    return sqrt( ((double)(l+m1)*(l-m1))/((2*l)*(2*l-1)) );
}

// --------------------------------------------------------------------------
//   CSHRotGenerator::little_v()
// --------------------------------------------------------------------------
double CSHRotGenerator::little_v(int l, int m1, int m2)
{
  if (abs(m2) < l)
  {
    /*
    return 
    0.5*sqrt(((1.0+delta(m1, 0)))*(l+CMath::iabs(m1)-1.0)*
    (l+CMath::iabs(m1))
    / ( (l+m2)*(l-m2) ) ) * (1.0-2.0*delta(m1, 0));
    */
    /// replacement for the delta function
    if(m1==0) {
      return -0.5*sqrt(2.0*(l+abs(m1)-1.0)*(l+abs(m1))
        / ( (l+m2)*(l-m2) )	) ;
    }
    else {
      return 0.5*sqrt( (l+abs(m1)-1.0)*(l+abs(m1) )
        / ( (l+m2)*(l-m2) ) );
    }
  }
  else
  {
    return 0.5*sqrt(((1.0+shm::delta(m1, 0)))*(l+abs(m1)-1.0)*(l+abs(m1))
      / ( (2.0*l)*(2.0*l-1.0) ) ) * (1.0-2.0*shm::delta(m1, 0));
  }
}

// --------------------------------------------------------------------------
//   CSHRotGenerator::little_w()
// --------------------------------------------------------------------------
double CSHRotGenerator::little_w(int l, int m1, int m2)
{
  if (m1 == 0) return 0;

  if (abs(m2) < l)
  {
    return -0.5*sqrt((l-abs(m1)-1.0)*(l-abs(m1))
      /
      ( (l+m2)*(l-m2) ) );
  }
  else
  {
    return -0.5*sqrt((l-abs(m1)-1.0)*(l-abs(m1))
      /
      ( (2.0*l)*(2.0*l-1.0) ) );
  }
}

// --------------------------------------------------------------------------
//   CSHRotGenerator::U()
// --------------------------------------------------------------------------
inline double CSHRotGenerator::U(int l, int m1, int m2)
{
  return  P(l, m1, m2, 0);
}

// --------------------------------------------------------------------------
//   CSHRotGenerator::V()
// --------------------------------------------------------------------------
inline double CSHRotGenerator::V (int l, int m1, int m2)
{
  if(m1>=0) {
    if(m1==0)      return P (l, 1, m2, 1)+P (l, -1, m2, -1);
    else if(m1==1) return M_SQRT2 * P (l, 0, m2, 1);
    else            return P (l, m1-1, m2, 1) - P (l, -m1+1, m2, -1);
  }
  else {
    if(m1 == -1)  return M_SQRT2 * P (l, 0, m2, -1);
    else           return P (l, -m1-1, m2, -1) + P (l, m1+1, m2, 1);
  }
}

// --------------------------------------------------------------------------
//   CSHRotGenerator::W()
// --------------------------------------------------------------------------
inline double CSHRotGenerator::W(int l, int m1, int m2)
{
  if (m1 > 0)
    return P(l, m1+1, m2, 1) + P(l, -m1-1, m2, -1);
  else
    return P(l, m1-1, m2, 1) - P(l, -m1+1, m2, -1);
}


// --------------------------------------------------------------------------
//  CSHRotGenerator::BuildMatrixY(
// --------------------------------------------------------------------------
void CSHRotGenerator::BuildMatrixY(CSHRotMatrix &shmat, int order, 
                                   double cos_beta, double sin_beta)
{
  if(order > shmat.GetOrder()) order=shmat.GetOrder();
  if(order<1) return;
  if( order > _order) _setNewOrder(order);

  _dR[0] = &shmat;

  _dR[0]->SetElement(0,0,0,1);
  if (order==1) return;

  _dR[0]->SetElement(1,  -1,-1,  1 );
  _dR[0]->SetElement(1,  -1, 0,  0 );
  _dR[0]->SetElement(1,  -1, 1,  0 );

  _dR[0]->SetElement(1,   0,-1,   0  );
  _dR[0]->SetElement(1,   0, 0,   cos_beta );
  _dR[0]->SetElement(1,   0, 1,  -sin_beta );

  _dR[0]->SetElement(1,   1,-1,   0  );
  _dR[0]->SetElement(1,   1, 0,   sin_beta );
  _dR[0]->SetElement(1,   1, 1,   cos_beta );

  if (order == 2) return;

  for(int l=2; l <order; l++) {
    for(int m1=-l; m1 <=l; m1++) {
      double* mat = _dR[0]->_matrix[shIndex(l,m1)];
      for(int m2=-l; m2 <= l; m2++) {
        const SPreRec& little = _getGenericValue(l, m1, m2);
        mat[l+m2] = 0;
        if(little.nz & 1) mat[l+m2] += little.u * U(l, m1, m2);
        if(little.nz & 2) mat[l+m2] += little.v * V(l, m1, m2);				
        if(little.nz & 4) mat[l+m2] += little.w * W(l, m1, m2);
      }
    }
  }
}

// --------------------------------------------------------------------------
//   CSHRotGenerator::BuildMatrix()
// --------------------------------------------------------------------------
void 
CSHRotGenerator::BuildMatrix(CSHRotMatrix &shmat, int order,
                             double r00, double r01, double r02,
                             double r10, double r11, double r12,
                             double r20, double r21, double r22)
{
  if(order > shmat.GetOrder()) order=shmat.GetOrder();
  if(order<1) return;
  if( order > _order) _setNewOrder(order);

  _dR[0] = &shmat;

  _dR[0]->SetElement(0,0,0,1);
  if (order==1) return;

  _dR[0]->SetElement(1,-1,-1,  r11 );
  _dR[0]->SetElement(1,-1, 0, -r21 );
  _dR[0]->SetElement(1,-1, 1,  r01 );

  _dR[0]->SetElement(1,0,-1,  -r12 );
  _dR[0]->SetElement(1,0, 0,   r22 );
  _dR[0]->SetElement(1,0, 1,  -r02 );

  _dR[0]->SetElement(1,1,-1,  r10 );
  _dR[0]->SetElement(1,1, 0, -r20 );
  _dR[0]->SetElement(1,1, 1,  r00 );

  if (order == 2) return;

  for(int l=2; l <order; l++)
  {
    for(int m1=-l; m1 <=l; m1++)
    {
      double* mat = _dR[0]->_matrix[shIndex(l,m1)];
      for(int m2=-l; m2 <= l; m2++)
      {
        const SPreRec& little = _getGenericValue(l, m1, m2);
        mat[l+m2] = 0;
        if(little.nz & 1)  // if (little.u != 0)
          mat[l+m2] += little.u * U(l, m1, m2);
        if(little.nz & 2)  // if (little.v != 0)
          mat[l+m2] += little.v * V(l, m1, m2);
        if(little.nz & 4)  // if (little.w != 0)
          mat[l+m2] += little.w * W(l, m1, m2);
      }
    }
  }
  // rotationMatrixDump();
}


// --------------------------------------------------------------------------
//   CSHRotGenerator::BuildMatrixZYZ()
// --------------------------------------------------------------------------
void 
CSHRotGenerator::BuildMatrixZYZ(CSHRotMatrix &shmat, int order,
                                double AZ1, double AY, double AZ2)
{
  double ca=cos(AZ1), cb=cos(AY);
  double cc=cos(AZ2), sa=sin(AZ1);
  double sb=sin(AY),  sc=sin(AZ2);

  double r00 = ca*cb*cc-sa*sc;
  double r10 = sa*cb*cc+ca*sc;
  double r20 = -sb*cc;
  double r01 = -ca*cb*sc-sa*cc;
  double r11 = -sa*cb*sc+ca*cc; 
  double r21 = sb*sc;
  double r02 = ca*sb;
  double r12 = sa*sb;
  double r22 = cb;

  BuildMatrix(shmat,order,
    r00, r01, r02,
    r10, r11, r12,
    r20, r21, r22);
}


// --------------------------------------------------------------------------
//   CSHRotGenerator::BuildMatrixAxisCosSin()
// --------------------------------------------------------------------------
void 
CSHRotGenerator::BuildMatrixAxisCosSin(CSHRotMatrix &shmat, 
                                       int    order, 
                                       const  float* axis, 
                                       double c,
                                       double s)
{
  double x=axis[0], y=axis[1], z=axis[2];

  double r00 = x*x+c*(1-x*x);
  double r01 = x*y*(1-c)-z*s;
  double r02 = x*z*(1-c)+y*s;

  double r10 = x*y*(1-c)+z*s;
  double r11 = y*y+c*(1-y*y);
  double r12 = y*z*(1-c)-x*s;

  double r20 = x*z*(1-c)-y*s;
  double r21 = y*z*(1-c)+x*s;
  double r22 = z*z+c*(1-z*z);

  BuildMatrix(shmat,order,
    r00, r01, r02,
    r10, r11, r12,
    r20, r21, r22);
}

// --------------------------------------------------------------------------
//  CSHRotGenerator::RotYdiff1GenCode()
// --------------------------------------------------------------------------
void CSHRotGenerator::RotYdiff1GenCode(ostream &o,
                                       int order,
                                       const char* dest, 
                                       const char* src)
{
  o << "if(order==0)return;"<< std::endl;
  o << dest << "[0]=" << src << "[0];" << std::endl;

  for(int l=1; l<order; l++) {
    o << "if(order==" << l << ")return;"<< std::endl;
    for(int m=-l; m<=l; m++) {
      int i=shIndex(l,m);
      o << std::setprecision(5);
      o << dest << "[" << i << "]=";
      o << src << "[" << i << "]*(1.0f+yy*" << _ddyDiag[i]<<")+";
      o << "y*(" << _dySubDiag[i] << "*" << src << "[" << i-1 << "]";
      if(i<order*order) o << "- "<<_dySubDiag[i+1] << "*" << src << "[" << i+1 << "]";
      o << ");" << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------
//   CSHRotGenerator::BuildMatrixNum_dAxis0()
// -----------------------------------------------------------------------------
void 
CSHRotGenerator::BuildMatrixNum_dAxis0(CSHRotMatrix &dR,
                                       int order,
                                       const float* axis)
{
  double h=0.00001;
  CSHRotMatrix fx_min_h(order);

  rotGenerator.BuildMatrixAxisAngle(dR,        order,  axis,  h);
  rotGenerator.BuildMatrixAxisAngle(fx_min_h , order,  axis, -h);

  dR.sub(fx_min_h);
  dR.mul(2.0/h);
}

// -----------------------------------------------------------------------------
//   CSHRotGenerator::BuildMatrixNum_ddAxis0()
// -----------------------------------------------------------------------------
void 
CSHRotGenerator::BuildMatrixNum_ddAxis0(CSHRotMatrix &ddR,
                                        int order,
                                        const  float* axis)
{
  double h=0.00001;
  CSHRotMatrix fx_min_h(order), fx_plus_h(order);

  rotGenerator.BuildMatrixAxisAngle(ddR,       order,  axis,  0);
  rotGenerator.BuildMatrixAxisAngle(fx_min_h , order,  axis, -h);
  rotGenerator.BuildMatrixAxisAngle(fx_plus_h, order,  axis,  h);

  ddR.mul(-2);
  ddR.add(fx_min_h);
  ddR.add(fx_plus_h);
  ddR.mul(1.0/(h*h));
}

// --------------------------------------------------------------------------
//  CZXZXZRotationObject::CZXZXZRotationObject()
// --------------------------------------------------------------------------
CZXZXZRotationObject::CZXZXZRotationObject(int order)
{
  _cnz1 = _cnz2 = _cny = _snz1 = _snz2 = _sny = 0;
  _tmpvec1 = _tmpvec2 = 0;
  SetNewOrder(order);
}

// --------------------------------------------------------------------------
//  CZXZXZRotationObject::~CZXZXZRotationObject()
// --------------------------------------------------------------------------
CZXZXZRotationObject::~CZXZXZRotationObject()
{
  _deleteData();
}

// --------------------------------------------------------------------------
//  CZXZXZRotationObject::SetNewOrder()
// --------------------------------------------------------------------------
void CZXZXZRotationObject::SetNewOrder(int order)
{
  _deleteData();

  if(order<=0)
    return;

  int orderPad    = PAD_FOUR(order);
  int sqrOrderPad = PAD_FOUR(order*order);

  _cnz1    = NEW_ALIGN16(float,orderPad);
  _cnz2    = NEW_ALIGN16(float,orderPad);
  _cny     = NEW_ALIGN16(float,orderPad);
  _snz1    = NEW_ALIGN16(float,orderPad);
  _snz2    = NEW_ALIGN16(float,orderPad);
  _sny     = NEW_ALIGN16(float,orderPad);
  _tmpvec1 = NEW_ALIGN16(float,sqrOrderPad);
  _tmpvec2 = NEW_ALIGN16(float,sqrOrderPad);
  assert(_cnz1);
  assert(_cnz2);
  assert(_cny);
  assert(_snz1);
  assert(_snz2);
  assert(_sny);
  assert(_tmpvec1);
  assert(_tmpvec2);

  _order = order;
}

// --------------------------------------------------------------------------
//  CZXZXZRotationObject::_deleteData
// --------------------------------------------------------------------------
void CZXZXZRotationObject::_deleteData()
{
  FREE_ALIGN16(_cnz1);
  FREE_ALIGN16(_cnz2);
  FREE_ALIGN16(_cny);
  FREE_ALIGN16(_snz1);
  FREE_ALIGN16(_snz2);
  FREE_ALIGN16(_sny);
  FREE_ALIGN16(_tmpvec1);
  FREE_ALIGN16(_tmpvec2);
  _cnz1 = _cnz2 = _cny = _snz1 = _snz2 = _sny = 0;
  _tmpvec1 = _tmpvec2 = 0;
}

/*
// --------------------------------------------------------------------------
//  some SIMD operators
// --------------------------------------------------------------------------
inline __m128 operator*(const __m128& a, const __m128& b)
{
  return _mm_mul_ps(a,b);
}

inline __m128 operator*(const __m128& a, const float& f)
{
  return _mm_mul_ps(a,_mm_set_ps1(f));
}

inline __m128 operator*(const float& f, const __m128& a)
{
  return a*f;
}

inline __m128 operator+(const __m128& a, const __m128& b)
{
  return _mm_add_ps(a,b);
}

inline __m128 operator-(const __m128& a, const __m128& b)
{
  return _mm_sub_ps(a,b);
}

inline __m128 operator-(const __m128& a)
{
  return _mm_sub_ps(_mm_setzero_ps(),a);
}

// --------------------------------------------------------------------------
//  CZXZXZRotationObject::TransformVectorUnroll()
// --------------------------------------------------------------------------
void 
CZXZXZRotationObject::TransformVectorUnrollSSE(float *outf,
                                               const float *inf, 
                                               int order,
                                               const CSHRotGenerator& rotGenerator)
{
  if(order>5) {
    exit(1);
    return;
  }

  //// Generated code for order 5:
  __m128 _tmpvec1[25];
  __m128 _tmpvec2[25];
  __m128 *out = (__m128*)outf;
  const __m128 *in = (const __m128*)inf;

  if(order<=0) return;
  out[0] = in[0];

  if(order==1) return;
  _tmpvec1[2] = in[2];
  _tmpvec1[1] = _cnz1[1]*in[1] - _snz1[1]*in[3];
  _tmpvec1[3] = _snz1[1]*in[1] + _cnz1[1]*in[3];

  _tmpvec2[1]=_tmpvec1[2];
  _tmpvec2[2]=-_tmpvec1[1];
  _tmpvec2[3]=_tmpvec1[3];

  _tmpvec1[2] = _tmpvec2[2];
  _tmpvec1[1] = _cny[1]*_tmpvec2[1] - _sny[1]*_tmpvec2[3];
  _tmpvec1[3] = _sny[1]*_tmpvec2[1] + _cny[1]*_tmpvec2[3];

  _tmpvec2[1]=-_tmpvec1[2];
  _tmpvec2[2]=_tmpvec1[1];
  _tmpvec2[3]=_tmpvec1[3];

  out[2] = _tmpvec2[2];
  out[1] = _cnz2[1]*_tmpvec2[1] - _snz2[1]*_tmpvec2[3];
  out[3] = _snz2[1]*_tmpvec2[1] + _cnz2[1]*_tmpvec2[3];

  if(order==2) return;

  _tmpvec1[6] = in[6];
  _tmpvec1[5] = _cnz1[1]*in[5] - _snz1[1]*in[7];
  _tmpvec1[7] = _snz1[1]*in[5] + _cnz1[1]*in[7];
  _tmpvec1[4] = _cnz1[2]*in[4] - _snz1[2]*in[8];
  _tmpvec1[8] = _snz1[2]*in[4] + _cnz1[2]*in[8];

  _tmpvec2[4]=_tmpvec1[7];
  _tmpvec2[5]=-_tmpvec1[5];
  _tmpvec2[6]=_tmpvec1[6]*-0.5f+_tmpvec1[8]*-0.86603f;
  _tmpvec2[7]=-_tmpvec1[4];
  _tmpvec2[8]=_tmpvec1[6]*-0.86603f+_tmpvec1[8]*0.5f;

  _tmpvec1[6] = _tmpvec2[6];
  _tmpvec1[5] = _cny[1]*_tmpvec2[5] - _sny[1]*_tmpvec2[7];
  _tmpvec1[7] = _sny[1]*_tmpvec2[5] + _cny[1]*_tmpvec2[7];
  _tmpvec1[4] = _cny[2]*_tmpvec2[4] - _sny[2]*_tmpvec2[8];
  _tmpvec1[8] = _sny[2]*_tmpvec2[4] + _cny[2]*_tmpvec2[8];

  _tmpvec2[4]=-_tmpvec1[7];
  _tmpvec2[5]=-_tmpvec1[5];
  _tmpvec2[6]=_tmpvec1[6]*-0.5f+_tmpvec1[8]*-0.86603f;
  _tmpvec2[7]=_tmpvec1[4];
  _tmpvec2[8]=_tmpvec1[6]*-0.86603f+_tmpvec1[8]*0.5f;

  out[6] = _tmpvec2[6];
  out[5] = _cnz2[1]*_tmpvec2[5] - _snz2[1]*_tmpvec2[7];
  out[7] = _snz2[1]*_tmpvec2[5] + _cnz2[1]*_tmpvec2[7];
  out[4] = _cnz2[2]*_tmpvec2[4] - _snz2[2]*_tmpvec2[8];
  out[8] = _snz2[2]*_tmpvec2[4] + _cnz2[2]*_tmpvec2[8];

  if(order==3) return;

  _tmpvec1[12] = in[12];
  _tmpvec1[11] = _cnz1[1]*in[11] - _snz1[1]*in[13];
  _tmpvec1[13] = _snz1[1]*in[11] + _cnz1[1]*in[13];
  _tmpvec1[10] = _cnz1[2]*in[10] - _snz1[2]*in[14];
  _tmpvec1[14] = _snz1[2]*in[10] + _cnz1[2]*in[14];
  _tmpvec1[9] = _cnz1[3]*in[9] - _snz1[3]*in[15];
  _tmpvec1[15] = _snz1[3]*in[9] + _cnz1[3]*in[15];

  _tmpvec2[9]=_tmpvec1[12]*-0.79057f+_tmpvec1[14]*0.61237f;
  _tmpvec2[10]=-_tmpvec1[10];
  _tmpvec2[11]=_tmpvec1[12]*-0.61237f+_tmpvec1[14]*-0.79057f;
  _tmpvec2[12]=_tmpvec1[9]*0.79057f+_tmpvec1[11]*0.61237f;
  _tmpvec2[13]=_tmpvec1[13]*-0.25f+_tmpvec1[15]*-0.96825f;
  _tmpvec2[14]=_tmpvec1[9]*-0.61237f+_tmpvec1[11]*0.79057f;
  _tmpvec2[15]=_tmpvec1[13]*-0.96825f+_tmpvec1[15]*0.25f;

  _tmpvec1[12] = _tmpvec2[12];
  _tmpvec1[11] = _cny[1]*_tmpvec2[11] - _sny[1]*_tmpvec2[13];
  _tmpvec1[13] = _sny[1]*_tmpvec2[11] + _cny[1]*_tmpvec2[13];
  _tmpvec1[10] = _cny[2]*_tmpvec2[10] - _sny[2]*_tmpvec2[14];
  _tmpvec1[14] = _sny[2]*_tmpvec2[10] + _cny[2]*_tmpvec2[14];
  _tmpvec1[9] = _cny[3]*_tmpvec2[9] - _sny[3]*_tmpvec2[15];
  _tmpvec1[15] = _sny[3]*_tmpvec2[9] + _cny[3]*_tmpvec2[15];

  _tmpvec2[9]=_tmpvec1[12]*0.79057f+_tmpvec1[14]*-0.61237f;
  _tmpvec2[10]=-_tmpvec1[10];
  _tmpvec2[11]=_tmpvec1[12]*0.61237f+_tmpvec1[14]*0.79057f;
  _tmpvec2[12]=_tmpvec1[9]*-0.79057f+_tmpvec1[11]*-0.61237f;
  _tmpvec2[13]=_tmpvec1[13]*-0.25f+_tmpvec1[15]*-0.96825f;
  _tmpvec2[14]=_tmpvec1[9]*0.61237f+_tmpvec1[11]*-0.79057f;
  _tmpvec2[15]=_tmpvec1[13]*-0.96825f+_tmpvec1[15]*0.25f;

  out[12] = _tmpvec2[12];
  out[11] = _cnz2[1]*_tmpvec2[11] - _snz2[1]*_tmpvec2[13];
  out[13] = _snz2[1]*_tmpvec2[11] + _cnz2[1]*_tmpvec2[13];
  out[10] = _cnz2[2]*_tmpvec2[10] - _snz2[2]*_tmpvec2[14];
  out[14] = _snz2[2]*_tmpvec2[10] + _cnz2[2]*_tmpvec2[14];
  out[9] = _cnz2[3]*_tmpvec2[9] - _snz2[3]*_tmpvec2[15];
  out[15] = _snz2[3]*_tmpvec2[9] + _cnz2[3]*_tmpvec2[15];

  if(order==4) return;

  _tmpvec1[20] = in[20];
  _tmpvec1[19] = _cnz1[1]*in[19] - _snz1[1]*in[21];
  _tmpvec1[21] = _snz1[1]*in[19] + _cnz1[1]*in[21];
  _tmpvec1[18] = _cnz1[2]*in[18] - _snz1[2]*in[22];
  _tmpvec1[22] = _snz1[2]*in[18] + _cnz1[2]*in[22];
  _tmpvec1[17] = _cnz1[3]*in[17] - _snz1[3]*in[23];
  _tmpvec1[23] = _snz1[3]*in[17] + _cnz1[3]*in[23];
  _tmpvec1[16] = _cnz1[4]*in[16] - _snz1[4]*in[24];
  _tmpvec1[24] = _snz1[4]*in[16] + _cnz1[4]*in[24];

  _tmpvec2[16]=_tmpvec1[21]*-0.93541f+_tmpvec1[23]*0.35355f;
  _tmpvec2[17]=_tmpvec1[17]*-0.75f+_tmpvec1[19]*0.66144f;
  _tmpvec2[18]=_tmpvec1[21]*-0.35355f+_tmpvec1[23]*-0.93541f;
  _tmpvec2[19]=_tmpvec1[17]*0.66144f+_tmpvec1[19]*0.75f;
  _tmpvec2[20]=_tmpvec1[20]*0.375f+_tmpvec1[22]*0.55902f+_tmpvec1[24]*0.73951f;
  _tmpvec2[21]=_tmpvec1[16]*0.93541f+_tmpvec1[18]*0.35355f;
  _tmpvec2[22]=_tmpvec1[20]*0.55902f+_tmpvec1[22]*0.5f+_tmpvec1[24]*-0.66144f;
  _tmpvec2[23]=_tmpvec1[16]*-0.35355f+_tmpvec1[18]*0.93541f;
  _tmpvec2[24]=_tmpvec1[20]*0.73951f+_tmpvec1[22]*-0.66144f+_tmpvec1[24]*0.125f;

  _tmpvec1[20] = _tmpvec2[20];
  _tmpvec1[19] = _cny[1]*_tmpvec2[19] - _sny[1]*_tmpvec2[21];
  _tmpvec1[21] = _sny[1]*_tmpvec2[19] + _cny[1]*_tmpvec2[21];
  _tmpvec1[18] = _cny[2]*_tmpvec2[18] - _sny[2]*_tmpvec2[22];
  _tmpvec1[22] = _sny[2]*_tmpvec2[18] + _cny[2]*_tmpvec2[22];
  _tmpvec1[17] = _cny[3]*_tmpvec2[17] - _sny[3]*_tmpvec2[23];
  _tmpvec1[23] = _sny[3]*_tmpvec2[17] + _cny[3]*_tmpvec2[23];
  _tmpvec1[16] = _cny[4]*_tmpvec2[16] - _sny[4]*_tmpvec2[24];
  _tmpvec1[24] = _sny[4]*_tmpvec2[16] + _cny[4]*_tmpvec2[24];

  _tmpvec2[16]=_tmpvec1[21]*0.93541f+_tmpvec1[23]*-0.35355f;
  _tmpvec2[17]=_tmpvec1[17]*-0.75f+_tmpvec1[19]*0.66144f;
  _tmpvec2[18]=_tmpvec1[21]*0.35355f+_tmpvec1[23]*0.93541f;
  _tmpvec2[19]=_tmpvec1[17]*0.66144f+_tmpvec1[19]*0.75f;
  _tmpvec2[20]=_tmpvec1[20]*0.375f+_tmpvec1[22]*0.55902f+_tmpvec1[24]*0.73951f;
  _tmpvec2[21]=_tmpvec1[16]*-0.93541f+_tmpvec1[18]*-0.35355f;
  _tmpvec2[22]=_tmpvec1[20]*0.55902f+_tmpvec1[22]*0.5f+_tmpvec1[24]*-0.66144f;
  _tmpvec2[23]=_tmpvec1[16]*0.35355f+_tmpvec1[18]*-0.93541f;
  _tmpvec2[24]=_tmpvec1[20]*0.73951f+_tmpvec1[22]*-0.66144f+_tmpvec1[24]*0.125f;

  out[20] = _tmpvec2[20];
  out[19] = _cnz2[1]*_tmpvec2[19] - _snz2[1]*_tmpvec2[21];
  out[21] = _snz2[1]*_tmpvec2[19] + _cnz2[1]*_tmpvec2[21];
  out[18] = _cnz2[2]*_tmpvec2[18] - _snz2[2]*_tmpvec2[22];
  out[22] = _snz2[2]*_tmpvec2[18] + _cnz2[2]*_tmpvec2[22];
  out[17] = _cnz2[3]*_tmpvec2[17] - _snz2[3]*_tmpvec2[23];
  out[23] = _snz2[3]*_tmpvec2[17] + _cnz2[3]*_tmpvec2[23];
  out[16] = _cnz2[4]*_tmpvec2[16] - _snz2[4]*_tmpvec2[24];
  out[24] = _snz2[4]*_tmpvec2[16] + _cnz2[4]*_tmpvec2[24];

  return;
}
*/

// --------------------------------------------------------------------------
//  CZXZXZRotationObject::TransformVectorUnroll()
// --------------------------------------------------------------------------
void 
CZXZXZRotationObject::TransformVectorUnroll(float *out, 
                                            const float *in, 
                                            int order,
                                            const CSHRotGenerator& rotGenerator)
{
  if(order>5) {
    TransformVector(out,in,order,rotGenerator);
    return;
  }

  //// Generated code for order 5:
  float _tmpvec1[25];
  float _tmpvec2[25];

  if(order<=0) return;
  out[0] = in[0];

  if(order==1) return;
  RotateBand1<float>(out,in,_U,_V,_N);

  if(order==2) return;

  _tmpvec1[6] = in[6];
  _tmpvec1[5] = _cnz1[1]*in[5] - _snz1[1]*in[7];
  _tmpvec1[7] = _snz1[1]*in[5] + _cnz1[1]*in[7];
  _tmpvec1[4] = _cnz1[2]*in[4] - _snz1[2]*in[8];
  _tmpvec1[8] = _snz1[2]*in[4] + _cnz1[2]*in[8];

  _tmpvec2[4]=_tmpvec1[7];
  _tmpvec2[5]=-_tmpvec1[5];
  _tmpvec2[6]=_tmpvec1[6]*-0.5f+_tmpvec1[8]*-0.86603f;
  _tmpvec2[7]=-_tmpvec1[4];
  _tmpvec2[8]=_tmpvec1[6]*-0.86603f+_tmpvec1[8]*0.5f;

  _tmpvec1[6] = _tmpvec2[6];
  _tmpvec1[5] = _cny[1]*_tmpvec2[5] - _sny[1]*_tmpvec2[7];
  _tmpvec1[7] = _sny[1]*_tmpvec2[5] + _cny[1]*_tmpvec2[7];
  _tmpvec1[4] = _cny[2]*_tmpvec2[4] - _sny[2]*_tmpvec2[8];
  _tmpvec1[8] = _sny[2]*_tmpvec2[4] + _cny[2]*_tmpvec2[8];

  _tmpvec2[4]=-_tmpvec1[7];
  _tmpvec2[5]=-_tmpvec1[5];
  _tmpvec2[6]=_tmpvec1[6]*-0.5f+_tmpvec1[8]*-0.86603f;
  _tmpvec2[7]=_tmpvec1[4];
  _tmpvec2[8]=_tmpvec1[6]*-0.86603f+_tmpvec1[8]*0.5f;

  out[6] = _tmpvec2[6];
  out[5] = _cnz2[1]*_tmpvec2[5] - _snz2[1]*_tmpvec2[7];
  out[7] = _snz2[1]*_tmpvec2[5] + _cnz2[1]*_tmpvec2[7];
  out[4] = _cnz2[2]*_tmpvec2[4] - _snz2[2]*_tmpvec2[8];
  out[8] = _snz2[2]*_tmpvec2[4] + _cnz2[2]*_tmpvec2[8];

  if(order==3) return;

  _tmpvec1[12] = in[12];
  _tmpvec1[11] = _cnz1[1]*in[11] - _snz1[1]*in[13];
  _tmpvec1[13] = _snz1[1]*in[11] + _cnz1[1]*in[13];
  _tmpvec1[10] = _cnz1[2]*in[10] - _snz1[2]*in[14];
  _tmpvec1[14] = _snz1[2]*in[10] + _cnz1[2]*in[14];
  _tmpvec1[9] = _cnz1[3]*in[9] - _snz1[3]*in[15];
  _tmpvec1[15] = _snz1[3]*in[9] + _cnz1[3]*in[15];

  _tmpvec2[9]=_tmpvec1[12]*-0.79057f+_tmpvec1[14]*0.61237f;
  _tmpvec2[10]=-_tmpvec1[10];
  _tmpvec2[11]=_tmpvec1[12]*-0.61237f+_tmpvec1[14]*-0.79057f;
  _tmpvec2[12]=_tmpvec1[9]*0.79057f+_tmpvec1[11]*0.61237f;
  _tmpvec2[13]=_tmpvec1[13]*-0.25f+_tmpvec1[15]*-0.96825f;
  _tmpvec2[14]=_tmpvec1[9]*-0.61237f+_tmpvec1[11]*0.79057f;
  _tmpvec2[15]=_tmpvec1[13]*-0.96825f+_tmpvec1[15]*0.25f;

  _tmpvec1[12] = _tmpvec2[12];
  _tmpvec1[11] = _cny[1]*_tmpvec2[11] - _sny[1]*_tmpvec2[13];
  _tmpvec1[13] = _sny[1]*_tmpvec2[11] + _cny[1]*_tmpvec2[13];
  _tmpvec1[10] = _cny[2]*_tmpvec2[10] - _sny[2]*_tmpvec2[14];
  _tmpvec1[14] = _sny[2]*_tmpvec2[10] + _cny[2]*_tmpvec2[14];
  _tmpvec1[9] = _cny[3]*_tmpvec2[9] - _sny[3]*_tmpvec2[15];
  _tmpvec1[15] = _sny[3]*_tmpvec2[9] + _cny[3]*_tmpvec2[15];

  _tmpvec2[9]=_tmpvec1[12]*0.79057f+_tmpvec1[14]*-0.61237f;
  _tmpvec2[10]=-_tmpvec1[10];
  _tmpvec2[11]=_tmpvec1[12]*0.61237f+_tmpvec1[14]*0.79057f;
  _tmpvec2[12]=_tmpvec1[9]*-0.79057f+_tmpvec1[11]*-0.61237f;
  _tmpvec2[13]=_tmpvec1[13]*-0.25f+_tmpvec1[15]*-0.96825f;
  _tmpvec2[14]=_tmpvec1[9]*0.61237f+_tmpvec1[11]*-0.79057f;
  _tmpvec2[15]=_tmpvec1[13]*-0.96825f+_tmpvec1[15]*0.25f;

  out[12] = _tmpvec2[12];
  out[11] = _cnz2[1]*_tmpvec2[11] - _snz2[1]*_tmpvec2[13];
  out[13] = _snz2[1]*_tmpvec2[11] + _cnz2[1]*_tmpvec2[13];
  out[10] = _cnz2[2]*_tmpvec2[10] - _snz2[2]*_tmpvec2[14];
  out[14] = _snz2[2]*_tmpvec2[10] + _cnz2[2]*_tmpvec2[14];
  out[9] = _cnz2[3]*_tmpvec2[9] - _snz2[3]*_tmpvec2[15];
  out[15] = _snz2[3]*_tmpvec2[9] + _cnz2[3]*_tmpvec2[15];

  if(order==4) return;

  _tmpvec1[20] = in[20];
  _tmpvec1[19] = _cnz1[1]*in[19] - _snz1[1]*in[21];
  _tmpvec1[21] = _snz1[1]*in[19] + _cnz1[1]*in[21];
  _tmpvec1[18] = _cnz1[2]*in[18] - _snz1[2]*in[22];
  _tmpvec1[22] = _snz1[2]*in[18] + _cnz1[2]*in[22];
  _tmpvec1[17] = _cnz1[3]*in[17] - _snz1[3]*in[23];
  _tmpvec1[23] = _snz1[3]*in[17] + _cnz1[3]*in[23];
  _tmpvec1[16] = _cnz1[4]*in[16] - _snz1[4]*in[24];
  _tmpvec1[24] = _snz1[4]*in[16] + _cnz1[4]*in[24];

  _tmpvec2[16]=_tmpvec1[21]*-0.93541f+_tmpvec1[23]*0.35355f;
  _tmpvec2[17]=_tmpvec1[17]*-0.75f+_tmpvec1[19]*0.66144f;
  _tmpvec2[18]=_tmpvec1[21]*-0.35355f+_tmpvec1[23]*-0.93541f;
  _tmpvec2[19]=_tmpvec1[17]*0.66144f+_tmpvec1[19]*0.75f;
  _tmpvec2[20]=_tmpvec1[20]*0.375f+_tmpvec1[22]*0.55902f+_tmpvec1[24]*0.73951f;
  _tmpvec2[21]=_tmpvec1[16]*0.93541f+_tmpvec1[18]*0.35355f;
  _tmpvec2[22]=_tmpvec1[20]*0.55902f+_tmpvec1[22]*0.5f+_tmpvec1[24]*-0.66144f;
  _tmpvec2[23]=_tmpvec1[16]*-0.35355f+_tmpvec1[18]*0.93541f;
  _tmpvec2[24]=_tmpvec1[20]*0.73951f+_tmpvec1[22]*-0.66144f+_tmpvec1[24]*0.125f;

  _tmpvec1[20] = _tmpvec2[20];
  _tmpvec1[19] = _cny[1]*_tmpvec2[19] - _sny[1]*_tmpvec2[21];
  _tmpvec1[21] = _sny[1]*_tmpvec2[19] + _cny[1]*_tmpvec2[21];
  _tmpvec1[18] = _cny[2]*_tmpvec2[18] - _sny[2]*_tmpvec2[22];
  _tmpvec1[22] = _sny[2]*_tmpvec2[18] + _cny[2]*_tmpvec2[22];
  _tmpvec1[17] = _cny[3]*_tmpvec2[17] - _sny[3]*_tmpvec2[23];
  _tmpvec1[23] = _sny[3]*_tmpvec2[17] + _cny[3]*_tmpvec2[23];
  _tmpvec1[16] = _cny[4]*_tmpvec2[16] - _sny[4]*_tmpvec2[24];
  _tmpvec1[24] = _sny[4]*_tmpvec2[16] + _cny[4]*_tmpvec2[24];

  _tmpvec2[16]=_tmpvec1[21]*0.93541f+_tmpvec1[23]*-0.35355f;
  _tmpvec2[17]=_tmpvec1[17]*-0.75f+_tmpvec1[19]*0.66144f;
  _tmpvec2[18]=_tmpvec1[21]*0.35355f+_tmpvec1[23]*0.93541f;
  _tmpvec2[19]=_tmpvec1[17]*0.66144f+_tmpvec1[19]*0.75f;
  _tmpvec2[20]=_tmpvec1[20]*0.375f+_tmpvec1[22]*0.55902f+_tmpvec1[24]*0.73951f;
  _tmpvec2[21]=_tmpvec1[16]*-0.93541f+_tmpvec1[18]*-0.35355f;
  _tmpvec2[22]=_tmpvec1[20]*0.55902f+_tmpvec1[22]*0.5f+_tmpvec1[24]*-0.66144f;
  _tmpvec2[23]=_tmpvec1[16]*0.35355f+_tmpvec1[18]*-0.93541f;
  _tmpvec2[24]=_tmpvec1[20]*0.73951f+_tmpvec1[22]*-0.66144f+_tmpvec1[24]*0.125f;

  out[20] = _tmpvec2[20];
  out[19] = _cnz2[1]*_tmpvec2[19] - _snz2[1]*_tmpvec2[21];
  out[21] = _snz2[1]*_tmpvec2[19] + _cnz2[1]*_tmpvec2[21];
  out[18] = _cnz2[2]*_tmpvec2[18] - _snz2[2]*_tmpvec2[22];
  out[22] = _snz2[2]*_tmpvec2[18] + _cnz2[2]*_tmpvec2[22];
  out[17] = _cnz2[3]*_tmpvec2[17] - _snz2[3]*_tmpvec2[23];
  out[23] = _snz2[3]*_tmpvec2[17] + _cnz2[3]*_tmpvec2[23];
  out[16] = _cnz2[4]*_tmpvec2[16] - _snz2[4]*_tmpvec2[24];
  out[24] = _snz2[4]*_tmpvec2[16] + _cnz2[4]*_tmpvec2[24];

  return;
}

// --------------------------------------------------------------------------
//  CZXZXZRotationObject::ToGlobalFrame()
// --------------------------------------------------------------------------
template<class TA>
void 
CZXZXZRotationObject::ToGlobalFrame(const TA *U, 
                                    const TA *V, 
                                    const TA *N,
                                    int order)
{
  assert(order>0);
  if(_order<order)SetNewOrder(order);

//  DEBUG << "CZXZXZRotationObject::ToGlobalFrame()" << std::endl;

  _U[0]=U[0]; _U[1]=U[1];  _U[2]=U[2];
  _V[0]=V[0]; _V[1]=V[1];  _V[2]=V[2];
  _N[0]=N[0]; _N[1]=N[1];  _N[2]=N[2];
  float cz1,sz1,cy,sy,cz2,sz2;
  if( EulerAnglesZYZ<float>(_U,_V,_N,cz1,sz1,cy,sy,cz2,sz2,1-1e-5) ) 
  {
    // I'm not sure this is right.
    if(_N[2]>0) { 
      cz1 = cy = 1;
      sz1 = sy = 0;
      cz2 = U[0]; sz2 = V[0]; 
    }
    else {
      cz1 = U[0];
      sz1 = V[0];
      cy  = -1;
      sy  = 0;
      cz1 = 1;
      sz1 = 0;
    }
  }
  recurrentSinCos<float>(_order,cz1,sz1,_cnz1,_snz1);
  recurrentSinCos<float>(_order,cy ,sy ,_cny ,_sny );
  recurrentSinCos<float>(_order,cz2,sz2,_cnz2,_snz2);
}

template void CZXZXZRotationObject::ToGlobalFrame(
  const float *U,const float *V,const float *N,int order);

__END_SPHAR_SOURCE

#endif
