// ===================================================================
// $Id: shrotmat.cpp,v 1.8 2005/11/29 15:51:51 xkrivanj Exp $
/// \file shrotmat.cpp
///
/// \brief Spherical harmonic rotation matrix.
///
/// SPHAR_REPLACEMENT_STRING
///
/// January 2004    Jaroslav Krivanek    Initial coding in shlib.cpp
/// May     2004    Jaroslav Krivanek    Code moved from shlib.cpp to shrot.cpp.
/// May     2004    Jaroslav Krivanek    Code moved from shrot.cpp to shrotmat.cpp.

// in case the source is included somewhere
#ifndef __SHROTMAT_CPP__
#define __SHROTMAT_CPP__

// SPHAR headers
#include "shconfig.h"
#include "shrotmat.h"

// standard headers
#include <iomanip>

__BEGIN_SPHAR_SOURCE

/////////////////////////////////////////////////////////////////////////////
//
//  class CSHRotMatrix
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  CSHRotMatrix::CSHRotMatrix()
// --------------------------------------------------------------------------
CSHRotMatrix::CSHRotMatrix(int order)
{
	assert(order>=0);
  _matrix = 0;
  _order  = 0;
  Resize(order);
}

// --------------------------------------------------------------------------
//  CSHRotMatrix::Resize()
// --------------------------------------------------------------------------
void CSHRotMatrix::Resize(int newOrder)
{
	assert(newOrder>=0);
  int l,m1;
  if(newOrder==_order) return;

  double **newMatrix = newOrder>0 ? new double*[newOrder * newOrder] : 0;

  int mo=min(_order,newOrder);
  // copy the old data
	for(l=0; l < mo; l++)
		for(m1=-l; m1 <=l; m1++)
      newMatrix[shIndex(l,m1)] = _matrix[shIndex(l,m1)];
  // delete excess records (newOrder < _order)
	for(l=newOrder; l < _order; l++)
		for(m1=-l; m1 <=l; m1++)
      delete _matrix[shIndex(l,m1)];
  // allocate new records (newOrder > _order)
	for(l=_order; l < newOrder; l++)
		for(m1=-l; m1 <=l; m1++)
			newMatrix[shIndex(l,m1)] = new double[2*l+1];
  // delete old pointer array
  delete _matrix;
  // update internals
  _matrix = newMatrix;
  _order  = newOrder;
}

// --------------------------------------------------------------------------
//  CSHRotMatrix::~CSHRotMatrix()
// --------------------------------------------------------------------------
CSHRotMatrix::~CSHRotMatrix()
{
	for(int l=0; l < _order; l++) {
		for(int m1=-l; m1 <=l; m1++) {
			delete [] _matrix[shIndex(l,m1)];
		}
	}
	delete [] _matrix;
}

// --------------------------------------------------------------------------
//  CSHRotMatrix::TransformVector()
// --------------------------------------------------------------------------
void 
CSHRotMatrix::TransformVector(float *out, const float* in, int order) const
{
	if(order<0) order = _order;
	else if(order>_order) order=_order;

	for(int l=0, lindex=0; l < order; l++, lindex+=2*l) {
		for(int m1=-l; m1 <= l ; m1++, out++) {
			*out = 0;
			const double* elem = _matrix[lindex+m1];
			for(int m2=-l; m2 <= l ; m2++)
				*out += in[lindex+m2] * (float)elem[l+m2]; // GetElement(l, m1, m2);
		}
	}
}

// --------------------------------------------------------------------------
//  CSHRotMatrix::TransformVectorGenCode()
// --------------------------------------------------------------------------
void 
CSHRotMatrix::TransformVectorGenCode(ostream& o, const char *out, const char* in, int order) const
{
	if(order<0) order = _order;
	else if(order>_order) order=_order;

	for(int l=0, lindex=0; l < order; l++, lindex+=2*l) {
    o << "if(order==" << l << ") return;" << std::endl;
		for(int m1=-l; m1 <= l ; m1++) {
			const double* elem = _matrix[lindex+m1];
      /*
			*out = 0;
			for(int m2=-l; m2 <= l ; m2++)
				*out += in[lindex+m2] * (float)elem[l+m2]; // GetElement(l, m1, m2);
      */
      o << out << "[" << shIndex(l,m1) << "]="; 
      for(int m2=-l; m2 <= l ; m2++) {
        if(fabs(elem[l+m2])>1e-10) {
          if(elem[l+m2]==1.0)
            o << in << "[" << lindex+m2 << "]+";
          else if(elem[l+m2]==-1.0)
            o << "-" << in << "[" << lindex+m2 << "]+";
          else
            o << in << "[" << lindex+m2 << "]*"
            << std::setprecision(5) << elem[l+m2] << "f+";
        }
      }
      o << "0.0f;"<<std::endl;
		}
	}
}

// --------------------------------------------------------------------------
//  CSHRotMatrix::TransformVectorXp90()
// --------------------------------------------------------------------------
void 
CSHRotMatrix::TransformVectorXp90(float *out, const float* in, int order) const
{
  assert(order<=_order);
	if(order<0) order = _order;

#if 0
  // reading matrix elements from 'this'
	for(int l=0, lindex=0; l < order; l++, lindex+=2*l) {
		for(int m1=-l; m1 <= l ; m1++, out++) {
			*out = 0;
			const double* elem = _matrix[lindex+m1];
			for(int m2=-l; m2 <= l ; m2++)
				*out += in[lindex+m2] * (float)elem[l+m2]; // GetElement(l, m1, m2);
		}
	}
#else

  // matrix elements hardcoded in C++
if(order==0) return;
out[0]=in[0]+0.0f;
if(order==1) return;
out[1]=-in[2]+0.0f;
out[2]=in[1]+0.0f;
out[3]=in[3]+0.0f;
if(order==2) return;
out[4]=-in[7]+0.0f;
out[5]=in[5]*-1.0f+0.0f;
out[6]=in[6]*-0.5f+in[8]*-0.86603f+0.0f;
out[7]=in[4]+0.0f;
out[8]=in[6]*-0.86603f+in[8]*0.5f+0.0f;
if(order==3) return;
out[9]=in[12]*0.79057f+in[14]*-0.61237f+0.0f;
out[10]=-in[10]+0.0f;
out[11]=in[12]*0.61237f+in[14]*0.79057f+0.0f;
out[12]=in[9]*-0.79057f+in[11]*-0.61237f+0.0f;
out[13]=in[13]*-0.25f+in[15]*-0.96825f+0.0f;
out[14]=in[9]*0.61237f+in[11]*-0.79057f+0.0f;
out[15]=in[13]*-0.96825f+in[15]*0.25f+0.0f;
if(order==4) return;
out[16]=in[21]*0.93541f+in[23]*-0.35355f+0.0f;
out[17]=in[17]*-0.75f+in[19]*0.66144f+0.0f;
out[18]=in[21]*0.35355f+in[23]*0.93541f+0.0f;
out[19]=in[17]*0.66144f+in[19]*0.75f+0.0f;
out[20]=in[20]*0.375f+in[22]*0.55902f+in[24]*0.73951f+0.0f;
out[21]=in[16]*-0.93541f+in[18]*-0.35355f+0.0f;
out[22]=in[20]*0.55902f+in[22]*0.5f+in[24]*-0.66144f+0.0f;
out[23]=in[16]*0.35355f+in[18]*-0.93541f+0.0f;
out[24]=in[20]*0.73951f+in[22]*-0.66144f+in[24]*0.125f+0.0f;
if(order==5) return;
out[25]=in[30]*-0.70156f+in[32]*0.68465f+in[34]*-0.19764f+0.0f;
out[26]=in[26]*-0.5f+in[28]*0.86603f+0.0f;
out[27]=in[30]*-0.52291f+in[32]*-0.30619f+in[34]*0.7955f+0.0f;
out[28]=in[26]*0.86603f+in[28]*0.5f+0.0f;
out[29]=in[30]*-0.48412f+in[32]*-0.66144f+in[34]*-0.57282f+0.0f;
out[30]=in[25]*0.70156f+in[27]*0.52291f+in[29]*0.48412f+0.0f;
out[31]=in[31]*0.125f+in[33]*0.40505f+in[35]*0.90571f+0.0f;
out[32]=in[25]*-0.68465f+in[27]*0.30619f+in[29]*0.66144f+0.0f;
out[33]=in[31]*0.40505f+in[33]*0.8125f+in[35]*-0.41926f+0.0f;
out[34]=in[25]*0.19764f+in[27]*-0.7955f+in[29]*0.57282f+0.0f;
out[35]=in[31]*0.90571f+in[33]*-0.41926f+in[35]*0.0625f+0.0f;
if(order==6) return;
out[36]=in[43]*-0.87945f+in[45]*0.46351f+in[47]*-0.10825f+0.0f;
out[37]=in[37]*-0.3125f+in[39]*0.80283f+in[41]*-0.50775f+0.0f;
out[38]=in[43]*-0.43301f+in[45]*-0.68465f+in[47]*0.5863f+0.0f;
out[39]=in[37]*0.80283f+in[39]*-0.0625f+in[41]*-0.59293f+0.0f;
out[40]=in[43]*-0.19764f+in[45]*-0.5625f+in[47]*-0.80283f+0.0f;
out[41]=in[37]*-0.50775f+in[39]*-0.59293f+in[41]*-0.625f+0.0f;
out[42]=in[42]*-0.3125f+in[44]*-0.45286f+in[46]*-0.49608f+in[48]*-0.67169f+0.0f;
out[43]=in[36]*0.87945f+in[38]*0.43301f+in[40]*0.19764f+0.0f;
out[44]=in[42]*-0.45286f+in[44]*-0.53125f+in[46]*-0.17116f+in[48]*0.69527f+0.0f;
out[45]=in[36]*-0.46351f+in[38]*0.68465f+in[40]*0.5625f+0.0f;
out[46]=in[42]*-0.49608f+in[44]*-0.17116f+in[46]*0.8125f+in[48]*-0.25388f+0.0f;
out[47]=in[36]*0.10825f+in[38]*-0.5863f+in[40]*0.80283f+0.0f;
out[48]=in[42]*-0.67169f+in[44]*0.69527f+in[46]*-0.25388f+in[48]*0.03125f+0.0f;
if(order==7) return;
out[49]=in[56]*0.64726f+in[58]*-0.69912f+in[60]*0.29811f+in[62]*-0.058463f+0.0f;
out[50]=in[50]*-0.1875f+in[52]*0.63738f+in[54]*-0.74739f+0.0f;
out[51]=in[56]*0.47496f+in[58]*0.073288f+in[60]*-0.78125f+in[62]*0.39836f+0.0f;
out[52]=in[50]*0.63738f+in[52]*-0.5f+in[54]*-0.5863f+0.0f;
out[53]=in[56]*0.42962f+in[58]*0.41984f+in[60]*-0.10364f+in[62]*-0.79273f+0.0f;
out[54]=in[50]*-0.74739f+in[52]*-0.5863f+in[54]*-0.3125f+0.0f;
out[55]=in[56]*0.4134f+in[58]*0.5741f+in[60]*0.53855f+in[62]*0.45768f+0.0f;
out[56]=in[49]*-0.64726f+in[51]*-0.47496f+in[53]*-0.42962f+in[55]*-0.4134f+0.0f;
out[57]=in[57]*-0.078125f+in[59]*-0.24357f+in[61]*-0.44879f+in[63]*-0.85624f+0.0f;
out[58]=in[49]*0.69912f+in[51]*-0.073288f+in[53]*-0.41984f+in[55]*-0.5741f+0.0f;
out[59]=in[57]*-0.24357f+in[59]*-0.60938f+in[61]*-0.57004f+in[63]*0.49435f+0.0f;
out[60]=in[49]*-0.29811f+in[51]*0.78125f+in[53]*0.10364f+in[55]*-0.53855f+0.0f;
out[61]=in[57]*-0.44879f+in[59]*-0.57004f+in[61]*0.67188f+in[63]*-0.14905f+0.0f;
out[62]=in[49]*0.058463f+in[51]*-0.39836f+in[53]*0.79273f+in[55]*-0.45768f+0.0f;
out[63]=in[57]*-0.85624f+in[59]*0.49435f+in[61]*-0.14905f+in[63]*0.015625f+0.0f;
if(order==8) return;
out[64]=in[73]*0.83561f+in[75]*-0.51633f+in[77]*0.18488f+in[79]*-0.03125f+0.0f;
out[65]=in[65]*-0.10937f+in[67]*0.46219f+in[69]*-0.7745f+in[71]*0.4178f+0.0f;
out[66]=in[73]*0.45768f+in[75]*0.47135f+in[77]*-0.70883f+in[79]*0.25674f+0.0f;
out[67]=in[65]*0.46219f+in[67]*-0.70312f+in[69]*-0.21819f+in[71]*0.49435f+0.0f;
out[68]=in[73]*0.27422f+in[75]*0.60515f+in[77]*0.33802f+in[79]*-0.66659f+0.0f;
out[69]=in[65]*-0.7745f+in[67]*-0.21819f+in[69]*0.26563f+in[71]*0.53102f+0.0f;
out[70]=in[73]*0.13073f+in[75]*0.38081f+in[77]*0.59086f+in[79]*0.69912f+0.0f;
out[71]=in[65]*0.4178f+in[67]*0.49435f+in[69]*0.53102f+in[71]*0.54688f+0.0f;
out[72]=in[72]*0.27344f+in[74]*0.39218f+in[76]*0.41133f+in[78]*0.45768f+in[80]*0.62671f+0.0f;
out[73]=in[64]*-0.83561f+in[66]*-0.45768f+in[68]*-0.27422f+in[70]*-0.13073f+0.0f;
out[74]=in[72]*0.39218f+in[74]*0.5f+in[76]*0.32775f+in[80]*-0.69912f+0.0f;
out[75]=in[64]*0.51633f+in[66]*-0.47135f+in[68]*-0.60515f+in[70]*-0.38081f+0.0f;
out[76]=in[72]*0.41133f+in[74]*0.32775f+in[76]*-0.28125f+in[78]*-0.73021f+in[80]*0.33329f+0.0f;
out[77]=in[64]*-0.18488f+in[66]*0.70883f+in[68]*-0.33802f+in[70]*-0.59086f+0.0f;
out[78]=in[72]*0.45768f+in[76]*-0.73021f+in[78]*0.5f+in[80]*-0.085582f+0.0f;
out[79]=in[64]*0.03125f+in[66]*-0.25674f+in[68]*0.66659f+in[70]*-0.69912f+0.0f;
out[80]=in[72]*0.62671f+in[74]*-0.69912f+in[76]*0.33329f+in[78]*-0.085582f+in[80]*0.0078125f+0.0f;
if(order==9) return;
out[81]=in[90]*-0.60905f+in[92]*0.69685f+in[94]*-0.36158f+in[96]*0.11158f+in[98]*-0.016573f+0.0f;
out[82]=in[82]*-0.0625f+in[84]*0.31561f+in[86]*-0.68179f+in[88]*0.65699f+0.0f;
out[83]=in[90]*-0.44315f+in[92]*0.056337f+in[94]*0.67233f+in[96]*-0.56833f+in[98]*0.15944f+0.0f;
out[84]=in[82]*0.31561f+in[84]*-0.71875f+in[86]*0.20252f+in[88]*0.58547f+0.0f;
out[85]=in[90]*-0.39636f+in[92]*-0.25195f+in[94]*0.39218f+in[96]*0.60515f+in[98]*-0.50931f+0.0f;
out[86]=in[82]*-0.68179f+in[84]*0.20252f+in[86]*0.5625f+in[88]*0.42159f+0.0f;
out[87]=in[90]*-0.37549f+in[92]*-0.42962f+in[94]*-0.138f+in[96]*0.29811f+in[98]*0.75268f+0.0f;
out[88]=in[82]*0.65699f+in[84]*0.58547f+in[86]*0.42159f+in[88]*0.21875f+0.0f;
out[89]=in[90]*-0.36685f+in[92]*-0.51301f+in[94]*-0.49435f+in[96]*-0.45768f+in[98]*-0.3852f+0.0f;
out[90]=in[81]*0.60905f+in[83]*0.44315f+in[85]*0.39636f+in[87]*0.37549f+in[89]*0.36685f+0.0f;
out[91]=in[91]*0.054688f+in[93]*0.16792f+in[95]*0.29543f+in[97]*0.46242f+in[99]*0.81713f+0.0f;
out[92]=in[81]*-0.69685f+in[83]*-0.056337f+in[85]*0.25195f+in[87]*0.42962f+in[89]*0.51301f+0.0f;
out[93]=in[91]*0.16792f+in[93]*0.45313f+in[95]*0.57728f+in[97]*0.38725f+in[99]*-0.53223f+0.0f;
out[94]=in[81]*0.36158f+in[83]*-0.67233f+in[85]*-0.39218f+in[87]*0.138f+in[89]*0.49435f+0.0f;
out[95]=in[91]*0.29543f+in[93]*0.57728f+in[95]*0.14063f+in[97]*-0.71624f+in[99]*0.21608f+0.0f;
out[96]=in[81]*-0.11158f+in[83]*0.56833f+in[85]*-0.60515f+in[87]*-0.29811f+in[89]*0.45768f+0.0f;
out[97]=in[91]*0.46242f+in[93]*0.38725f+in[95]*-0.71624f+in[97]*0.34766f+in[99]*-0.048318f+0.0f;
out[98]=in[81]*0.016573f+in[83]*-0.15944f+in[85]*0.50931f+in[87]*-0.75268f+in[89]*0.3852f+0.0f;
out[99]=in[91]*0.81713f+in[93]*-0.53223f+in[95]*0.21608f+in[97]*-0.048318f+in[99]*0.0039063f+0.0f;
if(order==10) return;
out[100]=in[111]*-0.80045f+in[113]*0.5438f+in[115]*-0.24319f+in[117]*0.065945f+in[119]*-0.0087346f+0.0f;
out[101]=in[101]*-0.035156f+in[103]*0.20644f+in[105]*-0.5438f+in[107]*0.72958f+in[109]*-0.35797f+0.0f;
out[102]=in[111]*-0.46456f+in[113]*-0.31561f+in[115]*0.70572f+in[117]*-0.42101f+in[119]*0.096319f+0.0f;
out[103]=in[101]*0.20644f+in[103]*-0.62891f+in[105]*0.50419f+in[107]*0.35433f+in[109]*-0.42673f+0.0f;
out[104]=in[111]*-0.31049f+in[113]*-0.53906f+in[115]*0.017469f+in[117]*0.69255f+in[119]*-0.36479f+0.0f;
out[105]=in[101]*-0.5438f+in[103]*0.50419f+in[105]*0.48437f+in[107]*-0.034939f+in[109]*-0.46285f+0.0f;
out[106]=in[111]*-0.19516f+in[113]*-0.48614f+in[115]*-0.49411f+in[117]*-0.091109f+in[119]*0.68786f+0.0f;
out[107]=in[101]*0.72958f+in[103]*0.35433f+in[105]*-0.034939f+in[107]*-0.32813f+in[109]*-0.48299f+0.0f;
out[108]=in[111]*-0.094722f+in[113]*-0.27885f+in[115]*-0.44538f+in[117]*-0.57487f+in[119]*-0.62002f+0.0f;
out[109]=in[101]*-0.35797f+in[103]*-0.42673f+in[105]*-0.46285f+in[107]*-0.48299f+in[109]*-0.49219f+0.0f;
out[110]=in[110]*-0.24609f+in[112]*-0.35124f+in[114]*-0.36183f+in[116]*-0.38378f+in[118]*-0.43066f+in[120]*-0.59363f+0.0f;
out[111]=in[100]*0.80045f+in[102]*0.46456f+in[104]*0.31049f+in[106]*0.19516f+in[108]*0.094722f+0.0f;
out[112]=in[110]*-0.35124f+in[112]*-0.46484f+in[114]*-0.36619f+in[116]*-0.18922f+in[118]*0.10058f+in[120]*0.69321f+0.0f;
out[113]=in[100]*-0.5438f+in[102]*0.31561f+in[104]*0.53906f+in[106]*0.48614f+in[108]*0.27885f+0.0f;
out[114]=in[110]*-0.36183f+in[112]*-0.36619f+in[114]*0.015625f+in[116]*0.45851f+in[118]*0.61372f+in[120]*-0.38452f+0.0f;
out[115]=in[100]*0.24319f+in[102]*-0.70572f+in[104]*-0.017469f+in[106]*0.49411f+in[108]*0.44538f+0.0f;
out[116]=in[110]*-0.38378f+in[112]*-0.18922f+in[114]*0.45851f+in[116]*0.46289f+in[118]*-0.61149f+in[120]*0.13595f+0.0f;
out[117]=in[100]*-0.065945f+in[102]*0.42101f+in[104]*-0.69255f+in[106]*0.091109f+in[108]*0.57487f+0.0f;
out[118]=in[110]*-0.43066f+in[112]*0.10058f+in[114]*0.61372f+in[116]*-0.61149f+in[118]*0.23047f+in[120]*-0.026922f+0.0f;
out[119]=in[100]*0.0087346f+in[102]*-0.096319f+in[104]*0.36479f+in[106]*-0.68786f+in[108]*0.62002f+0.0f;
out[120]=in[110]*-0.59363f+in[112]*0.69321f+in[114]*-0.38452f+in[116]*0.13595f+in[118]*-0.026922f+in[120]*0.0019531f+0.0f;
if(order==11) return;
out[121]=in[132]*0.57998f+in[134]*-0.68875f+in[136]*0.40329f+in[138]*-0.15847f+in[140]*0.038323f+in[142]*-0.0045805f+0.0f;
out[122]=in[122]*-0.019531f+in[124]*0.13073f+in[126]*-0.40544f+in[128]*0.68786f+in[130]*-0.58737f+0.0f;
out[123]=in[132]*0.41976f+in[134]*-0.13595f+in[136]*-0.55723f+in[138]*0.63604f+in[140]*-0.29501f+in[142]*0.05696f+0.0f;
out[124]=in[122]*0.13073f+in[124]*-0.5f+in[126]*0.64613f+in[130]*-0.56163f+0.0f;
out[125]=in[132]*0.37297f+in[134]*0.1369f+in[136]*-0.49511f+in[138]*-0.32426f+in[140]*0.65554f+in[142]*-0.2493f+0.0f;
out[126]=in[122]*-0.40544f+in[124]*0.64613f+in[126]*0.18359f+in[128]*-0.41751f+in[130]*-0.45838f+0.0f;
out[127]=in[132]*0.35034f+in[134]*0.31014f+in[136]*-0.13435f+in[138]*-0.52273f+in[140]*-0.40785f+in[142]*0.56872f+0.0f;
out[128]=in[122]*0.68786f+in[126]*-0.41751f+in[128]*-0.5f+in[130]*-0.32022f+0.0f;
out[129]=in[132]*0.33846f+in[134]*0.41655f+in[136]*0.23535f+in[138]*-0.047642f+in[140]*-0.39402f+in[142]*-0.70641f+0.0f;
out[130]=in[122]*-0.58737f+in[124]*-0.56163f+in[126]*-0.45838f+in[128]*-0.32022f+in[130]*-0.16406f+0.0f;
out[131]=in[132]*0.33321f+in[134]*0.46765f+in[136]*0.45638f+in[138]*0.43553f+in[140]*0.40022f+in[142]*0.33485f+0.0f;
out[132]=in[121]*-0.57998f+in[123]*-0.41976f+in[125]*-0.37297f+in[127]*-0.35034f+in[129]*-0.33846f+in[131]*-0.33321f+0.0f;
out[133]=in[133]*-0.041016f+in[135]*-0.12498f+in[137]*-0.21562f+in[139]*-0.32136f+in[141]*-0.46502f+in[143]*-0.7853f+0.0f;
out[134]=in[121]*0.68875f+in[123]*0.13595f+in[125]*-0.1369f+in[127]*-0.31014f+in[129]*-0.41655f+in[131]*-0.46765f+0.0f;
out[135]=in[133]*-0.12498f+in[135]*-0.34961f+in[137]*-0.49531f+in[139]*-0.49717f+in[141]*-0.25434f+in[143]*0.55223f+0.0f;
out[136]=in[121]*-0.40329f+in[123]*0.55723f+in[125]*0.49511f+in[127]*0.13435f+in[129]*-0.23535f+in[131]*-0.45638f+0.0f;
out[137]=in[133]*-0.21562f+in[135]*-0.49531f+in[137]*-0.38574f+in[139]*0.14035f+in[141]*0.68449f+in[143]*-0.26675f+0.0f;
out[138]=in[121]*0.15847f+in[123]*-0.63604f+in[125]*0.32426f+in[127]*0.52273f+in[129]*0.047642f+in[131]*-0.43553f+0.0f;
out[139]=in[133]*-0.32136f+in[135]*-0.49717f+in[137]*0.14035f+in[139]*0.62793f+in[141]*-0.4781f+in[143]*0.083523f+0.0f;
out[140]=in[121]*-0.038323f+in[123]*0.29501f+in[125]*-0.65554f+in[127]*0.40785f+in[129]*0.39402f+in[131]*-0.40022f+0.0f;
out[141]=in[133]*-0.46502f+in[135]*-0.25434f+in[137]*0.68449f+in[139]*-0.4781f+in[141]*0.14746f+in[143]*-0.014842f+0.0f;
out[142]=in[121]*0.0045805f+in[123]*-0.05696f+in[125]*0.2493f+in[127]*-0.56872f+in[129]*0.70641f+in[131]*-0.33485f+0.0f;
out[143]=in[133]*-0.7853f+in[135]*0.55223f+in[137]*-0.26675f+in[139]*0.083523f+in[141]*-0.014842f+in[143]*0.00097656f+0.0f;
if(order==12) return;
out[144]=in[157]*0.77144f+in[159]*-0.55833f+in[161]*0.28726f+in[163]*-0.10067f+in[165]*0.021967f+in[167]*-0.0023921f+0.0f;
out[145]=in[145]*-0.010742f+in[147]*0.080713f+in[149]*-0.28768f+in[151]*0.58636f+in[153]*-0.68381f+in[155]*0.31494f+0.0f;
out[146]=in[157]*0.46436f+in[159]*0.20165f+in[161]*-0.65706f+in[163]*0.52111f+in[165]*-0.19834f+in[167]*0.033117f+0.0f;
out[147]=in[145]*0.080713f+in[147]*-0.37207f+in[149]*0.65785f+in[151]*-0.31926f+in[153]*-0.42195f+in[155]*0.37724f+0.0f;
out[148]=in[157]*0.32929f+in[159]*0.45497f+in[161]*-0.23408f+in[163]*-0.54297f+in[165]*0.55492f+in[167]*-0.16439f+0.0f;
out[149]=in[145]*-0.28768f+in[147]*0.65785f+in[149]*-0.15723f+in[151]*-0.52668f+in[153]*-0.11374f+in[155]*0.41161f+0.0f;
out[150]=in[157]*0.2313f+in[159]*0.49004f+in[161]*0.27405f+in[163]*-0.30238f+in[165]*-0.58931f+in[167]*0.4388f+0.0f;
out[151]=in[145]*0.58636f+in[147]*-0.31926f+in[149]*-0.52668f+in[151]*-0.25684f+in[153]*0.15375f+in[155]*0.43273f+0.0f;
out[152]=in[157]*0.14842f+in[159]*0.39258f+in[161]*0.48401f+in[163]*0.34123f+in[165]*-0.074462f+in[167]*-0.68381f+0.0f;
out[153]=in[145]*-0.68381f+in[147]*-0.42195f+in[149]*-0.11374f+in[151]*0.15375f+in[153]*0.3457f+in[155]*0.44527f+0.0f;
out[154]=in[157]*0.072713f+in[159]*0.21529f+in[161]*0.3487f+in[163]*0.46436f+in[165]*0.54719f+in[167]*0.55833f+0.0f;
out[155]=in[145]*0.31494f+in[147]*0.37724f+in[149]*0.41161f+in[151]*0.43273f+in[153]*0.44527f+in[155]*0.45117f+0.0f;
out[156]=in[156]*0.22559f+in[158]*0.32109f+in[160]*0.32771f+in[162]*0.34047f+in[164]*0.36352f+in[166]*0.41011f+in[168]*0.56777f+0.0f;
out[157]=in[144]*-0.77144f+in[146]*-0.46436f+in[148]*-0.32929f+in[150]*-0.2313f+in[152]*-0.14842f+in[154]*-0.072713f+0.0f;
out[158]=in[156]*0.32109f+in[158]*0.43359f+in[160]*0.37077f+in[162]*0.26095f+in[164]*0.092871f+in[166]*-0.16464f+in[168]*-0.68381f+0.0f;
out[159]=in[144]*0.55833f+in[146]*-0.20165f+in[148]*-0.45497f+in[150]*-0.49004f+in[152]*-0.39258f+in[154]*-0.21529f+0.0f;
out[160]=in[156]*0.32771f+in[158]*0.37077f+in[160]*0.1167f+in[162]*-0.22371f+in[164]*-0.51184f+in[166]*-0.50411f+in[168]*0.41875f+0.0f;
out[161]=in[144]*-0.28726f+in[146]*0.65706f+in[148]*0.23408f+in[150]*-0.27405f+in[152]*-0.48401f+in[154]*-0.3487f+0.0f;
out[162]=in[156]*0.34047f+in[158]*0.26095f+in[160]*-0.22371f+in[162]*-0.52539f+in[164]*-0.19811f+in[166]*0.64697f+in[168]*-0.17914f+0.0f;
out[163]=in[144]*0.10067f+in[146]*-0.52111f+in[148]*0.54297f+in[150]*0.30238f+in[152]*-0.34123f+in[154]*-0.46436f+0.0f;
out[164]=in[156]*0.36352f+in[158]*0.092871f+in[160]*-0.51184f+in[162]*-0.19811f+in[164]*0.65723f+in[166]*-0.35145f+in[168]*0.050333f+0.0f;
out[165]=in[144]*-0.021967f+in[146]*0.19834f+in[148]*-0.55492f+in[150]*0.58931f+in[152]*0.074462f+in[154]*-0.54719f+0.0f;
out[166]=in[156]*0.41011f+in[158]*-0.16464f+in[160]*-0.50411f+in[162]*0.64697f+in[164]*-0.35145f+in[166]*0.091797f+in[168]*-0.0081119f+0.0f;
out[167]=in[144]*0.0023921f+in[146]*-0.033117f+in[148]*0.16439f+in[150]*-0.4388f+in[152]*0.68381f+in[154]*-0.55833f+0.0f;
out[168]=in[156]*0.56777f+in[158]*-0.68381f+in[160]*0.41875f+in[162]*-0.17914f+in[164]*0.050333f+in[166]*-0.0081119f+in[168]*0.00048828f+0.0f;
if(order==13) return;
out[169]=in[182]*-0.55674f+in[184]*0.67861f+in[186]*-0.43155f+in[188]*0.19801f+in[190]*-0.062616f+in[192]*0.012449f+in[194]*-0.0012449f+0.0f;
out[170]=in[170]*-0.0058594f+in[172]*0.048828f+in[174]*-0.19648f+in[176]*0.466f+in[178]*-0.67708f+in[180]*0.53235f+0.0f;
out[171]=in[182]*-0.40147f+in[184]*0.18821f+in[186]*0.45483f+in[188]*-0.64803f+in[190]*0.39943f+in[192]*-0.12913f+in[194]*0.01899f+0.0f;
out[172]=in[170]*0.048828f+in[172]*-0.26367f+in[174]*0.58944f+in[176]*-0.52813f+in[178]*-0.13542f+in[180]*0.53235f+0.0f;
out[173]=in[182]*-0.35516f+in[184]*-0.055501f+in[186]*0.51531f+in[188]*0.08745f+in[190]*-0.6299f+in[192]*0.43556f+in[194]*-0.10538f+0.0f;
out[174]=in[170]*-0.19648f+in[172]*0.58944f+in[174]*-0.42187f+in[176]*-0.38293f+in[178]*0.26922f+in[180]*0.46568f+0.0f;
out[175]=in[182]*-0.33185f+in[184]*-0.2178f+in[186]*0.28961f+in[188]*0.49108f+in[190]*0.012789f+in[192]*-0.64327f+in[194]*0.32163f+0.0f;
out[176]=in[170]*0.466f+in[172]*-0.52813f+in[174]*-0.38293f+in[176]*0.13477f+in[178]*0.45121f+in[180]*0.36815f+0.0f;
out[177]=in[182]*-0.31848f+in[184]*-0.32847f+in[186]*-0.014846f+in[188]*0.34013f+in[190]*0.50923f+in[192]*0.23938f+in[194]*-0.59846f+0.0f;
out[178]=in[170]*-0.67708f+in[172]*-0.13542f+in[174]*0.26922f+in[176]*0.45121f+in[178]*0.42773f+in[180]*0.25338f+0.0f;
out[179]=in[182]*-0.3109f+in[184]*-0.39838f+in[186]*-0.27694f+in[188]*-0.083252f+in[190]*0.16673f+in[192]*0.4414f+in[194]*0.6621f+0.0f;
out[180]=in[170]*0.53235f+in[172]*0.53235f+in[174]*0.46568f+in[176]*0.36815f+in[178]*0.25338f+in[180]*0.12891f+0.0f;
out[181]=in[182]*-0.30742f+in[184]*-0.43236f+in[186]*-0.42493f+in[188]*-0.41161f+in[190]*-0.39048f+in[192]*-0.35711f+in[194]*-0.29759f+0.0f;
out[182]=in[169]*0.55674f+in[171]*0.40147f+in[173]*0.35516f+in[175]*0.33185f+in[177]*0.31848f+in[179]*0.3109f+in[181]*0.30742f+0.0f;
out[183]=in[183]*0.032227f+in[185]*0.097772f+in[187]*0.16693f+in[189]*0.24351f+in[191]*0.33508f+in[193]*0.46294f+in[195]*0.75871f+0.0f;
out[184]=in[169]*-0.67861f+in[171]*-0.18821f+in[173]*0.055501f+in[175]*0.2178f+in[177]*0.32847f+in[179]*0.39838f+in[181]*0.43236f+0.0f;
out[185]=in[183]*0.097772f+in[185]*0.27905f+in[187]*0.41641f+in[189]*0.47611f+in[191]*0.41417f+in[193]*0.15606f+in[195]*-0.56268f+0.0f;
out[186]=in[169]*0.43155f+in[171]*-0.45483f+in[173]*-0.51531f+in[175]*-0.28961f+in[177]*0.014846f+in[179]*0.27694f+in[181]*0.42493f+0.0f;
out[187]=in[183]*0.16693f+in[185]*0.41641f+in[187]*0.43823f+in[189]*0.16917f+in[191]*-0.30448f+in[193]*-0.62629f+in[195]*0.30515f+0.0f;
out[188]=in[169]*-0.19801f+in[171]*0.64803f+in[173]*-0.08745f+in[175]*-0.49108f+in[177]*-0.34013f+in[179]*0.083252f+in[181]*0.41161f+0.0f;
out[189]=in[183]*0.24351f+in[185]*0.47611f+in[187]*0.16917f+in[189]*-0.40479f+in[191]*-0.4503f+in[193]*0.55233f+in[195]*-0.11714f+0.0f;
out[190]=in[169]*0.062616f+in[171]*-0.39943f+in[173]*0.6299f+in[175]*-0.012789f+in[177]*-0.50923f+in[179]*-0.16673f+in[181]*0.39048f+0.0f;
out[191]=in[183]*0.33508f+in[185]*0.41417f+in[187]*-0.30448f+in[189]*-0.4503f+in[191]*0.59912f+in[193]*-0.24672f+in[195]*0.029851f+0.0f;
out[192]=in[169]*-0.012449f+in[171]*0.12913f+in[173]*-0.43556f+in[175]*0.64327f+in[177]*-0.23938f+in[179]*-0.4414f+in[181]*0.35711f+0.0f;
out[193]=in[183]*0.46294f+in[185]*0.15606f+in[187]*-0.62629f+in[189]*0.55233f+in[191]*-0.24672f+in[193]*0.055908f+in[195]*-0.0044013f+0.0f;
out[194]=in[169]*0.0012449f+in[171]*-0.01899f+in[173]*0.10538f+in[175]*-0.32163f+in[177]*0.59846f+in[179]*-0.6621f+in[181]*0.29759f+0.0f;
out[195]=in[183]*0.75871f+in[185]*-0.56268f+in[187]*0.30515f+in[189]*-0.11714f+in[191]*0.029851f+in[193]*-0.0044013f+in[195]*0.00024414f+0.0f;

#define LFIX 14
  out += LFIX*LFIX;
	for(int l=LFIX, lindex=shIndex(l,0); l < order; l++, lindex+=2*l) {
		for(int m1=-l; m1 <= l ; m1++, out++) {
			*out = 0;
			const double* elem = _matrix[lindex+m1];
			for(int m2=-l; m2 <= l ; m2++)
				*out += in[lindex+m2] * (float)elem[l+m2]; // GetElement(l, m1, m2);
		}
	}
#undef LFIX
#endif
}

// --------------------------------------------------------------------------
//  CSHRotMatrix::TransformVectorXm90()
// --------------------------------------------------------------------------
void 
CSHRotMatrix::TransformVectorXm90(float *out, const float* in, int order) const
{
  assert(order<=_order);
  if(order<0) order = _order;

#if 0
  // reading matrix elements from 'this'
  for(int l=0, lindex=0; l < order; l++, lindex+=2*l) {
    for(int m1=-l; m1 <= l ; m1++, out++) {
      *out = 0;
      const double* elem = _matrix[lindex+m1];
      for(int m2=-l; m2 <= l ; m2++)
	*out += in[lindex+m2] * (float)elem[l+m2]; // GetElement(l, m1, m2);
    }
  }
#else
if(order==0) return;
  // matrix elements hardcoded in C++
out[0]=in[0]+0.0f;
if(order==1) return;
out[1]=in[2]+0.0f;
out[2]=-in[1]+0.0f;
out[3]=in[3]+0.0f;
if(order==2) return;
out[4]=in[7]+0.0f;
out[5]=in[5]*-1.0f+0.0f;
out[6]=in[6]*-0.5f+in[8]*-0.86603f+0.0f;
out[7]=-in[4]+0.0f;
out[8]=in[6]*-0.86603f+in[8]*0.5f+0.0f;
if(order==3) return;
out[9]=in[12]*-0.79057f+in[14]*0.61237f+0.0f;
out[10]=-in[10]+0.0f;
out[11]=in[12]*-0.61237f+in[14]*-0.79057f+0.0f;
out[12]=in[9]*0.79057f+in[11]*0.61237f+0.0f;
out[13]=in[13]*-0.25f+in[15]*-0.96825f+0.0f;
out[14]=in[9]*-0.61237f+in[11]*0.79057f+0.0f;
out[15]=in[13]*-0.96825f+in[15]*0.25f+0.0f;
if(order==4) return;
out[16]=in[21]*-0.93541f+in[23]*0.35355f+0.0f;
out[17]=in[17]*-0.75f+in[19]*0.66144f+0.0f;
out[18]=in[21]*-0.35355f+in[23]*-0.93541f+0.0f;
out[19]=in[17]*0.66144f+in[19]*0.75f+0.0f;
out[20]=in[20]*0.375f+in[22]*0.55902f+in[24]*0.73951f+0.0f;
out[21]=in[16]*0.93541f+in[18]*0.35355f+0.0f;
out[22]=in[20]*0.55902f+in[22]*0.5f+in[24]*-0.66144f+0.0f;
out[23]=in[16]*-0.35355f+in[18]*0.93541f+0.0f;
out[24]=in[20]*0.73951f+in[22]*-0.66144f+in[24]*0.125f+0.0f;
if(order==5) return;
out[25]=in[30]*0.70156f+in[32]*-0.68465f+in[34]*0.19764f+0.0f;
out[26]=in[26]*-0.5f+in[28]*0.86603f+0.0f;
out[27]=in[30]*0.52291f+in[32]*0.30619f+in[34]*-0.7955f+0.0f;
out[28]=in[26]*0.86603f+in[28]*0.5f+0.0f;
out[29]=in[30]*0.48412f+in[32]*0.66144f+in[34]*0.57282f+0.0f;
out[30]=in[25]*-0.70156f+in[27]*-0.52291f+in[29]*-0.48412f+0.0f;
out[31]=in[31]*0.125f+in[33]*0.40505f+in[35]*0.90571f+0.0f;
out[32]=in[25]*0.68465f+in[27]*-0.30619f+in[29]*-0.66144f+0.0f;
out[33]=in[31]*0.40505f+in[33]*0.8125f+in[35]*-0.41926f+0.0f;
out[34]=in[25]*-0.19764f+in[27]*0.7955f+in[29]*-0.57282f+0.0f;
out[35]=in[31]*0.90571f+in[33]*-0.41926f+in[35]*0.0625f+0.0f;
if(order==6) return;
out[36]=in[43]*0.87945f+in[45]*-0.46351f+in[47]*0.10825f+0.0f;
out[37]=in[37]*-0.3125f+in[39]*0.80283f+in[41]*-0.50775f+0.0f;
out[38]=in[43]*0.43301f+in[45]*0.68465f+in[47]*-0.5863f+0.0f;
out[39]=in[37]*0.80283f+in[39]*-0.0625f+in[41]*-0.59293f+0.0f;
out[40]=in[43]*0.19764f+in[45]*0.5625f+in[47]*0.80283f+0.0f;
out[41]=in[37]*-0.50775f+in[39]*-0.59293f+in[41]*-0.625f+0.0f;
out[42]=in[42]*-0.3125f+in[44]*-0.45286f+in[46]*-0.49608f+in[48]*-0.67169f+0.0f;
out[43]=in[36]*-0.87945f+in[38]*-0.43301f+in[40]*-0.19764f+0.0f;
out[44]=in[42]*-0.45286f+in[44]*-0.53125f+in[46]*-0.17116f+in[48]*0.69527f+0.0f;
out[45]=in[36]*0.46351f+in[38]*-0.68465f+in[40]*-0.5625f+0.0f;
out[46]=in[42]*-0.49608f+in[44]*-0.17116f+in[46]*0.8125f+in[48]*-0.25388f+0.0f;
out[47]=in[36]*-0.10825f+in[38]*0.5863f+in[40]*-0.80283f+0.0f;
out[48]=in[42]*-0.67169f+in[44]*0.69527f+in[46]*-0.25388f+in[48]*0.03125f+0.0f;
if(order==7) return;
out[49]=in[56]*-0.64726f+in[58]*0.69912f+in[60]*-0.29811f+in[62]*0.058463f+0.0f;
out[50]=in[50]*-0.1875f+in[52]*0.63738f+in[54]*-0.74739f+0.0f;
out[51]=in[56]*-0.47496f+in[58]*-0.073288f+in[60]*0.78125f+in[62]*-0.39836f+0.0f;
out[52]=in[50]*0.63738f+in[52]*-0.5f+in[54]*-0.5863f+0.0f;
out[53]=in[56]*-0.42962f+in[58]*-0.41984f+in[60]*0.10364f+in[62]*0.79273f+0.0f;
out[54]=in[50]*-0.74739f+in[52]*-0.5863f+in[54]*-0.3125f+0.0f;
out[55]=in[56]*-0.4134f+in[58]*-0.5741f+in[60]*-0.53855f+in[62]*-0.45768f+0.0f;
out[56]=in[49]*0.64726f+in[51]*0.47496f+in[53]*0.42962f+in[55]*0.4134f+0.0f;
out[57]=in[57]*-0.078125f+in[59]*-0.24357f+in[61]*-0.44879f+in[63]*-0.85624f+0.0f;
out[58]=in[49]*-0.69912f+in[51]*0.073288f+in[53]*0.41984f+in[55]*0.5741f+0.0f;
out[59]=in[57]*-0.24357f+in[59]*-0.60938f+in[61]*-0.57004f+in[63]*0.49435f+0.0f;
out[60]=in[49]*0.29811f+in[51]*-0.78125f+in[53]*-0.10364f+in[55]*0.53855f+0.0f;
out[61]=in[57]*-0.44879f+in[59]*-0.57004f+in[61]*0.67188f+in[63]*-0.14905f+0.0f;
out[62]=in[49]*-0.058463f+in[51]*0.39836f+in[53]*-0.79273f+in[55]*0.45768f+0.0f;
out[63]=in[57]*-0.85624f+in[59]*0.49435f+in[61]*-0.14905f+in[63]*0.015625f+0.0f;
if(order==8) return;
out[64]=in[73]*-0.83561f+in[75]*0.51633f+in[77]*-0.18488f+in[79]*0.03125f+0.0f;
out[65]=in[65]*-0.10937f+in[67]*0.46219f+in[69]*-0.7745f+in[71]*0.4178f+0.0f;
out[66]=in[73]*-0.45768f+in[75]*-0.47135f+in[77]*0.70883f+in[79]*-0.25674f+0.0f;
out[67]=in[65]*0.46219f+in[67]*-0.70312f+in[69]*-0.21819f+in[71]*0.49435f+0.0f;
out[68]=in[73]*-0.27422f+in[75]*-0.60515f+in[77]*-0.33802f+in[79]*0.66659f+0.0f;
out[69]=in[65]*-0.7745f+in[67]*-0.21819f+in[69]*0.26563f+in[71]*0.53102f+0.0f;
out[70]=in[73]*-0.13073f+in[75]*-0.38081f+in[77]*-0.59086f+in[79]*-0.69912f+0.0f;
out[71]=in[65]*0.4178f+in[67]*0.49435f+in[69]*0.53102f+in[71]*0.54688f+0.0f;
out[72]=in[72]*0.27344f+in[74]*0.39218f+in[76]*0.41133f+in[78]*0.45768f+in[80]*0.62671f+0.0f;
out[73]=in[64]*0.83561f+in[66]*0.45768f+in[68]*0.27422f+in[70]*0.13073f+0.0f;
out[74]=in[72]*0.39218f+in[74]*0.5f+in[76]*0.32775f+in[80]*-0.69912f+0.0f;
out[75]=in[64]*-0.51633f+in[66]*0.47135f+in[68]*0.60515f+in[70]*0.38081f+0.0f;
out[76]=in[72]*0.41133f+in[74]*0.32775f+in[76]*-0.28125f+in[78]*-0.73021f+in[80]*0.33329f+0.0f;
out[77]=in[64]*0.18488f+in[66]*-0.70883f+in[68]*0.33802f+in[70]*0.59086f+0.0f;
out[78]=in[72]*0.45768f+in[76]*-0.73021f+in[78]*0.5f+in[80]*-0.085582f+0.0f;
out[79]=in[64]*-0.03125f+in[66]*0.25674f+in[68]*-0.66659f+in[70]*0.69912f+0.0f;
out[80]=in[72]*0.62671f+in[74]*-0.69912f+in[76]*0.33329f+in[78]*-0.085582f+in[80]*0.0078125f+0.0f;
if(order==9) return;
out[81]=in[90]*0.60905f+in[92]*-0.69685f+in[94]*0.36158f+in[96]*-0.11158f+in[98]*0.016573f+0.0f;
out[82]=in[82]*-0.0625f+in[84]*0.31561f+in[86]*-0.68179f+in[88]*0.65699f+0.0f;
out[83]=in[90]*0.44315f+in[92]*-0.056337f+in[94]*-0.67233f+in[96]*0.56833f+in[98]*-0.15944f+0.0f;
out[84]=in[82]*0.31561f+in[84]*-0.71875f+in[86]*0.20252f+in[88]*0.58547f+0.0f;
out[85]=in[90]*0.39636f+in[92]*0.25195f+in[94]*-0.39218f+in[96]*-0.60515f+in[98]*0.50931f+0.0f;
out[86]=in[82]*-0.68179f+in[84]*0.20252f+in[86]*0.5625f+in[88]*0.42159f+0.0f;
out[87]=in[90]*0.37549f+in[92]*0.42962f+in[94]*0.138f+in[96]*-0.29811f+in[98]*-0.75268f+0.0f;
out[88]=in[82]*0.65699f+in[84]*0.58547f+in[86]*0.42159f+in[88]*0.21875f+0.0f;
out[89]=in[90]*0.36685f+in[92]*0.51301f+in[94]*0.49435f+in[96]*0.45768f+in[98]*0.3852f+0.0f;
out[90]=in[81]*-0.60905f+in[83]*-0.44315f+in[85]*-0.39636f+in[87]*-0.37549f+in[89]*-0.36685f+0.0f;
out[91]=in[91]*0.054688f+in[93]*0.16792f+in[95]*0.29543f+in[97]*0.46242f+in[99]*0.81713f+0.0f;
out[92]=in[81]*0.69685f+in[83]*0.056337f+in[85]*-0.25195f+in[87]*-0.42962f+in[89]*-0.51301f+0.0f;
out[93]=in[91]*0.16792f+in[93]*0.45313f+in[95]*0.57728f+in[97]*0.38725f+in[99]*-0.53223f+0.0f;
out[94]=in[81]*-0.36158f+in[83]*0.67233f+in[85]*0.39218f+in[87]*-0.138f+in[89]*-0.49435f+0.0f;
out[95]=in[91]*0.29543f+in[93]*0.57728f+in[95]*0.14063f+in[97]*-0.71624f+in[99]*0.21608f+0.0f;
out[96]=in[81]*0.11158f+in[83]*-0.56833f+in[85]*0.60515f+in[87]*0.29811f+in[89]*-0.45768f+0.0f;
out[97]=in[91]*0.46242f+in[93]*0.38725f+in[95]*-0.71624f+in[97]*0.34766f+in[99]*-0.048318f+0.0f;
out[98]=in[81]*-0.016573f+in[83]*0.15944f+in[85]*-0.50931f+in[87]*0.75268f+in[89]*-0.3852f+0.0f;
out[99]=in[91]*0.81713f+in[93]*-0.53223f+in[95]*0.21608f+in[97]*-0.048318f+in[99]*0.0039063f+0.0f;
if(order==10) return;
out[100]=in[111]*0.80045f+in[113]*-0.5438f+in[115]*0.24319f+in[117]*-0.065945f+in[119]*0.0087346f+0.0f;
out[101]=in[101]*-0.035156f+in[103]*0.20644f+in[105]*-0.5438f+in[107]*0.72958f+in[109]*-0.35797f+0.0f;
out[102]=in[111]*0.46456f+in[113]*0.31561f+in[115]*-0.70572f+in[117]*0.42101f+in[119]*-0.096319f+0.0f;
out[103]=in[101]*0.20644f+in[103]*-0.62891f+in[105]*0.50419f+in[107]*0.35433f+in[109]*-0.42673f+0.0f;
out[104]=in[111]*0.31049f+in[113]*0.53906f+in[115]*-0.017469f+in[117]*-0.69255f+in[119]*0.36479f+0.0f;
out[105]=in[101]*-0.5438f+in[103]*0.50419f+in[105]*0.48437f+in[107]*-0.034939f+in[109]*-0.46285f+0.0f;
out[106]=in[111]*0.19516f+in[113]*0.48614f+in[115]*0.49411f+in[117]*0.091109f+in[119]*-0.68786f+0.0f;
out[107]=in[101]*0.72958f+in[103]*0.35433f+in[105]*-0.034939f+in[107]*-0.32813f+in[109]*-0.48299f+0.0f;
out[108]=in[111]*0.094722f+in[113]*0.27885f+in[115]*0.44538f+in[117]*0.57487f+in[119]*0.62002f+0.0f;
out[109]=in[101]*-0.35797f+in[103]*-0.42673f+in[105]*-0.46285f+in[107]*-0.48299f+in[109]*-0.49219f+0.0f;
out[110]=in[110]*-0.24609f+in[112]*-0.35124f+in[114]*-0.36183f+in[116]*-0.38378f+in[118]*-0.43066f+in[120]*-0.59363f+0.0f;
out[111]=in[100]*-0.80045f+in[102]*-0.46456f+in[104]*-0.31049f+in[106]*-0.19516f+in[108]*-0.094722f+0.0f;
out[112]=in[110]*-0.35124f+in[112]*-0.46484f+in[114]*-0.36619f+in[116]*-0.18922f+in[118]*0.10058f+in[120]*0.69321f+0.0f;
out[113]=in[100]*0.5438f+in[102]*-0.31561f+in[104]*-0.53906f+in[106]*-0.48614f+in[108]*-0.27885f+0.0f;
out[114]=in[110]*-0.36183f+in[112]*-0.36619f+in[114]*0.015625f+in[116]*0.45851f+in[118]*0.61372f+in[120]*-0.38452f+0.0f;
out[115]=in[100]*-0.24319f+in[102]*0.70572f+in[104]*0.017469f+in[106]*-0.49411f+in[108]*-0.44538f+0.0f;
out[116]=in[110]*-0.38378f+in[112]*-0.18922f+in[114]*0.45851f+in[116]*0.46289f+in[118]*-0.61149f+in[120]*0.13595f+0.0f;
out[117]=in[100]*0.065945f+in[102]*-0.42101f+in[104]*0.69255f+in[106]*-0.091109f+in[108]*-0.57487f+0.0f;
out[118]=in[110]*-0.43066f+in[112]*0.10058f+in[114]*0.61372f+in[116]*-0.61149f+in[118]*0.23047f+in[120]*-0.026922f+0.0f;
out[119]=in[100]*-0.0087346f+in[102]*0.096319f+in[104]*-0.36479f+in[106]*0.68786f+in[108]*-0.62002f+0.0f;
out[120]=in[110]*-0.59363f+in[112]*0.69321f+in[114]*-0.38452f+in[116]*0.13595f+in[118]*-0.026922f+in[120]*0.0019531f+0.0f;
if(order==11) return;
out[121]=in[132]*-0.57998f+in[134]*0.68875f+in[136]*-0.40329f+in[138]*0.15847f+in[140]*-0.038323f+in[142]*0.0045805f+0.0f;
out[122]=in[122]*-0.019531f+in[124]*0.13073f+in[126]*-0.40544f+in[128]*0.68786f+in[130]*-0.58737f+0.0f;
out[123]=in[132]*-0.41976f+in[134]*0.13595f+in[136]*0.55723f+in[138]*-0.63604f+in[140]*0.29501f+in[142]*-0.05696f+0.0f;
out[124]=in[122]*0.13073f+in[124]*-0.5f+in[126]*0.64613f+in[130]*-0.56163f+0.0f;
out[125]=in[132]*-0.37297f+in[134]*-0.1369f+in[136]*0.49511f+in[138]*0.32426f+in[140]*-0.65554f+in[142]*0.2493f+0.0f;
out[126]=in[122]*-0.40544f+in[124]*0.64613f+in[126]*0.18359f+in[128]*-0.41751f+in[130]*-0.45838f+0.0f;
out[127]=in[132]*-0.35034f+in[134]*-0.31014f+in[136]*0.13435f+in[138]*0.52273f+in[140]*0.40785f+in[142]*-0.56872f+0.0f;
out[128]=in[122]*0.68786f+in[126]*-0.41751f+in[128]*-0.5f+in[130]*-0.32022f+0.0f;
out[129]=in[132]*-0.33846f+in[134]*-0.41655f+in[136]*-0.23535f+in[138]*0.047642f+in[140]*0.39402f+in[142]*0.70641f+0.0f;
out[130]=in[122]*-0.58737f+in[124]*-0.56163f+in[126]*-0.45838f+in[128]*-0.32022f+in[130]*-0.16406f+0.0f;
out[131]=in[132]*-0.33321f+in[134]*-0.46765f+in[136]*-0.45638f+in[138]*-0.43553f+in[140]*-0.40022f+in[142]*-0.33485f+0.0f;
out[132]=in[121]*0.57998f+in[123]*0.41976f+in[125]*0.37297f+in[127]*0.35034f+in[129]*0.33846f+in[131]*0.33321f+0.0f;
out[133]=in[133]*-0.041016f+in[135]*-0.12498f+in[137]*-0.21562f+in[139]*-0.32136f+in[141]*-0.46502f+in[143]*-0.7853f+0.0f;
out[134]=in[121]*-0.68875f+in[123]*-0.13595f+in[125]*0.1369f+in[127]*0.31014f+in[129]*0.41655f+in[131]*0.46765f+0.0f;
out[135]=in[133]*-0.12498f+in[135]*-0.34961f+in[137]*-0.49531f+in[139]*-0.49717f+in[141]*-0.25434f+in[143]*0.55223f+0.0f;
out[136]=in[121]*0.40329f+in[123]*-0.55723f+in[125]*-0.49511f+in[127]*-0.13435f+in[129]*0.23535f+in[131]*0.45638f+0.0f;
out[137]=in[133]*-0.21562f+in[135]*-0.49531f+in[137]*-0.38574f+in[139]*0.14035f+in[141]*0.68449f+in[143]*-0.26675f+0.0f;
out[138]=in[121]*-0.15847f+in[123]*0.63604f+in[125]*-0.32426f+in[127]*-0.52273f+in[129]*-0.047642f+in[131]*0.43553f+0.0f;
out[139]=in[133]*-0.32136f+in[135]*-0.49717f+in[137]*0.14035f+in[139]*0.62793f+in[141]*-0.4781f+in[143]*0.083523f+0.0f;
out[140]=in[121]*0.038323f+in[123]*-0.29501f+in[125]*0.65554f+in[127]*-0.40785f+in[129]*-0.39402f+in[131]*0.40022f+0.0f;
out[141]=in[133]*-0.46502f+in[135]*-0.25434f+in[137]*0.68449f+in[139]*-0.4781f+in[141]*0.14746f+in[143]*-0.014842f+0.0f;
out[142]=in[121]*-0.0045805f+in[123]*0.05696f+in[125]*-0.2493f+in[127]*0.56872f+in[129]*-0.70641f+in[131]*0.33485f+0.0f;
out[143]=in[133]*-0.7853f+in[135]*0.55223f+in[137]*-0.26675f+in[139]*0.083523f+in[141]*-0.014842f+in[143]*0.00097656f+0.0f;
if(order==12) return;
out[144]=in[157]*-0.77144f+in[159]*0.55833f+in[161]*-0.28726f+in[163]*0.10067f+in[165]*-0.021967f+in[167]*0.0023921f+0.0f;
out[145]=in[145]*-0.010742f+in[147]*0.080713f+in[149]*-0.28768f+in[151]*0.58636f+in[153]*-0.68381f+in[155]*0.31494f+0.0f;
out[146]=in[157]*-0.46436f+in[159]*-0.20165f+in[161]*0.65706f+in[163]*-0.52111f+in[165]*0.19834f+in[167]*-0.033117f+0.0f;
out[147]=in[145]*0.080713f+in[147]*-0.37207f+in[149]*0.65785f+in[151]*-0.31926f+in[153]*-0.42195f+in[155]*0.37724f+0.0f;
out[148]=in[157]*-0.32929f+in[159]*-0.45497f+in[161]*0.23408f+in[163]*0.54297f+in[165]*-0.55492f+in[167]*0.16439f+0.0f;
out[149]=in[145]*-0.28768f+in[147]*0.65785f+in[149]*-0.15723f+in[151]*-0.52668f+in[153]*-0.11374f+in[155]*0.41161f+0.0f;
out[150]=in[157]*-0.2313f+in[159]*-0.49004f+in[161]*-0.27405f+in[163]*0.30238f+in[165]*0.58931f+in[167]*-0.4388f+0.0f;
out[151]=in[145]*0.58636f+in[147]*-0.31926f+in[149]*-0.52668f+in[151]*-0.25684f+in[153]*0.15375f+in[155]*0.43273f+0.0f;
out[152]=in[157]*-0.14842f+in[159]*-0.39258f+in[161]*-0.48401f+in[163]*-0.34123f+in[165]*0.074462f+in[167]*0.68381f+0.0f;
out[153]=in[145]*-0.68381f+in[147]*-0.42195f+in[149]*-0.11374f+in[151]*0.15375f+in[153]*0.3457f+in[155]*0.44527f+0.0f;
out[154]=in[157]*-0.072713f+in[159]*-0.21529f+in[161]*-0.3487f+in[163]*-0.46436f+in[165]*-0.54719f+in[167]*-0.55833f+0.0f;
out[155]=in[145]*0.31494f+in[147]*0.37724f+in[149]*0.41161f+in[151]*0.43273f+in[153]*0.44527f+in[155]*0.45117f+0.0f;
out[156]=in[156]*0.22559f+in[158]*0.32109f+in[160]*0.32771f+in[162]*0.34047f+in[164]*0.36352f+in[166]*0.41011f+in[168]*0.56777f+0.0f;
out[157]=in[144]*0.77144f+in[146]*0.46436f+in[148]*0.32929f+in[150]*0.2313f+in[152]*0.14842f+in[154]*0.072713f+0.0f;
out[158]=in[156]*0.32109f+in[158]*0.43359f+in[160]*0.37077f+in[162]*0.26095f+in[164]*0.092871f+in[166]*-0.16464f+in[168]*-0.68381f+0.0f;
out[159]=in[144]*-0.55833f+in[146]*0.20165f+in[148]*0.45497f+in[150]*0.49004f+in[152]*0.39258f+in[154]*0.21529f+0.0f;
out[160]=in[156]*0.32771f+in[158]*0.37077f+in[160]*0.1167f+in[162]*-0.22371f+in[164]*-0.51184f+in[166]*-0.50411f+in[168]*0.41875f+0.0f;
out[161]=in[144]*0.28726f+in[146]*-0.65706f+in[148]*-0.23408f+in[150]*0.27405f+in[152]*0.48401f+in[154]*0.3487f+0.0f;
out[162]=in[156]*0.34047f+in[158]*0.26095f+in[160]*-0.22371f+in[162]*-0.52539f+in[164]*-0.19811f+in[166]*0.64697f+in[168]*-0.17914f+0.0f;
out[163]=in[144]*-0.10067f+in[146]*0.52111f+in[148]*-0.54297f+in[150]*-0.30238f+in[152]*0.34123f+in[154]*0.46436f+0.0f;
out[164]=in[156]*0.36352f+in[158]*0.092871f+in[160]*-0.51184f+in[162]*-0.19811f+in[164]*0.65723f+in[166]*-0.35145f+in[168]*0.050333f+0.0f;
out[165]=in[144]*0.021967f+in[146]*-0.19834f+in[148]*0.55492f+in[150]*-0.58931f+in[152]*-0.074462f+in[154]*0.54719f+0.0f;
out[166]=in[156]*0.41011f+in[158]*-0.16464f+in[160]*-0.50411f+in[162]*0.64697f+in[164]*-0.35145f+in[166]*0.091797f+in[168]*-0.0081119f+0.0f;
out[167]=in[144]*-0.0023921f+in[146]*0.033117f+in[148]*-0.16439f+in[150]*0.4388f+in[152]*-0.68381f+in[154]*0.55833f+0.0f;
out[168]=in[156]*0.56777f+in[158]*-0.68381f+in[160]*0.41875f+in[162]*-0.17914f+in[164]*0.050333f+in[166]*-0.0081119f+in[168]*0.00048828f+0.0f;
if(order==13) return;
out[169]=in[182]*0.55674f+in[184]*-0.67861f+in[186]*0.43155f+in[188]*-0.19801f+in[190]*0.062616f+in[192]*-0.012449f+in[194]*0.0012449f+0.0f;
out[170]=in[170]*-0.0058594f+in[172]*0.048828f+in[174]*-0.19648f+in[176]*0.466f+in[178]*-0.67708f+in[180]*0.53235f+0.0f;
out[171]=in[182]*0.40147f+in[184]*-0.18821f+in[186]*-0.45483f+in[188]*0.64803f+in[190]*-0.39943f+in[192]*0.12913f+in[194]*-0.01899f+0.0f;
out[172]=in[170]*0.048828f+in[172]*-0.26367f+in[174]*0.58944f+in[176]*-0.52813f+in[178]*-0.13542f+in[180]*0.53235f+0.0f;
out[173]=in[182]*0.35516f+in[184]*0.055501f+in[186]*-0.51531f+in[188]*-0.08745f+in[190]*0.6299f+in[192]*-0.43556f+in[194]*0.10538f+0.0f;
out[174]=in[170]*-0.19648f+in[172]*0.58944f+in[174]*-0.42187f+in[176]*-0.38293f+in[178]*0.26922f+in[180]*0.46568f+0.0f;
out[175]=in[182]*0.33185f+in[184]*0.2178f+in[186]*-0.28961f+in[188]*-0.49108f+in[190]*-0.012789f+in[192]*0.64327f+in[194]*-0.32163f+0.0f;
out[176]=in[170]*0.466f+in[172]*-0.52813f+in[174]*-0.38293f+in[176]*0.13477f+in[178]*0.45121f+in[180]*0.36815f+0.0f;
out[177]=in[182]*0.31848f+in[184]*0.32847f+in[186]*0.014846f+in[188]*-0.34013f+in[190]*-0.50923f+in[192]*-0.23938f+in[194]*0.59846f+0.0f;
out[178]=in[170]*-0.67708f+in[172]*-0.13542f+in[174]*0.26922f+in[176]*0.45121f+in[178]*0.42773f+in[180]*0.25338f+0.0f;
out[179]=in[182]*0.3109f+in[184]*0.39838f+in[186]*0.27694f+in[188]*0.083252f+in[190]*-0.16673f+in[192]*-0.4414f+in[194]*-0.6621f+0.0f;
out[180]=in[170]*0.53235f+in[172]*0.53235f+in[174]*0.46568f+in[176]*0.36815f+in[178]*0.25338f+in[180]*0.12891f+0.0f;
out[181]=in[182]*0.30742f+in[184]*0.43236f+in[186]*0.42493f+in[188]*0.41161f+in[190]*0.39048f+in[192]*0.35711f+in[194]*0.29759f+0.0f;
out[182]=in[169]*-0.55674f+in[171]*-0.40147f+in[173]*-0.35516f+in[175]*-0.33185f+in[177]*-0.31848f+in[179]*-0.3109f+in[181]*-0.30742f+0.0f;
out[183]=in[183]*0.032227f+in[185]*0.097772f+in[187]*0.16693f+in[189]*0.24351f+in[191]*0.33508f+in[193]*0.46294f+in[195]*0.75871f+0.0f;
out[184]=in[169]*0.67861f+in[171]*0.18821f+in[173]*-0.055501f+in[175]*-0.2178f+in[177]*-0.32847f+in[179]*-0.39838f+in[181]*-0.43236f+0.0f;
out[185]=in[183]*0.097772f+in[185]*0.27905f+in[187]*0.41641f+in[189]*0.47611f+in[191]*0.41417f+in[193]*0.15606f+in[195]*-0.56268f+0.0f;
out[186]=in[169]*-0.43155f+in[171]*0.45483f+in[173]*0.51531f+in[175]*0.28961f+in[177]*-0.014846f+in[179]*-0.27694f+in[181]*-0.42493f+0.0f;
out[187]=in[183]*0.16693f+in[185]*0.41641f+in[187]*0.43823f+in[189]*0.16917f+in[191]*-0.30448f+in[193]*-0.62629f+in[195]*0.30515f+0.0f;
out[188]=in[169]*0.19801f+in[171]*-0.64803f+in[173]*0.08745f+in[175]*0.49108f+in[177]*0.34013f+in[179]*-0.083252f+in[181]*-0.41161f+0.0f;
out[189]=in[183]*0.24351f+in[185]*0.47611f+in[187]*0.16917f+in[189]*-0.40479f+in[191]*-0.4503f+in[193]*0.55233f+in[195]*-0.11714f+0.0f;
out[190]=in[169]*-0.062616f+in[171]*0.39943f+in[173]*-0.6299f+in[175]*0.012789f+in[177]*0.50923f+in[179]*0.16673f+in[181]*-0.39048f+0.0f;
out[191]=in[183]*0.33508f+in[185]*0.41417f+in[187]*-0.30448f+in[189]*-0.4503f+in[191]*0.59912f+in[193]*-0.24672f+in[195]*0.029851f+0.0f;
out[192]=in[169]*0.012449f+in[171]*-0.12913f+in[173]*0.43556f+in[175]*-0.64327f+in[177]*0.23938f+in[179]*0.4414f+in[181]*-0.35711f+0.0f;
out[193]=in[183]*0.46294f+in[185]*0.15606f+in[187]*-0.62629f+in[189]*0.55233f+in[191]*-0.24672f+in[193]*0.055908f+in[195]*-0.0044013f+0.0f;
out[194]=in[169]*-0.0012449f+in[171]*0.01899f+in[173]*-0.10538f+in[175]*0.32163f+in[177]*-0.59846f+in[179]*0.6621f+in[181]*-0.29759f+0.0f;
out[195]=in[183]*0.75871f+in[185]*-0.56268f+in[187]*0.30515f+in[189]*-0.11714f+in[191]*0.029851f+in[193]*-0.0044013f+in[195]*0.00024414f+0.0f;

#define LFIX 14
  out += LFIX*LFIX;
  for(int l=LFIX, lindex=shIndex(l,0); l < order; l++, lindex+=2*l) {
    for(int m1=-l; m1 <= l ; m1++, out++) {
      *out = 0;
      const double* elem = _matrix[lindex+m1];
      for(int m2=-l; m2 <= l ; m2++)
	*out += in[lindex+m2] * (float)elem[l+m2]; // GetElement(l, m1, m2);
    }
  }
#undef LFIX
#endif
}


// --------------------------------------------------------------------------
//  CSHRotMatrix::TransformVectorYdY()
// --------------------------------------------------------------------------
#if 0
// nice but slow version
void 
CSHRotMatrix::TransformVectorYdY(float       *out, 
                                 const float *in,
                                 const float angle,
                                 int   order) const
{
  if (order < 0)
    order = _order;
  else
    if (order>_order)
      order = _order;

  for(int l = 0; l < order; l++) {
    for(int m = -l; m <= l; m++) {
      out[shIndex(l, m)] = in[shIndex(l,m)];
      if(m > -l)
	out[shIndex(l, m)] += 
	  GetElement(l, m, m-1) * in[shIndex(l, m-1)] * angle;
      if (m < l)
	out[shIndex(l, m)] += 
	  GetElement(l, m, m+1) * in[shIndex(l, m+1)] * angle;
    }
  }
}
#else
// hacked but fast version
void 
CSHRotMatrix::TransformVectorYdY(float* out, 
                                 const float *in,
                                 const float *ddyDiag,
                                 const float angle,
                                 int   order) const
{
  if (order < 0)
    order = _order;
  else
    if (order > _order)
      order=_order;
  
  if (order < 1)
    return;

  out[0] = in[0];

  float aangle = 0.5f*shm::sqr2(angle);

  double **mat = _matrix+1;
  out++;in++;ddyDiag++;
  for(int l=1, index=1; l<order; l++) {
    int m=-l;
    out[0] = in[0] + (mat[0][l+m+1]*in[+1]*angle) + (in[0]*ddyDiag[0]*aangle);
    index++;mat++;out++;in++;ddyDiag++;
    // m = -l+1 .. l-1
    for(m=-l+1; m<l; m++) {
      out[0] = in[0] + (in[1]*mat[0][l+m+1] + in[-1]*mat[0][l+m-1])*angle +
        (in[0]*ddyDiag[0]*aangle);
      index++;mat++;out++;in++;ddyDiag++;
    }
    // m = l
    out[0] = in[0] + (mat[0][l+m-1]*in[-1]*angle) + (in[0]*ddyDiag[0]*aangle);
    index++;mat++;out++;in++;ddyDiag++;
  }
}
#endif


// --------------------------------------------------------------------------
//   CSHRotMatrix::mul()
// --------------------------------------------------------------------------
const CSHRotMatrix& 
CSHRotMatrix::mul(const double a, int order)
{
  if (order<0)
    order = _order;
  else
    if (order>_order)
      order=_order;

  for(int l=0; l<order; l++)
    for(int m1=-l; m1 <= l ; m1++)
      for(int m2=-l; m2 <= l ; m2++)
	_matrix[shIndex(l,m1)][l+m2] *= a;
  
  return *this;
}

// --------------------------------------------------------------------------
//   CSHRotMatrix::mul()
// --------------------------------------------------------------------------
const CSHRotMatrix& 
CSHRotMatrix::mul(const CSHRotMatrix& m, int order)
{
  if (order<0)
    order = _order;
  else
    if (order>_order)
      order = _order;

  if (order > m.GetOrder())
    order = m.GetOrder();

  for(int l=0; l<order; l++)
    for(int m1=-l; m1 <= l ; m1++)
      for(int m2=-l; m2 <= l ; m2++)
	_matrix[shIndex(l,m1)][l+m2] *= m(l,m1,m2);
  
  return *this;
}

// --------------------------------------------------------------------------
//   CSHRotMatrix::sub()
// --------------------------------------------------------------------------
const CSHRotMatrix& 
CSHRotMatrix::sub(const CSHRotMatrix& m, int order)
{
  if (order < 0)
    order = _order;
  else
    if (order > _order)
      order=_order;

  if(order>m.GetOrder()) order=m.GetOrder();

  for(int l=0; l<order; l++)
    for(int m1=-l; m1 <= l ; m1++)
      for(int m2=-l; m2 <= l ; m2++)
	_matrix[shIndex(l,m1)][l+m2] -= m(l,m1,m2);
  
  return *this;
}

// --------------------------------------------------------------------------
//   CSHRotMatrix::add()
// --------------------------------------------------------------------------
const CSHRotMatrix& 
CSHRotMatrix::add(const CSHRotMatrix& m, int order)
{
  if (order < 0)
    order = _order;
  else
    if (order > _order)
      order = _order;

  if (order>m.GetOrder())
    order=m.GetOrder();

  for(int l=0; l<order; l++)
    for(int m1=-l; m1 <= l ; m1++)
      for(int m2=-l; m2 <= l ; m2++)
	_matrix[shIndex(l,m1)][l+m2] += m(l,m1,m2);
  
  return *this;
}

// --------------------------------------------------------------------------
//  EpsilonEqual()
// --------------------------------------------------------------------------
bool
EpsilonEqual(const CSHRotMatrix& mat1, 
	     const CSHRotMatrix& mat2, 
	     double eps, 
	     int order)
{
  if (order < 0)
    order = mat1.GetOrder();
  else
    if (order > mat1.GetOrder())
      order = mat1.GetOrder();

  if (order>mat2.GetOrder())
    return false;

  for(int l=0; l<order; l++) {
    for(int m1=-l; m1 <= l ; m1++) {
      for(int m2=-l; m2 <= l ; m2++) {
	double diff = shm::fabs( mat1(l,m1,m2)-mat2(l,m1,m2) );
	if(diff>eps) { PRINTVAR(diff); return false; }
	
	// if( gmath::fabs( mat1(l,m1,m2) - mat2(l,m1,m2) ) > eps ) return false;
      }
    }
  }
  
  // PRINTVAR(maxdiff);
  return true;
}


// -----------------------------------------------------------------------
//   CSHRotMatrix::assign()
// -----------------------------------------------------------------------
const CSHRotMatrix& 
CSHRotMatrix::assign(const CSHRotMatrix& m, int order)
{
  if (order < 0)
    order = _order;
  else
    if (order > _order)
      order=_order;

  if (order > m.GetOrder())
    order = m.GetOrder();

  for(int l=0; l<order; l++)
    for(int m1=-l; m1 <= l ; m1++)
      for(int m2=-l; m2 <= l ; m2++)
	_matrix[shIndex(l,m1)][l+m2] = m(l,m1,m2);
  
  return *this;
}

// -----------------------------------------------------------------------
//   CSHRotMatrix::setz()
// -----------------------------------------------------------------------
const CSHRotMatrix& 
CSHRotMatrix::setz(int order)
{
  if (order < 0)
    order = _order;
  else
    if (order > _order)
      order=_order;

  for(int l=0; l<order; l++)
    for(int m1=-l; m1 <= l ; m1++)
      for(int m2=-l; m2 <= l ; m2++)
	_matrix[shIndex(l,m1)][l+m2] = 0;
  
  return *this;
}

// --------------------------------------------------------------------------
//   CSHRotMatrix::Describe()
// --------------------------------------------------------------------------
void
CSHRotMatrix::Describe(ostream& o, int order) const
{
  if (order < 0)
    order = _order;
  else
    if (order > _order)
      order=_order;

  for(int l=0; l<order; l++) {
    for(int m1=-l; m1 <= l ; m1++) {
      for(int m2=-l; m2 <= l ; m2++) {
	o 
	  << std::setw(5) << std::setprecision(2) 
	  << ((fabsf(_matrix[shIndex(l,m1)][l+m2])>1e-3) ?
	      _matrix[shIndex(l,m1)][l+m2]:0.0) << "";
        // if(m2<l) o << " & ";
      }
      o << "\\\\ \n";
    }
    o << std::endl;
  }

  return;
}

// --------------------------------------------------------------------------
//   CSHRotMatrix::Describe()
// --------------------------------------------------------------------------
void
CSHRotMatrix::SetIdentity(CSHRotMatrix& m, int order)
{
  if (order < 0)
    order = m.GetOrder();
  else
    if(order > m.GetOrder())
      order = m.GetOrder();

  for(int l=0; l<order; l++)
    for(int m1=-l; m1 <= l ; m1++)
      for(int m2=-l; m2 <= l ; m2++)
	m(l,m1,m2) = ( m1==m2 ? 1 : 0 );

  return;
}

__END_SPHAR_SOURCE

#endif
