// ===================================================================
// $Id: shroty.cpp,v 1.10 2005/11/29 15:51:51 xkrivanj Exp $
/// \file shroty.cpp
///
/// \brief Spherical harmonic rotations (rotations around Y axis).
///
/// SPHAR_REPLACEMENT_STRING
///
/// Initial coding by Jaroslav Krivanek, January 2004.

// in case the source is included somewhere
#ifndef __SHROTY_CPP__
#define __SHROTY_CPP__

// SPHAR headers
#include "shconfig.h"
#include "shrot.h"
#include "shrotmat.h"

__BEGIN_SPHAR_SOURCE


// -----------------------------------------------------------------------------
//   cos_diff()
// -----------------------------------------------------------------------------
/// Find the n-th derivative of the 'cos' function.
template<class TReal>
inline TReal cos_diff(TReal cos_beta, TReal sin_beta, int nth)
{
  assert(nth>=0);
  /* 
  0 th ...  cos(beta)
  1 st ... -sin(beta)
  2 nd ... -cos(beta)
  3 rd ...  sin(beta)
  4 th ...  cos(beta)
  ...
  */
  switch(nth%4) {
    case 0: return  cos_beta;
    case 1: return -sin_beta;
    case 2: return -cos_beta;
    case 3: return  sin_beta;
    default:
      FATAL << "Should not happen." << std::endl;
      FATAL_EXIT;
  }
}

// -----------------------------------------------------------------------------
//   sin_diff()
// -----------------------------------------------------------------------------
/// Find the n-th derivative of the 'cos' function.
template<class TReal>
inline TReal sin_diff(TReal cos_beta, TReal sin_beta, int nth)
{
  assert(nth>=0);
  /* 
  0 th ...  sin(beta)
  1 st ...  cos(beta)
  2 nd ... -sin(beta)
  3 rd ... -cos(beta)
  4 th ...  sin(beta)
  ...
  */
  return cos_diff(cos_beta,sin_beta,nth+3);
}


// -----------------------------------------------------------------------------
//   CSHRotGenerator::BuildMatrix_dY()
// -----------------------------------------------------------------------------
void CSHRotGenerator::BuildMatrix_dY(CSHRotMatrix **dR, 
                                     int           nth,
                                     int           order,
                                     double        cos_beta, 
                                     double        sin_beta)
{
  assert(nth>=0);
  assert(nth<=MAX_DERIVATIVE);

  // k-th derivative matrix
  int k;

  // check order
  if(order<1) return;
  if(order > _order) _setNewOrder(order);

  // copy matrices to local array & check sufficient order
  for(k=0; k<=nth; k++) {
    assert(dR[k]->GetOrder() >= order);
    _dR[k] = dR[k];
  }

  // set zeroth order
  dR[0]->SetElement(0,0,0,1);
  for(k=1; k<=nth; k++)
    dR[k]->SetElement(0,0,0,0);

  if (order==1) return;

  // set first order
  for(k=0; k<=nth; k++) {
    // compute the k-th derivative of cos and sin
    double dcos_dbeta = cos_diff(cos_beta,sin_beta,k);
    double dsin_dbeta = sin_diff(cos_beta,sin_beta,k);

    // set matrix elements
    dR[k]->SetElement(1,  -1,-1, (k==0)?1:0 );
    dR[k]->SetElement(1,  -1, 0,    0 );
    dR[k]->SetElement(1,  -1, 1,    0 );

    dR[k]->SetElement(1,   0,-1,    0 );
    dR[k]->SetElement(1,   0, 0,   dcos_dbeta );
    dR[k]->SetElement(1,   0, 1,  -dsin_dbeta );

    dR[k]->SetElement(1,   1,-1,   0  );
    dR[k]->SetElement(1,   1, 0,   dsin_dbeta );
    dR[k]->SetElement(1,   1, 1,   dcos_dbeta );
  }

  if (order == 2) return;

  // set higher orders
  for(int l=2; l <order; l++) {
    for(k=0; k<=nth; k++) {
      for(int m1=-l; m1 <=l; m1++) {
        double* mat = dR[k]->_matrix[shIndex(l,m1)];
        for(int m2=-l; m2 <= l; m2++) {
          const SPreRec& little = _getGenericValue(l, m1, m2);
          mat[l+m2] = 0;
          if(little.nz & 1) mat[l+m2] += little.u * dUY(l, m1, m2, k);
          if(little.nz & 2) mat[l+m2] += little.v * dVY(l, m1, m2, k);				
          if(little.nz & 4) mat[l+m2] += little.w * dWY(l, m1, m2, k);
        }
      }
    }
  }
}

// -----------------------------------------------------------------------------
//   CSHRotGenerator::dTerm()
// -----------------------------------------------------------------------------
inline double CSHRotGenerator::dTerm(int fl, int fm1, int fm2,
                                     int gl, int gm1, int gm2,
                                     int nth)
{
  // implements
  // \sum_{i=0}^n \binom{n}{i} \frac{d^i f}{dx^i} \frac{d^{n-i} g}{dx^{n-i}}

  double sum = 0;
  for(int i=0; i<=nth; i++)
    sum += 
    shm::binom(nth,i) * (*_dR[i])(fl,fm1,fm2) * (*_dR[nth-i])(gl,gm1,gm2);

  return sum;
}

// -----------------------------------------------------------------------------
//   CSHRotGenerator::dPY()
// -----------------------------------------------------------------------------
double CSHRotGenerator::dPY(int l, int m1, int m2, int index, int nth)
{
  ///////////////////////////////////////////////////////////////////////////
  //  (A . B)' => A'.B + A.B'  // sin'(x) => cos(x)  // cos'(x) => -sin(x) //
  ///////////////////////////////////////////////////////////////////////////

  if (abs(m2) < l)
  {
    return dTerm(1, index, 0,    l-1, m1, m2, nth);
  }
  if (m2 == l)
  {
    return
      dTerm(1, index,  1,    l-1, m1, l-1,   nth) -
      dTerm(1, index, -1,    l-1, m1, -l+1,  nth);
  }
  if (m2 == -l)
  {
    return 
      dTerm(1, index, 1,     l-1, m1, -l+1,  nth) +
      dTerm(1, index, -1,    l-1, m1, l-1,   nth);
  }
  return 0;
}

// -----------------------------------------------------------------------------
//   CSHRotGenerator::dWY()
// -----------------------------------------------------------------------------
inline double CSHRotGenerator::dWY(int l, int m1, int m2, int nth)
{
  if (m1 > 0)
    return dPY(l, m1+1, m2, 1, nth) + dPY(l, -m1-1, m2, -1, nth);
  else
    return dPY(l, m1-1, m2, 1, nth) - dPY(l, -m1+1, m2, -1, nth);
}

// -----------------------------------------------------------------------------
//   CSHRotGenerator::dUY()
// -----------------------------------------------------------------------------
inline double CSHRotGenerator::dUY(int l, int m1, int m2, int nth)
{
  return  dPY(l, m1, m2, 0, nth);
}

// -----------------------------------------------------------------------------
//   CSHRotGenerator::dVY()
// -----------------------------------------------------------------------------
inline double CSHRotGenerator::dVY(int l, int m1, int m2, int nth)
{
  if(m1>=0) {
    if(m1==0)      return dPY(l, 1, m2, 1, nth)+dPY(l, -1, m2, -1, nth);
    else if(m1==1) return M_SQRT2 * dPY(l, 0, m2, 1, nth);
    else           return dPY(l, m1-1, m2, 1, nth) - dPY(l, -m1+1, m2, -1, nth);
  }
  else {
    if(m1 == -1)  return M_SQRT2 * dPY(l, 0, m2, -1, nth);
    else          return dPY(l, -m1-1, m2, -1, nth) + dPY(l, m1+1, m2, 1, nth);		
  }
}

// -----------------------------------------------------------------------------
//   CSHRotGenerator::BuildMatrix_dY0()
// -----------------------------------------------------------------------------
void CSHRotGenerator::BuildMatrix_dY0(CSHRotMatrix &dR, int order)
{
  CSHRotMatrix R(order);
  CSHRotMatrix *dRarray[] = {&R,&dR};
  BuildMatrix_dY(dRarray,1,order,cos(0.0),sin(0.0));
}

// -----------------------------------------------------------------------------
//   CSHRotGenerator::BuildMatrixNum_ddY0()
// -----------------------------------------------------------------------------
void CSHRotGenerator::BuildMatrixNum_ddY0(CSHRotMatrix &ddR, int order)
{
  double h=0.00001;
  CSHRotMatrix fx_min_h(order), fx_plus_h(order);

  rotGenerator.BuildMatrixY(ddR,        order,  cos(0.0), sin(0.0));
  rotGenerator.BuildMatrixY(fx_min_h ,  order,  cos(-h),  sin(-h));
  rotGenerator.BuildMatrixY(fx_plus_h,  order,  cos(h),   sin(h));

  ddR.mul(-2);
  ddR.add(fx_min_h);
  ddR.add(fx_plus_h);
  ddR.mul(1.0/(h*h));
}

// -----------------------------------------------------------------------------
//   CSHRotGenerator::BuildMatrixNum_dddY0()
// -----------------------------------------------------------------------------
void CSHRotGenerator::BuildMatrixNum_dddY0(CSHRotMatrix &dddR, int order)
{
  // based on second-order center-difference formula:
  // f'''(t) = [f(t + 2h) - 2f(t + h) + 2f(t - h) - f(t - 2h)] )] / 2h^3

  double h=0.00001;
  CSHRotMatrix tmp(order);

  rotGenerator.BuildMatrixY(dddR, order, cos(2*h),  sin(2*h));
  rotGenerator.BuildMatrixY(tmp,  order, cos(h),  sin(h));
  tmp.mul(-2);
  dddR.add(tmp);
  rotGenerator.BuildMatrixY(tmp,  order, cos(-h),  sin(-h));
  tmp.mul(2);
  dddR.add(tmp);
  rotGenerator.BuildMatrixY(tmp,  order, cos(-2*h),  sin(-2*h));
  tmp.mul(-1);
  dddR.add(tmp);

  dddR.mul(0.5/(h*h*h));
}

// --------------------------------------------------------------------------
//   CSHRotGenerator::P()
// --------------------------------------------------------------------------
double CSHRotGenerator::P(int l, int m1, int m2, int index)
{
#define R (*_dR[0])
  if (abs(m2) < l)
  {
    return R(1, index, 0) * R(l-1, m1, m2);
  }
  if (m2 == l)
  {
    return R(1, index,  1) * R(l-1, m1, l-1) -
      R(1, index, -1) * R(l-1, m1, -l+1);
  }
  if (m2 == -l)
  {
    return R(1, index, 1) * R(l-1, m1, -l+1)+
      R(1, index, -1) * R(l-1, m1, l-1);
  }
  return 0;
#undef R
}

__END_SPHAR_SOURCE

#endif
