// ===================================================================
// $Id: shtab.cpp,v 1.8 2005/11/29 15:51:51 xkrivanj Exp $
/// \file shtab.cpp
/// 
/// \brief Storage for tabulated (hemi)spherical basis.
///
/// SPHAR_REPLACEMENT_STRING
///
/// Initial coding by Jaroslav Krivanek, June 2004.

// in case the source is included somewhere
#ifndef __SHTAB_CPP__
#define __SHTAB_CPP__

// SPHAR headers
#include "shconfig.h"
#include "shtab.h"
#include "shvec.h"
#include "shbasis.h"

__BEGIN_SPHAR_SOURCE

// --------------------------------------------------------------------------
//  CSphTabulated::CSphTabulated()
// --------------------------------------------------------------------------
template<class TV,class TA>
CSphTabulated<TV,TA>::CSphTabulated(int maxOrder, int res, 
                             const CSphBasis<TV,TA>* shBasis)
{
  assert(maxOrder>0);
  assert(res>0);

  if(maxOrder<1 || res<1) {
    _maxOrder     = 0;
    _maxVecLen    = 0;
    _xRes         = 0;
    _yRes         = 0;
    _upper        = 0;
    _upper_dtheta = 0;
    _upper_dphi   = 0;
    _lower        = 0;
    _lower_dtheta = 0;
    _lower_dphi   = 0;
    _defaultVal   = 0;
    return;
  }

  int x,y,i;

  int sqrOrder   = maxOrder*maxOrder;
  _maxOrder  = maxOrder;
  _maxVecLen = PAD_FOUR(sqrOrder);
  _xRes      = res;
  _yRes      = res;

  STATUS << " - precomputig tabulated basis (" << _xRes << "x" << _yRes 
    << ")...";

  int numEntries = _numEntries();

  // allocate default value
  _defaultVal = NEW_ALIGN16( TV, _maxVecLen );
  assert(_defaultVal);
  shv<TV>::setz(_maxVecLen,_defaultVal);

  // allocate upper hemisphere
  _upper        = new TV*[numEntries];
  _upper_dtheta = new TV*[numEntries];
  _upper_dphi   = new TV*[numEntries];
  assert(_upper);
  assert(_upper_dtheta);
  assert(_upper_dphi);
  shv<TV*>::movvs(numEntries,_upper,_defaultVal);
  shv<TV*>::movvs(numEntries,_upper_dtheta,_defaultVal);
  shv<TV*>::movvs(numEntries,_upper_dphi,_defaultVal);

  // allocate lower hemisphere
  if(// 0 
    shBasis->GetDomain() == CSphBasis<TV,TA>::ESphere
    ) {
    _lower        = new TV*[numEntries];
    _lower_dtheta = new TV*[numEntries];
    _lower_dphi   = new TV*[numEntries];
    assert(_lower);
    assert(_lower_dtheta);
    assert(_lower_dphi);
    shv<TV*>::movvs(numEntries,_lower,_defaultVal);
    shv<TV*>::movvs(numEntries,_lower_dtheta,_defaultVal);
    shv<TV*>::movvs(numEntries,_lower_dphi,_defaultVal);
  }
  else {
    _lower = _lower_dtheta = _lower_dphi = 0;
  }

  /// the main cycle
  int   yRes1  = _yRes + 1;
  TA deltax = 2.0/_xRes;
  TA deltay = 2.0/_yRes;
  TA vect[3];

  for(x=0; x <= _xRes; x++) {
    vect[0] = x*deltax-1;

    for(y=0; y <= _yRes; y++) {
      vect[1] =  y*deltay-1;			

      TA sqrDist = shm::sqr2(vect[0]) + shm::sqr2(vect[1]);

      if (sqrDist > TA(1.0)) continue;

      vect[2] = 0.5 - 0.5*sqrDist;
      TA inorm = 1.0 / sqrt(sqrDist + shm::sqr2(vect[2]));
      TA vect1[3] = {vect[0] * inorm, vect[1] * inorm, vect[2] * inorm};

      int index = x*yRes1+y;

      _upper       [index] = NEW_ALIGN16( TV , _maxVecLen );
      _upper_dtheta[index] = NEW_ALIGN16( TV , _maxVecLen );
      _upper_dphi  [index] = NEW_ALIGN16( TV , _maxVecLen );
      assert(_upper       [index]);
      assert(_upper_dtheta[index]);
      assert(_upper_dphi  [index]);
      for(i=sqrOrder; i<_maxVecLen; i++)
        _upper[index][i]=_upper_dtheta[index][i]=_upper_dphi[index][i]=0;
      shBasis->EvalDerivCartV(_maxOrder,vect1,_upper[index],
        _upper_dtheta[index],_upper_dphi[index]);

      if(_lower) {
        vect1[2] = -vect1[2];
        _lower       [index] = NEW_ALIGN16( TV , _maxVecLen );
        _lower_dtheta[index] = NEW_ALIGN16( TV , _maxVecLen );
        _lower_dphi  [index] = NEW_ALIGN16( TV , _maxVecLen );
        assert(_lower       [index]);
        assert(_lower_dtheta[index]);
        assert(_lower_dphi  [index]);
        for(i=sqrOrder; i<_maxVecLen; i++)
          _lower[index][i]=_lower_dtheta[index][i]=_lower_dphi[index][i]=0;
        shBasis->EvalDerivCartV(_maxOrder,vect1,_lower[index],
          _lower_dtheta[index],_lower_dphi[index]);
      }      
    }
  }

  STATUS << "done." << std::endl;
}

// --------------------------------------------------------------------------
//  CSphTabulated::~CSphTabulated()
// --------------------------------------------------------------------------
template<class TV,class TA>
CSphTabulated<TV,TA>::~CSphTabulated()
{
  int i;
  int numEntries = _numEntries();
  if(_upper) {
    for(i=0; i<numEntries; i++) {
      if(_upper[i] != _defaultVal) {
        FREE_ALIGN16(_upper[i]);
        FREE_ALIGN16(_upper_dtheta[i]);
        FREE_ALIGN16(_upper_dphi[i]);
      }
    }
    delete [] _upper;
  }
  if(_lower) {
    for(i=0; i<numEntries; i++) {
      if(_lower[i] != _defaultVal) {
        FREE_ALIGN16(_lower[i]);
        FREE_ALIGN16(_lower_dtheta[i]);
        FREE_ALIGN16(_lower_dphi[i]);
      }
    }
    delete [] _lower;
  }

  FREE_ALIGN16(_defaultVal);
}

// --------------------------------------------------------------------------
//  CSphTabulated::GetCoeffs4()
// --------------------------------------------------------------------------
template<class TV,class TA>
void
CSphTabulated<TV,TA>::GetCoeffs4(int order, const TA *vect, TV *dest,
                          TV *dest_dtheta, TV* dest_dphi) const
{
  assert(order>0);
  assert(order<=_maxOrder);
  assert(dest);

  TA u,v;
  TV **array,**array_dtheta,**array_dphi;
  int sqrOrder = order*order;

  // decide which hemisphere
  if (vect[2] < 0) 
  {
    assert(_lower);
    array        = _lower;
    array_dtheta = _lower_dtheta;
    array_dphi   = _lower_dphi;
    u=TA(0.5)*(TA(1.0)+(vect[0]/(TA(1.0)-vect[2])));
    v=TA(0.5)*(TA(1.0)+(vect[1]/(TA(1.0)-vect[2])));
  }
  else {
    array        = _upper;
    array_dtheta = _upper_dtheta;
    array_dphi   = _upper_dphi;   
    u=TA(0.5)*(TA(1.0)+(vect[0]/(TA(1.0)+vect[2])));
    v=TA(0.5)*(TA(1.0)+(vect[1]/(TA(1.0)+vect[2])));
  }

  if (u<0.0 || u>1 || v<0 || v>1)
    shv<TV>::setz(sqrOrder,dest);
  
  
#if 1
  // NEAREST NEIGHBOR INTERPOLATION 
  int uindex = (int)(u * _xRes + TA(0.4999));
  int vindex = (int)(v * _yRes + TA(0.4999));
  int index = uindex*(_yRes+1)+vindex;
  shv<TV>::movvv(sqrOrder,dest,array[index]);
  if(dest_dtheta) shv<TV>::movvv(sqrOrder,dest_dtheta,array_dtheta[index]);
  if(dest_dphi)   shv<TV>::movvv(sqrOrder,dest_dphi,  array_dphi[index]);
  return;
#else
  //BILINEAR INTERPOLATION
  int uindex = (int)(u * _xRes);
  int vindex = (int)(v * _yRes);
  int vecLen   = PAD_FOUR(order*order);
  
  // get the interpolated vectors
  const TV* s00 = array[ uindex    *(_yRes+1)+vindex  ] ;
  const TV* s10 = array[ (uindex+1)*(_yRes+1)+vindex  ] ;
  const TV* s01 = array[ (uindex)  *(_yRes+1)+vindex+1] ;
  const TV* s11 = array[ (uindex+1)*(_yRes+1)+vindex+1] ;
  
  // compute the interpolation weights
  TV du = fabs(uindex-u*_xRes);
  TV dv = fabs(vindex-v*_yRes);
  
  // perform the interpolation for all channels
  shv<TV>::setz4   (vecLen, dest);
  shv<TV>::madvsip4(vecLen, dest, s00, (1-du)*(1-dv) );
  shv<TV>::madvsip4(vecLen, dest, s10, (  du)*(1-dv) );
  shv<TV>::madvsip4(vecLen, dest, s01, (1-du)*(  dv) );
  shv<TV>::madvsip4(vecLen, dest, s11, (1-du)*(  dv) );
#endif
}

// explicit instatiation
template class CSphTabulated<float,float>;
template class CSphTabulated<float,double>;
template class CSphTabulated<double,float>;
template class CSphTabulated<double,double>;

__END_SPHAR_SOURCE

#endif
